/*
 * Log.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_IO_PKG;

import java.util.Vector;
import java.io.ByteArrayOutputStream;
import java.io.PrintWriter;

IMPORT_DEBUG

/**
 * A Janos Log interface.  A log maintains a buffer of a fixed number
 * of messages.  All messages passed to the log via <code>logMsg()</code>
 * are sent to all listeners on the log.
 * <p>
 * Really just a glorified ring buffer.
 * <p>
 * Patterned after syslog(3).
 */
public class Log
	implements LogLevel
{
	/**
	 * The entities in the ring-buffer.  Each message has its level
	 * recorded along with it.
	 */
	private class Message 
	{
		public int level; // only need 3 bits...
		public String msg;
	}

	/**
	 * The ring buffer.  The msg_ring is never empty.
	 */
	private final Message[] msg_ring;

	/**
	 * Size of the ring buffer.
	 */
	private static final int RING_SIZE = 25;
	private int ring_head, ring_tail;

	/**
	 * Listeners on this log.  Should be Vector<LogListener>.
	 */
	private final Vector listeners;

	/**
	 * Stores the prefix for the log messages
	 */
	private String prefix;

	/**
	 * Used for converting exception stack traces into strings.  
	 * A single global object is used for all traces.
	 *
	 * @see #logThrowable()
	 */
	private final ByteArrayOutputStream byteStream = new ByteArrayOutputStream(250);

	/**
	 * Used for converting exception stack traces into strings.  
	 * A single global object is used for all traces.
	 *
	 * @see #logThrowable()
	 */
	private final PrintWriter byteWriter = new PrintWriter(byteStream);


	/**
	 * Create a log.  All messages sent to this log will
	 * be prefixed with the given prefix string.
	 */
	public Log(String prefix)
	{
		int i;
		
		msg_ring = new Message[RING_SIZE];
		listeners = new Vector(5);
		for( i = 0; i < RING_SIZE; i++ )
		{
			msg_ring[i] = new Message();
		}
		
		this.prefix = prefix;

		// Add a message.  Ensures log is never empty.
		if (prefix == null)
			msg_ring[0].msg = "*** Log Started";
		else 
			msg_ring[0].msg = prefix + " *** Log Started";
		msg_ring[0].level = DBG;
		ring_head = 0;
		ring_tail = 0;
	}

	/**
	 * Set the prefix message.
	 */
	public void setPrefix(String prefix)
	{
		this.prefix = prefix;
	}

	private void broadcastMessage(int level, String msg)
	{
		int i;
		for (i = 0; i < listeners.size(); i++)
		{
			LogListener ll;
			ll = ((LogListener) listeners.elementAt(i));
			
			ll.logMsg(level, msg);
		}
	}
	
	/**
	 * Add a message to this log.  Added to the ring buffer.
	 * All active listeners will recieve the message, too.
	 */
	public void logMsg(int level, String msg)
	{
		// Put the prefix on the message
		if( prefix != null )
			msg = prefix + ":" + msg;

		assert(level >= 0);
		assert(level <= DBG);

		// add to the ring buffer
		ring_tail = (ring_tail + 1) % RING_SIZE;
		msg_ring[ring_tail].level = level;
		msg_ring[ring_tail].msg = msg;
		if (ring_tail == ring_head)
			ring_head = (ring_head + 1) % RING_SIZE;
		
		// pass out to all listeners
		broadcastMessage(level, msg);
	}
	
	public void logMsg(String msg)
	{
		this.logMsg(DBG, msg);
	}

	/**
	 * Add a message to the log thats a stack trace.
	 */
	public void logThrowable(int level, Throwable th)
	{
		String msg;
		
		synchronized(this.byteWriter)
		{
			if( this.prefix != null )
				this.byteWriter.println(prefix);
			th.printStackTrace(this.byteWriter);
			this.byteWriter.flush();
			msg = this.byteStream.toString();
			this.byteStream.reset();
		}
		
		broadcastMessage(level, msg);
	}
	
	public void logThrowable(Throwable th)
	{
		this.logThrowable(DBG, th);
	}

	/**
	 * Add a listener to this log.
	 * The listener will immediately receive all buffered messages
	 * and eventually receive all future messages.
	 */
	public void addListener(LogListener ll)
	{
		if (!listeners.contains(ll))
		{
			listeners.addElement(ll);
			int i = ring_head;
			do
			{
				if( msg_ring[i].msg != null )
					ll.logMsg(msg_ring[i].level,
						  msg_ring[i].msg);
				i = (i + 1) % RING_SIZE;
			}
			while (i != ring_tail);
		}
	}

	public void removeListener(LogListener ll)
	{
		listeners.removeElement(ll);
	}

	public String toString()
	{
		return "Log[msg_ring="
			+ this.msg_ring
			+ "; ring_head="
			+ this.ring_head
			+ "; ring_tail="
			+ this.ring_tail
			+ "; listeners="
			+ this.listeners
			+ "; prefix="
			+ this.prefix
			+ "; byteStream="
			+ this.byteStream
			+ "; byteWriter="
			+ this.byteWriter
			+ "]";
	}
}
