/*
 * Address.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 *
 * NOTE: The toString() and fromString() functions in this class are
 * covered by: 
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 *
 */
package JANOS_NODEOS_PKG;

import java.net.InetAddress;

import java.util.StringTokenizer;
import java.util.NoSuchElementException;

import edu.utah.janosvm.kit.comm.commspace.CommSpaceable;

/**
 * An address object hides the details and implementation of
 * addressing in Janos.  Addresses can only be created through an
 * AddressFactory.
 * <p>
 * An Address has two parts a logical "active network" address and an
 * underlying "physical address".  The physical address in this
 * implementation of Address is the combination of a protocol
 * specification and an address specification; This is similar to what
 * is described in the NodeOS specification for addressing.
 * <p>
 * The "active address" is a 32-bit IP-like number.
 * <p>
 * Address objects are immutable.  They must be immutable because
 * anyone(?) can read them out of routing tables...
 * <p>
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class Address
	implements CommSpaceable
{
	/**
	 * The physical protocol specification.  This defines
	 * the context in which the addressSpec is interpreted.
	 *
	 * @see #addressSpec
	 */
	private final String protocolSpec;

	/**
	 * The physical address specification.
	 *
	 * @see #protocolSpec
	 */
	private final String addressSpec;
	
	/**
	 * The logical level address that maps to the underlying
	 * physical address.
	 * <p>
	 * XXX This is somewhat ANTSR specific, as it assumes a logical
	 * address is a 32-bit integer.
	 */
	private final int anAddress;

	/**
	 * Construct a new Address.  The address has two parts: an
	 * "active address" and a "physical address".  The "active address"
	 * is a 32-bit integer.  The "physical address" consists of the
	 * protocol spec and its associated address spec.
	 *
	 * @param anAddress The virtual active net address
	 * @param protocolSpec The protocol specification for the `physical'
	 *   connection
	 * @param addressSpec The address specification for the protocol
	 */
	PACKAGE Address(int anAddress, String protocolSpec, String addressSpec)
	{
		this.anAddress = anAddress;
		this.protocolSpec = protocolSpec;
		this.addressSpec = addressSpec;
	}

	/**
	 * Construct a new address from the given Address.
	 */
	PACKAGE Address(Address other)
	{
		this.anAddress = other.anAddress;
		this.protocolSpec = other.protocolSpec;
		this.addressSpec = other.addressSpec;
	}

	public int getANAddress()
	{
		return this.anAddress;
	}
	
	public String getProtocolSpec()
	{
		return this.protocolSpec;
	}

	public String getAddressSpec()
	{
		return this.addressSpec;
	}

	public boolean equals(Object obj)
	{
		boolean retval = false;

		if( obj instanceof Address )
		{
			Address addr = (Address)obj;

			if( (addr.anAddress == this.anAddress)
			    && addr.protocolSpec.equals(this.protocolSpec)
			    && addr.addressSpec.equals(this.addressSpec) )
			{
				retval = true;
			}
		}
		return retval;
	}
	
	public int hashCode()
	{
		return this.anAddress;
	}

	public Object clone()
	{
		return new Address(this.anAddress,
				   CloneFactory.clone(this.protocolSpec),
				   CloneFactory.clone(this.addressSpec));
	}

	/**
	 * Deep copy to be performed when this is pulled out of a
	 * CommSpaceElement
	 * <p>
	 * This appears due to the CommSpaceable interface
	 */
	public Object deepCopy()
	{
		return this.clone();
	}
	
	public String toString()
	{
		return "Address[anAddr=" +toString(anAddress)+
			"; ProtocolSpec=" +this.protocolSpec+
			"; AddressSpec=" +this.addressSpec+ "]";
	}

	/**
	 * Generate a string representation of the given 32-bit integer address.
	 * Generates standard dotted-quad notation.
	 */
	public static String toString (int a)
	{
		StringBuffer dotted = new StringBuffer(3*4 + 4);
		dotted.append((a>>24) & 0xFF)
			.append(".")
			.append((a>>16) & 0xFF)
			.append(".")
			.append((a>>8) & 0xFF)
			.append(".")
			.append(a & 0xFF);
		return dotted.toString();
	}

	/**
	 * Generate a 32-bit integer from the given string address.
	 * The string must be a dotted-quad, e.g. "192.168.1.2".
	 *
	 * @param name A dotted-quad string address
	 * @return the 32-bit integer represented by name
	 * @exception AddressFormatException if there aren't 4 parts, any of the 4 parts don't parse,
	 * or are out of range.
	 */
	public static int fromString(String name)
		throws AddressFormatException
	{
		StringTokenizer st = new StringTokenizer(name, ".");
		if (st.countTokens() != 4)
			throw new AddressFormatException("can't parse " +name);
		
		int[] b = new int[4];
		try
		{
			for (int i = 0; i < 4; i++) {
				b[i] = Integer.parseInt(st.nextToken());
				if ((b[i] < 0) || (b[i] > 255))
					throw new AddressFormatException("address part at "
									 +i+ " is out of range in " +name);
			}
		}
		catch (NumberFormatException e) 
		{
			throw new AddressFormatException("non-numeric address: " +name);
		}
		catch (NoSuchElementException _)
		{
			// This should never happen
			panic("trouble in Address.fromString!");
		}
		
		return ((b[0]<<24) | (b[1]<<16) | (b[2]<<8) | b[3]); 
	}

}
