/*
 * AddressSpecPortAddr.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.net.InetAddress;
import java.net.UnknownHostException;

import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 * An implementation of an AddressSpec for UDP and TCP style addresses (e.g. a
 * 16 bit port number coupled with a 32 bit IP address.
 * <p>
 * This is the internal representation of most "AddressSpec"'s.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class AddressSpecPortAddr
	implements AddressSpec
{
	/**
	 * Constant used to indicate that this spec matches any IP address.
	 */
	PACKAGE static final InetAddress WILDCARD_ADDRESS = null;

	/**
	 * Constant used to indicate that the receiver of the spec should use
	 * the source interface's IP address.
	 */
	PACKAGE static InetAddress       SOURCE_ADDRESS;

	/**
	 * Constant used to indicate that this spec matches any UDP/TCP port
	 * number.
	 */
	PACKAGE static final int         WILDCARD_PORT    = -1;

	static {
		try
		{
			SOURCE_ADDRESS = InetAddress.getLocalHost();
		}
		catch(UnknownHostException e)
		{
			Node.printStackTrace(e);
			Node.fatalError("Can't get the local host "
					+ "InetAddress?");
		}
	}
	
	/**
	 * Source internet address for this AddressSpec
	 */
	private InetAddress saddr = WILDCARD_ADDRESS;
	
	/**
	 * Source port this AddressSpec represents.
	 */
	private int sport = WILDCARD_PORT;

	/**
	 * Destination internet address for this AddressSpec
	 */
	private InetAddress daddr = WILDCARD_ADDRESS;
	
	/**
	 * Destination port this AddressSpec represents.
	 */
	private int dport = WILDCARD_PORT;

	/**
	 * @param addr The TCP/UDP address spec to parse.
	 */
	PACKAGE AddressSpecPortAddr(String addr)
		throws PacketAddressFormatException
	{
		this.parseAddress(addr);
	}

	/**
	 * Parses UDP address strings.  The format is:
	 *  "<src address>:<dest port>:<dest address>:<src port>
	 *
	 * @param addr The TCP/UDP address spec to parse.
	 */
	private void parseAddress(String addr)
		throws PacketAddressFormatException
	{
		try
		{
			StringTokenizer st = new StringTokenizer(addr, ":");
			String frag;

			/* Source Address */
			frag = st.nextToken();
			if (frag.equals("$source"))
			{
				/* XXX This is a jnodeos hack. */
				this.saddr = SOURCE_ADDRESS;
			}
			else if (frag.equals("*"))
			{
				this.saddr = WILDCARD_ADDRESS;
			}
			else
			{
				throw new PacketAddressFormatException(
					"Janos cannot generate packets with "
					+ "a custom IP address ("
					+ frag
					+ ", in "
					+ addr
					+ ").");
			}

			/* Destination Port */
			if (!st.hasMoreTokens())
			{
				throw new PacketAddressFormatException(
					"A destination UDP port is required "
					+ "for an OutChannel.");
			}
			frag = st.nextToken();
			if (frag.equals("*"))
				this.dport = WILDCARD_PORT;
			else
				this.dport = Integer.parseInt(frag);

			/* Destination Address */
			if (!st.hasMoreTokens())
			{
				throw new PacketAddressFormatException(
					"A destination IP address is required "
					+ "for an OutChannel.");
			}
			frag = st.nextToken();
			if (frag.equals("*"))
				this.daddr = WILDCARD_ADDRESS;
			else
				this.daddr = InetAddress.getByName(frag);

			/* Source Port */
			if( st.hasMoreTokens() )
			{
				frag = st.nextToken();
				if (frag.equals("*"))
					this.sport = WILDCARD_PORT;
				else
					this.sport = Integer.parseInt(frag);
			}
			else
			{
				/* Assume wildcard */
				this.sport = WILDCARD_PORT;
			}
		}
		catch(NumberFormatException e)
		{
			throw new PacketAddressFormatException
				("Cannot parse port number: "
				 + e.getMessage());
		}
		catch(UnknownHostException e)
		{
			throw new PacketAddressFormatException
				("Cannot parse address " +addr+": " +e);
		}
		catch(NoSuchElementException e)
		{
			throw new PacketAddressFormatException
				("Address " +addr+ " formatted incorrectly.");
		}
	}

	/**
	 * @return The destination address.
	 */
	PACKAGE InetAddress destAddr()
	{
		return daddr;
	}
	
	/**
	 * @return The destination port number.
	 */
	PACKAGE int destPort()
	{
		return dport;
	}

	/**
	 * @return The source address.
	 */
	PACKAGE InetAddress sourceAddr()
	{
		return saddr;
	}
	
	/**
	 * @return The source port number.
	 */
	PACKAGE int sourcePort()
	{
		return sport;
	}

	public String getSource()
	{
		String retval;
		
		if( this.saddr == WILDCARD_ADDRESS )
		{
			retval = "*:" + this.sport;
		}
		else if( this.saddr == SOURCE_ADDRESS )
		{
			retval = "$source:" + this.sport;
		}
		else
		{
			retval = this.saddr.toString() + this.sport;
		}
		return retval;
	}

	public String getDestination()
	{
		String retval;
		
		if( this.daddr == WILDCARD_ADDRESS )
		{
			retval = "*:" + this.dport;
		}
		else
		{
			retval = this.daddr.toString() + this.dport;
		}
		return retval;
	}

	public int hashCode()
	{
		int retval = 0;

		if( this.saddr != WILDCARD_ADDRESS )
			retval = this.saddr.hashCode();
		if( this.daddr != WILDCARD_ADDRESS )
			retval ^= this.daddr.hashCode();
		retval ^= (this.sport << 16) ^ this.dport;
		return retval;
	}

	/**
	 * Two AddressSpecPortAddr objects are equal if the 
	 * source and destination ports and addresses are equal.
	 */
	public boolean equals(Object o)
	{
		boolean retval = false;

		if ( o instanceof AddressSpecPortAddr )
		{
			AddressSpecPortAddr other = (AddressSpecPortAddr)o;

			retval =
				(other.sport == this.sport)
				&& (other.dport == this.dport)
				&& (other.saddr.equals(this.saddr))
				&& (other.daddr.equals(this.daddr));
		}
		return( retval );
	}
	
	public String toString()
	{
		return "AddressSpecPortAddr[sa="
			+ ((this.saddr == WILDCARD_ADDRESS) ?
			   "*" : ((this.saddr == SOURCE_ADDRESS) ?
				  "$source" : this.saddr.toString()))
			+ "; sp="
			+ this.sport
			+ "; da="
			+ ((this.daddr == WILDCARD_ADDRESS) ?
			   "*" : this.daddr.toString())
			+ "; dp="
			+ this.dport
			+ "]";
	}
}
