/*
 * DemultiplexKey.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * A key for matching incoming packets.
 *
 * @see InChannel
 * @see CutThroughChannel
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class DemultiplexKey
	extends NATIVE_DEMULTIPLEXKEY
{
	CHECK_NATIVE_NODEOS

	/**
	 * Create a new demultiplex key.  By default, the key
	 * is empty (and thus it matches everything).
	 */
	public DemultiplexKey(Domain domain)
	{
		super(domain.getNativeDomain());
		require(Domain.currentDomain() == domain);
	}

	/**
	 * Add a segment that matches the given sequence of bytes,
	 * skipping offset bytes from the previous segment.  The mask
	 * says which <emph>bytes</emph> of sequence are relevant.  If
	 * mask is null, all bytes are relevant.
	 * <p>
	 * XXX currently mask must be null.
	 * <p>
	 * XXX should throw a custom exception for bogon parameters (vs. Error).
	 *
	 * @param length The length of the bytes to match
	 * @param sequence The sequence of bytes to match
	 * @param mask If non-null, its a bit mask that indicates which bytes
	 * in `sequence' should be matched.
	 */
	public void add(int length, byte sequence[], byte mask[])
	{
		if ((length <= 0)
		    || ((sequence != null) && (length > sequence.length))
		    || ((mask != null) && (mask.length != length)))
			throw new Error("bad match parameters"); // XXX need real exception
		
		// XXX other sanity checks?

		super.add(length, sequence, mask);
	}

	/**
	 * Add a segment matching the given byte offset bytes from
	 * the last segment.
	 * <p>
	 * Equivalent to:
	 * <pre>
	 *    byte[] barray = new byte[1];
	 *    barray[0] = b;
	 *    match(offset, 1, barray, null);
	 * </pre>
	 */
	public void add(byte b)
	{
		super.addByte(b);
	}

	/**
	 * Add a segment matching the given short offset bytes from
	 * the last segment.
	 * <p>
	 * XXX currently does not convert s to network order.  Should it?
	 */
	public void add(short s)
	{
		/* XXX convert s to network order? */

		super.addShort(s);
	}

	/**
	 * Add a segment matching the given int offset bytes from
	 * the last segment.
	 * <p>
	 * XXX currently does not convert i to network order.  Should it?
	 */
	public void match(int i)
	{
		/* XXX convert i to network order? */

		super.addInt(i);
	}

	/**
	 * Add a segment matching the given long offset bytes from
	 * the last segment.
	 * <p>
	 * XXX currently does not convert l to network order.  Should it?
	 */
	public void match(long l)
	{
		/* XXX convert l to network order? */

		super.addLong(l);
	}

	public static final int ETH_ANYSADDR = 1;
	public static final int ETH_ANYDADDR = 2;
	public static final int ETH_ANYETYPE = 4;
	public static final int ETH_ALLFLAGS = 7;
	
	/**
	 * Add a segment to match an Ethernet header.
	 * <p>
	 * @param flags The flags that indicate which parameters should be
	 *   ignored (i.e. wildcards).
	 * @param src The source address to match.
	 * @param dst The destination address to match.
	 * @param type The type of ethernet packet to match.
	 */
	public void addETH(int flags,
			   byte src[], byte dst[], short type)
	{
		if (((flags & ~ETH_ALLFLAGS) != 0)
		    || ((src != null) && (src.length != 6))
		    || ((dst != null) && (dst.length != 6)))
			throw new Error("bad match parameters");
		
		super.addETH(flags, src, dst, type);
	}
	
	public static final int IPV4_ANYSADDR = 1;
	public static final int IPV4_ANYDADDR = 2;
	public static final int IPV4_ANYPROTO = 4;
	public static final int IPV4_ALLFLAGS = 7;
	
	/**
	 * Add a segment to match an IPv4 header.  Offset is the offset
	 * from the end of the previous segment (or from the beginning of the
	 * buffer for the first segment).
	 * <p>
	 * Use this instead of the vanilla match() to get a "payloadOffset"
	 * in the deliver func that will skip IP options.  (A vanilla
	 * match() call will leave payloadOffset in the first byte after
	 * the "protocol" byte, which is not necessairly the first byte
	 * of the payload in an IP packet (e.g., one with options).
	 * <p>
	 * @param flags The flags that indicate which parameters should be
	 *   ignored (i.e. wildcards).
	 * @param source the 32-bit IPv4 source address to match in the header.
	 * @param dest the 32-bit IPv4 destination address to match in the header.
	 * @param protocol the 16-bit IPv4 protocol identifier.
	 */
	public void addIPv4(int flags,
			    int source, int srcMask,
			    int dest, int dstMask,
			    byte protocol)
	{
		if (((flags & ~IPV4_ALLFLAGS) != 0))
			throw new Error("bad match parameters"); // XXX need real exception
		
		// XXX other sanity checks?

		super.addIPv4(flags,
			      source, srcMask,
			      dest, dstMask,
			      protocol);
	}

	/**
	 * Add a segment to match a UDP header, needs to be preceded by a
	 * call to <code>matchIPv4</code>.
	 */
	public void addUDP(int srcPort, int destPort)
	{
		if ((srcPort < -1)
		    || (srcPort > (64 * 1024))
		    || (destPort < -1)
		    || (destPort > (64 * 1024)))
			throw new Error("bad matchUDP() parameters"); // XXX need real exception
		
		// XXX other sanity checks?
		
		super.addUDP(srcPort, destPort);
	}

	/**
	 * Add a segment to match a TCP header, needs to be preceded by a
	 * call to <code>matchIPv4</code>.
	 */
	public void addTCP(int srcPort, int destPort)
	{
		if ((srcPort < -1)
		    || (srcPort > (64 * 1024))
		    || (destPort < -1)
		    || (destPort > (64 * 1024)))
			throw new Error("bad matchUDP() parameters"); // XXX need real exception
		
		// XXX other sanity checks?

		super.addTCP(srcPort, destPort);
	}
	
	/**
	 * Add a segment that matches an ANEP header.
	 *
	 * @param version The ANEP version number to match
	 * @param allVersions Match any version of ANEP
	 * @param protocolID The protocolID to match
	 */
	public void addANEP(byte version,
			    boolean allVersions, short protocolID)
	{
		if ((version < 0))
			throw new Error("bad matchUDP() parameters"); // XXX need real exception
		
		// XXX other sanity checks?
		super.addANEP(version, allVersions, protocolID);
	}

	/**
	 * Test a packet to see if this key would match that packet.
	 * <p>
	 * XXX payloadLength has odd semantics.  Either explain why, or fix
	 * the callers of this method.
	 *
	 * @param payload The packet data
	 * @param payloadStart The place to start matching in `payload'
	 * @param payloadLength The length of data in payload, including bytes
	 * up to payloadStart.
	 *
	 * @return 0 if there was an error otherwise the index of the first byte after the
	 * end of the key.
	 */
	public final int matches(byte payload[],
				 int payloadStart,
				 int payloadLength)
	{
		if ((payloadStart < 0)
		    || (payloadStart > payload.length)
		    || (payloadStart + payloadLength > payload.length)
		    || (payloadLength > payload.length))
			throw new Error("Bad payload for matches()"); // XXX need real exception
		
		return super.matches(payload, payloadStart, payloadLength);
	}

	/**
	 * Get the "length" of this key.  The "length" of a key is the
	 * distance between the lowest absolute offset the key matches
	 * and the highest byte the key matches.  For keys that include
	 * variable length components (e.g., addANEP() or addIPv4() were used),
	 * the value returned is just a guess.
	 */
	public final int length()
	{
		return super.length();
	}
	
	PACKAGE NATIVE_DEMULTIPLEXKEY getNativeKey()
	{
		return this;
	}

	public String toString()
	{
		return super.toString();
	}
}
