/*
 * Domain.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;

import JVM_SYS_PKG.ConfigFile;
import JVM_SYS_PKG.DeadTeamException;

import JVM_RESOURCES_PKG.ClassPathResource;

import java.io.IOException;

/**
 * Java wrapper for a domain.  This is the current domain's local
 * interface to its own domain object.
 * <p>
 * If you know BeOS or V, this is like a "Team".  If you prefer Mach
 * (or know what Fluke is), this is similar to a "Task".  If all you
 * know is Unix, then think of this as a "process".  You might also
 * call it a "resource container" or possibly a "protection domain."
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class Domain
	extends NATIVE_DOMAIN
{
	CHECK_NATIVE_NODEOS

	/**
	 * The Domain ClassLoader for this domain.
	 * <p>
	 * The loader also acts as a flag for a destroyed domain.  If
	 * loader is null, the domain is destroyed.  Note that loader is
	 * null until the first thread in the new domain is started, so
	 * don't rely on destroyed() until the domain actually running.
	 *
	 * @see DomainInitializer#run()
	 */
	private DomainClassLoader loader;

	/**
	 * Cached reference to the kernel's domain handle that points
	 * to this domain.
	 */
	PACKAGE DomainHandle kernelHandle;

	/**
	 * Collection of values local to this domain.
	 */
	private final Hashtable locals = new Hashtable();
	
	/**
	 * Create a new Domain, which is a child of the current Domain.
	 * Starts a new thread executing within the new domain.
	 * <p>
	 * This is the only interface for creating a Domain.
	 * <p>
	 * The termHandler will be invoked by a thread in the parent
	 * Domain's if the child Domain is terminated.  This handler has a
	 * short limit for execution.
	 * <p>
	 * If termHandler is null, then no termination handler is called.
	 * <p>
	 * XXX Should be called "create()" to sync back with NodeOS spec.
	 * <p>
	 * XXX don't overload Runnable for the termHandler.  Use a different,
	 * jnodeos interface that doesn't overlap with the other callback interfaces.
	 *
	 * @param cred The principal associated with this Domain
	 * @param tPoolSpec The specification for the new Domain's ThreadPool
	 * @param memSpec The memory limit for this domain
 	 * @param domainStartSpec Class loading and startup information
	 * @param termHandler Termination handler invoked by the NodeOS when
	 * an immediate child domain is being terminated.
	 * @return a DomainHandle onto the new child domain.
	 */
	public DomainHandle newDomain(Credentials cred,
				  ThreadPoolSpec tPoolSpec,
				  MemSpec memSpec,
				  DomainStartSpec domainStartSpec,
				  Runnable termHandler,
				  DomainClassSpec fcs)
		throws DomainStartException
	{
		require(cred);
		require(tPoolSpec);
		require(memSpec);
		require(domainStartSpec);
		// XXX require(termHandler);
		
		TRACE(Trace.NODEOS_DOMAINSTART, "Domain.newDomain()");
		
		return super.createDomain(cred,
					tPoolSpec,
					memSpec,
					domainStartSpec,
					termHandler,
					fcs);
	}

	public DomainHandle newDomain(Credentials cred,
				  ThreadPoolSpec tPoolSpec,
				  MemSpec memSpec,
				  DomainStartSpec domainStartSpec,
				  Runnable termHandler)
		throws DomainStartException
	{
		return this.newDomain(cred,
				    tPoolSpec,
				    memSpec,
				    domainStartSpec,
				    termHandler,
				    DomainClassSpec.getDefault());
	}

	/**
	 * Just like 'newDomain', but for the root domain which may require
	 * special magic in the backend.
	 *
	 * @see #newDomain
	 */
	PACKAGE static DomainHandle createRootDomain(Credentials cred,
						 ThreadPoolSpec tPoolSpec,
						 MemSpec memSpec,
						 DomainStartSpec domainStartSpec)
		throws DomainStartException
	{
		return NATIVE_DOMAIN.createRootDomain(cred,
						  tPoolSpec,
						  memSpec,
						  domainStartSpec);
	}

	/**
	 * Initalize a new Domain.  Does not "start" the domain.
	 * Registers the domain in the global DomainRegistry.
	 * <p>
	 * Only invoked via DomainHandle's constructor.  See newDomain
	 * <p>
	 * This is invoked in the context of the new domain by a visiting
	 * thread from the parent domain.
	 *
	 * @param cred the credentials for the new domain. XXX.
	 * @param tPoolspec the thread pool spec for the first thread pool in the new domain
	 * @param memSpec memory spec for the new domain
	 * @param termHandler callback (in parent!) to be invoked when domain is nuked.
	 * @param fcs The set of classes this domain will be allowed to access.
	 * @see #newDomain()
	 * @see DomainHandle
	 */
	PACKAGE Domain(final Credentials cred,
		     final ThreadPoolSpec tPoolSpec,
		     final MemSpec memSpec,
		     final Object termHandler,
		     final DomainClassSpec fcs)
	{
		super(cred.getNativeCreds(),
		      tPoolSpec.getNativeSpec(),
		      memSpec.getNativeSpec(),
		      termHandler);

		TRACE(Trace.NODEOS_DOMAINSTART, "Domain.<init>(cred=" +cred+ ")");

		this.loader = new DomainClassLoader(this, fcs);

		// register() allocates a new DomainHandle, too
		this.kernelHandle = DomainRegister.register(this);
	}

	PACKAGE Hashtable getLocals()
	{
		return this.locals;
	}

	/**
	 * Start domain shutdown/cleanup for this domain.
	 * Unregisters domain from the global DomainRegistry.
	 */
	public synchronized void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);

		/*
		 * Run the Domain termination callback in the parent.
		 * Blocks until the term handler finishes.
		 */
		super.kickTermHandler();
		
		// DomainRegister.unregister() is synchronized internally
		DomainRegister.unregister(this);

		// Dump any IP port numbers reserved by this domain.
		IPPortDelegate.releasePorts();
		
		// loader is the flag for redundant calls to destroy
		if (loader != null)
		{
			loader.destroy();
			loader = null;

			super.destroy();
		}
	}

	public synchronized boolean destroyed()
	{
		// destroyed() is only reliable for threads inside the domain.
		assert(this == Domain.currentDomain());
		
		// XXX ???
		return (loader == null);
	}
	
	/**
	 * Create an explicit thread.  This is a new thread which is
	 * not subject to whatever concurrency limit this Domain's
	 * ThreadPool has.  But, it will be scheduled by the
	 * ThreadPool's scheduler.
	 * 
	 * @allocations: This function will allocate a new Thread object.
	 * @param r the Runnable to start the target thread on.
	 * @param name the String name of the new thread.
	 */
	public void threadStart(java.lang.Runnable r, String name)
	{
		super.threadStart(r, name);
	}

	public Object getObject()
	{
		return super.getObject();
	}

	public void setObject(Object obj)
	{
		super.setObject(obj);
	}

	public String toString()
	{
		return "Domain["
			+ "name=" + this.getName() + ";"
			+ super.toString()
	                + "]";
	}

	/**
	 * Add a new class to an existing domain.
	 */
	public void addClass(String name, byte[] bytecode)
	{
		assert(this == Domain.currentDomain());
		assert(name != null);
		assert(bytecode != null);
		assert(loader);
		loader.addClass(name, bytecode);
	}
	
	/**
	 * Add a package to the list of "system" packages.  System
	 * packages will be loaded by the system class loader.
	 * All other classes will be loaded by the per-domain
	 * classloader.
	 * <p>
	 * All sub-domains *share* the same definition of "system"
	 * packages for now.  This is because only one EE System
	 * is loaded on the NodeOS.  This will have to change.
	 * <p>
	 * Implicitly allows the package.
	 */
	public void addSystemPackage(String name)
	{
		assert(this == Domain.currentDomain());
		assert(name != null);
		assert(loader != null);

		loader.addSystemPackage(name);
	}

	/**
	 * Seal a package so that no other classes can be added to it.
	 *
	 * @param name The name of the package to seal.
	 */
	public void sealPackage(String name)
	{
		assert(this == Domain.currentDomain());
		assert(name != null);
		assert(loader != null);

		loader.sealPackage(name);
	}

	/**
	 * Export a class from the class loader to the domain.
	 *
	 * @param name The name of the class to export.
	 */
	public void exportClass(String name)
	{
		assert(this == Domain.currentDomain());
		assert(name != null);
		assert(loader != null);

		loader.exportClass(name);
	}

	/**
	 * Enable access controls on the class loader.
	 */
	public void enableAccessControls(boolean doit)
	{
		assert(this == Domain.currentDomain());
		assert(loader != null);

		loader.enableAccessControls(doit);
	}

	public Class loadClass(String name)
		throws java.lang.ClassNotFoundException
	{
		require(name != null);
		require(loader);
		
		return this.loader.loadClass(name);
	}

	PACKAGE Class loadClass(String name, boolean resolve)
		throws java.lang.ClassNotFoundException
	{
		require(name != null);
		require(loader);
		
		return this.loader.loadClass(name, resolve);
	}

	/**
	 * XXX Is this evil?  Should it just be loadFile from class path?
	 *
	 * XXX should never return null.
	 */
	public static byte[] readByteCode(String classname)
		throws ClassNotFoundException
	{
		byte[] bytes = null;

		require(classname);

		String cname = classname.replace('.', '/').concat(".class");

		try
		{
			bytes = ClassPathResource.loadFile(cname);
		}
		catch(DeadTeamException e)
		{
		}
		catch(IOException ioe)
		{
		}

		if (bytes == null)
			throw new ClassNotFoundException("Could not find bytecode for "
							 + classname
							 + ".");
		return bytes;
	}

	public void getStats(DomainStats fs)
	{
		super.getStats(fs.getNative());
	}
	
	/**
	 * Lookup something in the JanosVM config file.
	 * XXX This is a hack to make it easy to get config parameters.
	 */
	public static String lookupConfig(String section,
					  String tag,
					  int instance)
	{
		return ConfigFile.lookup(section, tag, instance);
	}

	/**
	 * Lookup something in the JanosVM config file.
	 * XXX
	 */
	public static String lookupConfig(String section, String tag)
	{
		return lookupConfig(section, tag, 0);
	}
	
	/**
	 * Get the current domain.  The current Domain is the
	 * domain associated with the current Thread.
	 */
	public final static Domain currentDomain()
	{
		NATIVE_DOMAIN ndomain;

		ndomain = NATIVE_DOMAIN.native_currentDomain();
		assert((ndomain == null) || (ndomain instanceof Domain));
		return (Domain)ndomain;
	}

	PACKAGE NATIVE_DOMAIN getNativeDomain()
	{
		return this;
	}

#ifdef DEBUG
	// always non-null.
	private String name = "Unnamed Domain";
#endif /*DEBUG*/
	
	public void setName(String name)
	{
#ifdef DEBUG
		this.name = name;
#endif /*DEBUG*/	
	}

	public String getName()
	{
#ifdef DEBUG
		return name;
#else
		return null;
#endif /*DEBUG*/
	}
}

