/*
 * DomainClassLoader.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.io.IOException;

import JVM_SYS_PKG.DeadTeamException;

import JVM_RESOURCES_PKG.NameSpaceResource;
import JVM_RESOURCES_PKG.ClassPathResource;

/**
 * Loads class files for a Domain.  This ClassLoader is critical to the security 
 * of the system.  It must prevent dynamic code from installing
 * untrusted code in trusted packages or overriding trusted code.
 * <p>
 * There are three kinds of classes that a Domain is concerned with:
 * system classes, local classes and dynamic classes.  All are
 * optionally subject to package access control checks.
 * <p>
 * System classes are classes that get loaded once, and get loaded by
 * the JVM's root class loader.  The "system" classes are defined on a
 * package basis.  See Main for the code that initializes the set of
 * system packages (this include the base Java packages: java.lang,
 * java.util; and the various Janos NodeOS classes: janos.nodeos,
 * janos.util, etc.)  The set of system packages is recorded in the
 * DomainClassSpec object associated with this loader.
 * <p>
 * Local classes are classes that get loaded off of the node's local
 * disk, but get loaded into each Domain.  These consist of Applications
 * loaded at end nodes.  And maybe some day extensions will fall in
 * this category(?).  Like system classes, local classes are
 * categorized on a package basis.  The set of local packages is
 * implicitly those that are neither dynamic nor system packages.
 * <p>
 * Dynamic classes are classes that are installed via a CodeGroup.
 * They exclusively apply to the unit of code that is shipped around.
 * <p>
 * Aside from these classifications, access to classes can be
 * controlled.  By default domain access control is disabled.  Once
 * enabled, then only system classes and local classes listed in the
 * exported class table will be made visible to the domain.  This
 * prevents Domains from accessing system classes they shouldn't have
 * access to.
 * <p>
 * XXX Not sure if this is correct wrt the new JDK1.2 ClassLoader
 * semantics (e.g., should we be overriding findClass??).
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class DomainClassLoader 
	extends ClassLoader
{
	/**
	 * A mapping from classname to bytecode.  This is the set of
	 * classes that are loaded over the net into this ClassLoader.
	 */
	private final Hashtable dynamicClasses = new Hashtable();

	/**
	 * The set of classes that this loader can access.
	 */
	private final DomainClassSpec classSpec;

	/**
	 * Determines whether the class loader should enforce the above
	 * class spec.
	 */
	private boolean accessControl = false;

#ifdef DEBUG
	/**
	 * For coherent debugging, need to know what domain I'm
	 * associated with.
	 */
	private final Domain domain;
#endif

	/**
	 * Create a DomainClassLoader.  The domain parameter is only
	 * used in debugging builds.
	 *
	 * @param domain the Domain this loader will be associated with
	 */
	PACKAGE DomainClassLoader(Domain domain, DomainClassSpec fcs)
	{
		require(domain);
		DEBUGONLY(this.domain = domain);
		this.classSpec = (DomainClassSpec)fcs.clone();
	}

	/**
	 * Add a new class to an existing domain.  This should
	 * be called from inside the domain that the class is
	 * being added to.
	 * <p>
	 * Neither argument may be null.
	 *
	 * @param name the FQN of the class to be added
	 * @param bytecode the bytecode representation of that class
	 */
	public void addClass(String name, byte[] bytecode)
	{
		String pkgName;
		int offset;
		
		TRACE(Trace.NODEOS_CLASSLOAD, "Adding class " +name+ " to domain " + domain);
		assert(name != null);
		assert(bytecode != null);

		offset = name.lastIndexOf('.');
		if( (offset == 0) || (offset == -1) )
		{
			throw new SecurityException(
				"Can't add class "
				+ name
				+ " to the root package");
		}
		pkgName = name.substring(0, offset);
		if( /* XXX ? */ !this.classSpec.isExportedClass(name) &&
		    this.classSpec.isSealedPackage(pkgName) )
		{
			throw new SecurityException(
				"Can't add class "
				+ name
				+ " to a sealed package");
		}
		dynamicClasses.put(name, bytecode);
	}

	/**
	 * Seal a package so that more classes cannot be added to it.
	 */
	public void sealPackage(String pkgName)
	{
		require(pkgName != null);

		this.classSpec.sealPackage(pkgName);
	}

	/**
	 * Add a system package to the loader.
	 */
	public void addSystemPackage(String pkgName)
	{
		require(pkgName != null);
		
		TRACE(Trace.NODEOS_CLASSLOAD, "Adding system package " +pkgName+ " in domain " + domain);
		this.classSpec.addSystemPackage(pkgName);
	}
	
	/**
	 * Allow access to a Java package from this ClassLoader.
	 */
	public void exportClass(String className)
	{
		require(className != null);
		
		TRACE(Trace.NODEOS_CLASSLOAD, "Allowing class " +className+ " in domain " + domain);
		this.classSpec.exportClass(className);
	}

	/**
	 * Turn package access controls on or off.
	 */
	public void enableAccessControls(boolean doit)
	{
		if (doit)
		{
			TRACE(Trace.NODEOS_CLASSLOAD, "Enabled access controls in domain " + domain);
		}
		else
		{
			TRACE(Trace.NODEOS_CLASSLOAD, "Disabled access controls in domain " + domain);
		}
		
		accessControl = doit;
	}

	/**
	 * Read the bytecode for the given class off the local file system.
	 * <p>
	 * This uses the JanosVM's (real/emulated) ClassPathResource to load
	 * the file so you're subject to its access restrictions/implementation
	 * limitations.
	 *
	 * @param classname The class to read off the disk.
	 * @return The byte code for the class.
	 * @exception ClassNotFoundException thrown if the class is not found on
	 *    disk.
	 */
	private static byte[] readLocalBytecode(String classname)
		throws ClassNotFoundException
	{
		require(classname);

		// XXX no cache, as all the caching is done elsewhere...
		// check in cache first
		// byte [] p = (byte [])classcache.get(classname);
		//
		// if (p != null)
		//	return p;
		
		String cname = classname.replace('.', '/').concat(".class");
		byte[] bytes = null;

		try
		{
			bytes = ClassPathResource.loadFile(cname);
		}
		catch(DeadTeamException e)
		{
		}
		catch(IOException ioe)
		{
			Node.println("Caught IOException? " + ioe);
			Node.printStackTrace(ioe);
		}

		if (bytes == null)
			throw new ClassNotFoundException("Could not find bytecode for "
							 + classname
							 + ".");
		return bytes;
	}

	/**
	 * Look for the given class on the local filesystem.
	 */
	private Class readLocalClass(String name)
		throws ClassNotFoundException
	{
		byte[] bytes;

		require(name);

		TRACE(Trace.NODEOS_CLASSLOAD, "Reading class " +name+ " from disk in domain " + domain);

		bytes = readLocalBytecode(name);
		assert(bytes != null);
		return defineClass(name, bytes, 0, bytes.length);
	}

	/**
	 * @see #loadClass(String,boolean)
	 */
	public Class loadClass(String name)
		throws java.lang.ClassNotFoundException
	{
		Class c;
		
		c = this.loadClass(name, true);

		return c;
	}

	/**
	 * The standard ClassLoader loadClass interface.
	 * <p>
	 * This loadClass looks for already-loaded classes, looks in the
	 * local dynamicClasses, and then checks the package name, then
	 * looks in the system class loader.
	 * <p>
	 * XXX Sealed packages may or may not be a good idea for restricting
	 * the ability to add classes to a package.
	 * <p>
	 * XXX also, there is a problem that defineClass() is called before
	 * the package access check is done.  Thus <clinit> might run before
	 * the access control check is made.  That would be bad.
	 */
	public synchronized Class loadClass(String name, boolean resolve)
		throws ClassNotFoundException
	{
		Class cl;

		require(name);

		TRACE(Trace.NODEOS_CLASSLOAD, "Loading class " +name+ " in domain " + domain);

		// If its already loaded, just get it and get out.
		cl = findLoadedClass(name);
		if (cl != null)
		{
			TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " was already loaded.");
			// Fall through to resolve and return
		}
		else
		{
			String pkgName;
			int offset;
			
			TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " not already loaded in domain " + domain);

			// Figure out if it has a package
			offset = name.lastIndexOf('.');
			if (offset == 0 || offset == -1)
			{
				TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " is in the no-name package.");

				// XXX No-name package is always inaccessible?
				throw new ClassNotFoundException(name + ": The no-name package is always inaccessible.");
			}
			
			pkgName = name.substring(0, offset);
			TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " is in the '" +pkgName+ "' package (accessControl="+accessControl+").");
			if (this.classSpec.isSystemPackage(pkgName))
			{
				if (accessControl &&
				    !this.classSpec.isExportedClass(name))
				{
					TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " is an un-exported system class (denied).");

					throw new ClassNotFoundException(
						"Access to system class "
						+ pkgName
						+ " is not enabled.  "
						+ "Cannot load class "
						+ name);
				}

				TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " is a system class, loading in domain " +domain);
				// TRACE(Trace.NODEOS_CLASSLOAD, "  (via loader: " +this.getClass().getClassLoader()+ ")");

				/*
				 * System packages are loaded by the system
				 * class loader.
				 */
				// cl = this.getClass().getClassLoader().findClass(name);
				// cl = this.getClass().getClassLoader().loadClass(name, resolve);
				// cl = this.getClass().getClassLoader().loadClass(name);
				// cl = Class.forName(name, true, this.getClass().getClassLoader());
				// cl = Class.forName(name);
				cl = NameSpaceResource.loadTeamClass(name);

				TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " loaded via " +cl.getClassLoader());

				ensure(cl != null); // throws otherwise

				// Fall through to resolve and return
			}
			else
			{
				TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " seems to be dynamic.");

				// See if such a class was dynamically installed
				byte[] b = (byte[]) dynamicClasses.get(name);
				if(b != null)
				{

					// XXX package control doesn't work here...

					// XXX May throw weird exceptions/errors, no?
					cl = defineClass(name, b, 0, b.length);
					assert(cl != null);
					TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " loaded from dynamicClasses.");
				
					// XXX Can drop bytecode from dynamicClasses now...

					// fall through to resolve and return
				}
				else
				{
					assert(cl == null);

					TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " not found in dynamicClasses hash.");
					/*
					 * Check that the package is accessible.
					 */
					if (accessControl &&
					    !this.classSpec.isExportedClass(name))
					{
						TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " is an un-exported dynamic class (denied).");
						throw new ClassNotFoundException("Access to the " +name+ " class is not enabled.");
					}

					cl = readLocalClass(name);
					assert(cl != null); // throws otherwise

					TRACE(Trace.NODEOS_CLASSLOAD, "... class " +name+ " loaded from disk.");
				}
				
				ensure(cl != null);
			}

			ensure(cl != null);
		}
		
		ensure(cl != null);
		
		if (resolve)
			resolveClass(cl);

		return cl;
	}

	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
	}

#ifdef DEBUG
	/**
	 * When debugging is enabled, the toString()
	 * method goes to some lengths to avoid printing
	 * the object default identifier (i.e, @82e9350)
	 * because that screws up the regression tests.
	 */
	private static int dbg_currentIdNumber = 0;
	private int dbg_idNumber = dbg_currentIdNumber++;

	public String toString()
	{
		return DomainClassLoader.class.getName() + "#" + dbg_idNumber;
	}
#else
	public String toString()
	{
		return "DomainClassLoader@0x" + Integer.toHexString(this.hashCode());
	}
#endif
}
