/*
 * DomainClassSpec.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import JVM_RESOURCES_PKG.NameSpaceResource;

/**
 * A specification for which classes a domain can access.
 */
public final class DomainClassSpec
{
	/**
	 * Default set of java classes that should be exported to untrusted
	 * code.
	 */
	public static final String defaultJavaClasses[] = {
		/* Most exceptions... */
		"java.lang.Throwable",
		"java.lang.Exception",
		"java.lang.Error",
		"java.lang.ClassNotFoundException",
		"java.lang.InstantiationException",
		"java.lang.InternalError",
		"java.lang.InterruptedException",
		"java.lang.SecurityException",
		"java.lang.NumberFormatException",
		"java.lang.StringIndexOutOfBoundsException",
		"java.lang.CloneNotSupportedException",
		"java.lang.UnsupportedOperationException",

		/* Utility classes */
		"java.lang.Byte",
		"java.lang.Boolean",
		"java.lang.Math",
		"java.lang.Integer",
		"java.lang.Long",
		"java.lang.Object",
		"java.lang.String",
		"java.lang.StringBuffer",
		"java.lang.Cloneable",
		"java.lang.Runnable",

		/* Bah, this one shouldn't be visible... */
		"java.io.IOException",

		"java.util.Hashtable",
		"java.util.Enumeration",
		"java.util.Vector",
		"java.util.NoSuchElementException",
	};

	/**
	 * Default set of classes that are considered "system" code.
	 */
	public static final String defaultSystemPackages[] = {
		"java.lang",
		"java.io",
		"java.util",
		"java.security",
		"java.security.spec",
		JANOS_NODEOS_PKG_S,
		JANOS_UTIL_PKG_S,
		JANOS_ROUTE_PKG_S,
		JANOS_IO_PKG_S,
	};

	/**
	 * Default set of packages that should be "sealed" so that untrusted
	 * code can't add any classes to them.
	 */
	public static final String defaultSealedPackages[] = {
		"java.lang",
		"java.io",
		"java.util",
		"java.security",
		"java.security.spec",
		JANOS_NODEOS_PKG_S,
		JANOS_UTIL_PKG_S,
		JANOS_ROUTE_PKG_S,
		JANOS_IO_PKG_S,
	};

	/**
	 * Reference to the default spec that should be used by new domains.
	 */
	private static DomainClassSpec defaultSpec;

	/**
	 * The set of packages that contain "system" code.
	 */
	private final Hashtable systemPackages = new Hashtable();

	/**
	 * The set of classes that are safe to export to untrusted code.
	 */
	private final Hashtable exportedClasses = new Hashtable();

	/**
	 * The set of packages which should be "sealed" so that untrusted
	 * code can't add any classes to them.
	 */
	private final Hashtable sealedPackages = new Hashtable();

	private final Vector eeNames = new Vector();
	
	public DomainClassSpec()
	{
		/*
		 * XXX The Kaffe hash table will cache an object when you call
		 * these functions, so we do it here to prevent any segfaults.
		 */
		this.systemPackages.elements();
		this.exportedClasses.elements();
		this.sealedPackages.elements();
	}

	public synchronized void addSystemPackage(String pkgName)
	{
		if( !this.isSystemPackage(pkgName) )
		{
			this.systemPackages.put(pkgName, pkgName);
		}
	}

	public synchronized boolean isSystemPackage(String pkgName)
	{
		boolean retval = false;
		
		if( this.systemPackages.get(pkgName) != null )
			retval = true;
		return retval;
	}

	public synchronized void exportClass(String className)
	{
		if( !this.isExportedClass(className) )
		{
			this.exportedClasses.put(className, className);
		}
	}

	public synchronized boolean isExportedClass(String className)
	{
		boolean retval = false;

		if( this.exportedClasses.get(className) != null )
			retval = true;
		return retval;
	}

	public synchronized void sealPackage(String pkgName)
	{
		if( !this.isSealedPackage(pkgName) )
		{
			this.sealedPackages.put(pkgName, pkgName);
		}
	}

	public synchronized boolean isSealedPackage(String pkgName)
	{
		boolean retval = false;
		
		if( this.sealedPackages.get(pkgName) != null )
			retval = true;
		return retval;
	}

	public synchronized void addServerName(String name)
	{
		this.eeNames.addElement(name);
	}

	PACKAGE synchronized Vector getServerNames()
	{
		return this.eeNames;
	}

	public DomainClassSpec absorb(DomainClassSpec fcs)
	{
		Enumeration elems;
		int lpc, size;
		String str;
		
		elems = fcs.systemPackages.elements();
		while( elems.hasMoreElements() )
		{
			str = (String)CloneFactory.
				clone((String)elems.nextElement());
			this.addSystemPackage(str);
		}
		elems = fcs.exportedClasses.elements();
		while( elems.hasMoreElements() )
		{
			str = (String)CloneFactory.
				clone((String)elems.nextElement());
			this.exportClass(str);
		}
		elems = fcs.sealedPackages.elements();
		while( elems.hasMoreElements() )
		{
			str = (String)CloneFactory.
				clone((String)elems.nextElement());
			this.sealPackage(str);
		}
		size = fcs.eeNames.size();
		for( lpc = 0; lpc < size; lpc++ )
		{
			this.eeNames.addElement(
				CloneFactory.clone((String)fcs.eeNames.
						   elementAt(lpc)));
		}
		return this;
	}
	
	public synchronized Object clone()
	{
		return new DomainClassSpec().absorb(this);
	}

	public static DomainClassSpec merge(DomainClassSpec fcs[])
	{
		DomainClassSpec retval;
		int lpc;
		
		retval = new DomainClassSpec();
		for( lpc = 0; lpc < fcs.length; lpc++ )
		{
			synchronized( fcs[lpc] )
			{
				retval.absorb(fcs[lpc]);
			}
		}
		return retval;
	}

	public static void exportServerClasses(String eeName)
	{
		try
		{
			NameSpaceResource.exportClassGroup(eeName);
		}
		catch(Exception e) // have exportClassGroup() to throw specific, checked exceptions.
		{
			// XXX Handle this better
			Node.printStackTrace(e);
			Node.fatalError("Couldn't export server classes "
					+ eeName);
		}
	}
	
	PACKAGE static void setDefault(DomainClassSpec fcs)
	{
		/*
		 * XXX Need to get around lazy allocation to stop cross
		 * process pointers.
		 */
		fcs.systemPackages.elements();
		fcs.exportedClasses.elements();
		fcs.sealedPackages.elements();
		DomainClassSpec.defaultSpec = fcs;
	}
	
	public static DomainClassSpec getDefault()
	{
		return DomainClassSpec.defaultSpec;
	}
}
