/*
 * DomainInitializer.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.lang.reflect.Constructor;

import JVM_SYS_PKG.Team;

import JVM_RESOURCES_PKG.GCPolicy;

/**
 * Callback for initializing a new Domain (in the context of one of its
 * own threads).
 * <p>
 * This class's <code>run()</code> should be a method on a Domain
 * object, but starting a thread requires a <emph>public</emph>
 * <code>run()</code> method, which would be visible to the world on
 * Domain.
 * <p>
 * XXX probably should be combined with DomainStartSpec?
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class DomainInitializer
	implements Runnable
{
	/**
	 * The spec that determines how the domain should be started up.
	 */
	private final DomainStartSpec fss;
	
	/**
	 * Must be called in the context of the domain for which
	 * it is being used.
	 */
	PACKAGE DomainInitializer(DomainStartSpec fss)
	{
		/* Make a copy in this domain's memory */
		this.fss = (DomainStartSpec)fss.clone();
	}

	/**
	 * The run() function that the domain initializing thread will be
	 * executing.
	 */
	public void run()
	{
		Domain domain;

		TRACE(Trace.NODEOS_DOMAINSTART, "DomainInitializer.run()");

		domain = Domain.currentDomain();
		assert(domain != null);

		try
		{
			GCPolicy gp;

			/* Start up the GC threads within this Domains context */
			gp = GCPolicy.currentPolicy();
			if( (gp != null) && !gp.startThreads() )
			{
				Node.println("Unable to start up GC threads in new domain.");
				domain.destroy();
			}
			/* Start the startclass in the domain. */
			startInitialClass(domain, this.fss);
		}
		catch (ExceptionInInitializerError eiie)
		{
			Node.println("DomainInitialzier.run(): Trying to start "
				     + this.fss.getStartClass()
				     + " in new domain(" +domain+ ") and got:");
			Node.printStackTrace(eiie.getException());

			domain.destroy();
		}
		catch (Throwable e) // XXX
		{
			Node.println("DomainInitialzier.run(): Trying to start "
				     + this.fss.getStartClass()
				     + " in new domain(" +domain+ ") and got:");
			Node.printStackTrace(e);

			domain.destroy();
		}
	}
	
	/**
	 * Start the domain on its way.
	 */
	private static void startInitialClass(Domain domain,
					      DomainStartSpec fss)
		throws Exception
	{
		DomainRun domainRun;
		DomainClassLoader loader;

		require(domain);
		require(fss);
		require(fss.getStartClass());
		require(fss.getArguments());
		require(Team.current());
		require(Team.current().owns(fss.getStartClass()));
		
		TRACE(Trace.NODEOS_DOMAINSTART, "DomainInitializer.startInitialClass(" +fss.getStartClass()+ ")");

		Class c;
		
		c = domain.loadClass(fss.getStartClass(), true);
		domainRun = createStartClassInstance(c);
		
		TRACE(Trace.NODEOS_DOMAINSTART, "DomainInitializer.startInitialClass(): running startup class.");
		domainRun.run(domain,
			    fss.getArguments(),
			    fss.getCommSpaceHandle());
	}
	
	private static DomainRun createStartClassInstance(Class c)
		throws Error,  // XXX Error should be a Janos-specific Error
		NoSuchMethodException,
		IllegalAccessException,
		InstantiationException,
		java.lang.reflect.InvocationTargetException // XXX XXX XXX
	{
		Object instance;

		require(c);

		/*
		 * This seemed so easy:
		 *    domain.domainRun = (DomainRun) c.newInstance();
		 *
		 * But no, it's not, because we don't want to require
		 * c to have a public constructor.  See
		 * ClassInstantiationKey.
		 *
		 * Or, we need to have a VM method that will execute a
		 * private/package constructor on the class.  We trust
		 * the EE runtime to only pass safe classes.
		 */
		try
		{
			Class[] paramTypes;
			Object[] params;
			Constructor con;

			paramTypes = new Class[]{ ClassInstantiationKey.class };
			con = c.getConstructor(paramTypes);
							   
			params = new Object[] { new ClassInstantiationKey(c) };

			instance = con.newInstance(params);
		}
		catch(NoSuchMethodException e)
		{
			// Try the no-arg public constructor version(?)
			instance = c.newInstance();
		}

		if (! (instance instanceof DomainRun))
			throw new Error(c.getName() + " is not an instanceof DomainRun.");

		return (DomainRun)instance;
	}
}
