/*
 * DomainRegister.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX

import JVM_SYS_PKG.Team;
import JVM_SYS_PKG.TeamHandle;
import JVM_SYS_PKG.DeadTeamException;
import JVM_SYS_PKG.TeamEjectionException;

/**
 * Maintain a map of all the Domains in the system.  Currently
 * only used by the DomainInfo InfoHook.
 * <p>
 * <b>NOTE:</b> We don't use a hashtable to map domains
 * from name to domain object, becuase names are mutable and
 * are set after the domain is registered.
 * <p>
 * XXX currently this class is only useful for debugging builds.
 * <p>
 * XXX rename to DomainRegistry
 *
 * @see Domain#newDomain()
 * @see Domain#destroy()
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class DomainRegister
{
	PACKAGE static Team nodeosTeam = Team.current();

	/**
	 * This is a real global variable.  It is maintained
	 * in the kernel.
	 */
	private static final Vector domains = new Vector(30);

	/**
	 * InfoHook class that reports the current set of domains in the
	 * system via the NodeOS Status interface.
	 */
	private static class DomainInfo
		extends InfoHook
	{
		PACKAGE DomainInfo()
		{
			super("Domain");
		}

		public String[] getInfo()
		{
			String retval[];

			synchronized(domains)
			{
				retval = new String[domains.size()];
				
				for (int i = 0; i < domains.size(); i++)
				{
					DomainHandle elem =
						(DomainHandle)domains.elementAt(i);
					retval[i] = elem.toString();
				}
			}
			return retval;
		}
	}

	static {
		Status.addInfoHook(new DomainInfo());
	}

	/**
	 * Register the given domain.
	 *
	 * @allocation: a new DomainHandle to hold a reference to the domain
	 */
	PACKAGE static DomainHandle register(final Domain domain)
	{
		DomainHandle retval = null;
		
		// Switch into kernel context, create a DomainHandle pointing
		// to the given domain and register the handle in the table.

		try
		{
			nodeosTeam.switchTo();
			{
				retval = new DomainHandle(domain);
				synchronized(domains)
				{
					domains.addElement(retval);
				}
			}
			nodeosTeam.returnFrom();
		}
		catch (DeadTeamException dte)
		{
		}
		catch(TeamEjectionException dte)
		{
		}
		catch (Exception whatever)
		{
			FATALEXCEPTION(whatever, "Exception in DomainRegister.register()");
		}
		return retval;
	}

	/**
	 * Unregister a domain.
	 * <p>
	 * NOTE: this method is invoked in a context where most of the
	 * Domain has been destroyed.  Beware.
	 * <p>
	 * XXX this is bogus-slow.  Could cache index in domain object?
	 */
	PACKAGE static void unregister(final Domain domain)
	{
		TRACE(Trace.NODEOS_DESTROY, "Unregistering domain " +domain);

		try
		{
			nodeosTeam.switchTo();
			{
				synchronized(domains)
				{
					domains.removeElement(domain.kernelHandle);
				}
			}
			nodeosTeam.returnFrom();
		}
		catch (DeadTeamException dte)
		{
			/* XXX Bah, we'll leak the DomainHandle... */
		}
		catch(TeamEjectionException dte)
		{
		}
		catch (Exception whatever)
		{
			FATALEXCEPTION(whatever, "Exception in DomainRegister.unregister()");
		}
	}

#ifdef DEBUG
	/**
	 * Return the domain object corresponding to the given name
	 */
	public static DomainHandle findDomain(String name)
	{
		DomainHandle retval = null;
		int i;

		for( i = 0; (i < domains.size()) && (retval == null); i++ )
		{
			DomainHandle elem = (DomainHandle)domains.elementAt(i);
			if( elem.getName().equals(name) )
				retval = elem;
		}
		return retval;
	}
#endif
}
