/*
 * IPPortDelegate.Java
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.util.Enumeration;

import java.net.BindException;

import JVM_SYS_PKG.Team;
import JVM_SYS_PKG.TeamHandle;
import JVM_SYS_PKG.DeadTeamException;
import JVM_SYS_PKG.TeamEjectionException;

/**
 * The IPPortDelegate is a simple class used to manage which domains have
 * access to the TCP/UDP port space, and provides a handy way to manufacture
 * port numbers.
 */
public final class IPPortDelegate
{
	/**
	 * The range of port numbers that we'll manufacture.
	 */
	private final int PORT_RANGE_MIN = 2048;
	private final int PORT_RANGE_MAX = 65000;

	/**
	 * Identifies for the TCP port space.
	 */
	private static final int TCP_SPACE = 0;
	
	/**
	 * Identifies for the UDP port space.
	 */
	private static final int UDP_SPACE = 1;
	
	/**
	 * The port reservations across the whole node.
	 */
	private final static Hashtable globalPorts[] = {
		new Hashtable(),
		new Hashtable()
	};

	/**
	 * The port reservations for this object/domain.
	 */
	private final Hashtable localPorts = new Hashtable();

	/**
	 * Specifies which port space (TCP/UDP) that we're working in.
	 */
	private final int portSpace;
	
	/**
	 * A dummy class to distinguish the TCP port space in the team/domain.
	 */
	PACKAGE static class TCPSpace
	{
		PACKAGE final static Class thisClass = TCPSpace.class;
	}

	/**
	 * A dummy class to distinguish the UDP port space in the team/domain.
	 */
	PACKAGE static class UDPSpace
	{
		PACKAGE final static Class thisClass = UDPSpace.class;
	}

	/**
	 * @return The UDP port space delegate.
	 */
	public static IPPortDelegate getUDP()
	{
		Team team = Team.current();
		IPPortDelegate retval;

		synchronized(team)
		{
			if( (retval = (IPPortDelegate)
			     team.getClassLocal(UDPSpace.thisClass)) == null )
			{
				retval = new IPPortDelegate(UDP_SPACE);
				team.setClassLocal(UDPSpace.thisClass, retval);
			}
		}
		return retval;
	}
	
	/**
	 * @return The TCP port space delegate.
	 */
	public static IPPortDelegate getTCP()
	{
		Team team = Team.current();
		IPPortDelegate retval;

		synchronized(team)
		{
			if( (retval = (IPPortDelegate)
			     team.getClassLocal(TCPSpace.thisClass)) == null )
			{
				retval = new IPPortDelegate(TCP_SPACE);
				team.setClassLocal(TCPSpace.thisClass, retval);
			}
		}
		return retval;
	}

	/**
	 * Release any reserved ports.
	 */
	PACKAGE static void releasePorts()
	{
		Team team = Team.current();
		IPPortDelegate del;

		synchronized(team)
		{
			if( (del = (IPPortDelegate)
			     team.getClassLocal(TCPSpace.thisClass)) != null )
			{
				del.destroy();
			}
			if( (del = (IPPortDelegate)
			     team.getClassLocal(UDPSpace.thisClass)) != null )
			{
				del.destroy();
			}
		}
	}

	/**
	 * Construct an IPPortDelegate for the given space.
	 *
	 * @param space One of the above *_SPACE identifiers.
	 */
	private IPPortDelegate(int space)
	{
		this.portSpace = space;
	}

	/**
	 * @return A randomly selected and reserved ported number.
	 */
	public int reservePort()
	{
		int tries = 256, tmp, retval = -1;
		double rand = Math.random();

		tmp = PORT_RANGE_MIN +
			(int)(rand * (PORT_RANGE_MAX - PORT_RANGE_MIN));
		while( (tries > 0) && (retval == -1) )
		{
			try
			{
				this.reservePort(tmp);
				retval = tmp;
			}
			catch(BindException e)
			{
				tmp++;
			}
		}
		return retval;
	}

	/**
	 * @param port The port number to reserve.
	 * @exception Throws a BindException if the given port number is
	 * @exception already in use.
	 */
	public synchronized void reservePort(int port)
		throws BindException
	{
		Integer i = new Integer(port);
		
		if( this.localPorts.get(i) != null )
		{
			throw new BindException();
		}
		else
		{
			try
			{
				TeamHandle kernel = Team.current().
					getKernelTeam();
				
				kernel.switchTo();
				synchronized(globalPorts[this.portSpace])
				{
					if( globalPorts[this.portSpace].get(i)
					    == null )
					{
						Integer ki = new Integer(port);
							
						globalPorts[this.portSpace].
							put(ki, ki);
					}
					else
					{
						throw new BindException();
					}
				}
				kernel.returnFrom();
				this.localPorts.put(i, i);
			}
			catch(DeadTeamException e)
			{
				throw new BindException();
			}
			catch(TeamEjectionException e)
			{
				throw new BindException();
			}
		}
	}

	/**
	 * @param port The port number to release.
	 */
	public synchronized void unreservePort(int port)
	{
		Integer i = new Integer(port);
		
		if( this.localPorts.get(i) != null )
		{
			this.localPorts.remove(i);
			
			try
			{
				TeamHandle kernel = Team.current().
					getKernelTeam();
				
				kernel.switchTo();
				globalPorts[this.portSpace].remove(i);
				kernel.returnFrom();
			}
			catch(DeadTeamException e)
			{
				/* XXX Can this happen? */
			}
			catch(TeamEjectionException e)
			{
				/* XXX Can this happen? */
			}
		}
	}

	/**
	 * Release any reserved ports.
	 */
	private void destroy()
	{
		Enumeration keys = this.localPorts.keys();

		try
		{
			TeamHandle kernel = Team.current().getKernelTeam();
			
			kernel.switchTo();
			while( keys.hasMoreElements() )
			{
				Integer i = (Integer)keys.nextElement();
				
				globalPorts[this.portSpace].remove(i);
			}
			kernel.returnFrom();
		}
		catch(DeadTeamException e)
		{
			/* XXX Can this happen? */
		}
		catch(TeamEjectionException e)
		{
			/* XXX Can this happen? */
		}
	}
}
