/*
 * Main.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

IMPORT_PROFILER

/**
 * Boot the Janos Java NodeOS.  Initializes the underlying native
 * NodeOS, sets up the basic system packages in the ClassLoader, and
 * creates the root domain.
 * <p>
 * The Janos Java NodeOS must always startup with this class.  It doesn't
 * do much besides initialize the NodeOS and then fire up its first
 * argument in the root domain.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class Main
{
	/**
	 * All Janos active routers start with this function.  We load
	 * up native libraries, define the set of "system" packages,
	 * and fire up the root domain.  The arguments to main() should
	 * be a class to instantiate in the root domain and whatever
	 * arguments it wants.
	 *
	 * @see DomainClassLoader for info on system packages.
	 */
	public static void main(String[] args)
	{
		int lpc;
		
		NATIVE_SETUP.initialize();

		PROFILESTAGE("SysInit");

		Node.println("Janos Java NodeOS v1.2.0");
		
		require(args);

		String initClass = null;
		String[] initArgs = null;
		String initPackage;
		int pkgend;

		for( lpc = 0; lpc < args.length; lpc++ )
		{
			if( (initClass == null) && args[lpc].startsWith("--") )
			{
				if( args[lpc].equals("--debug") )
				{
					if( (lpc + 1) < args.length )
					{
						Trace.setFlags(args[lpc + 1]);
						lpc += 1;
					}
					else
					{
						Node.println("ERROR: No flags passed to `--debug'.");
						Node.shutdown();
					}
				}
				else
				{
					Node.println("ERROR: Unknown flag "
						     + args[lpc]
						     + " passed to Main");
					Node.shutdown();
				}
			}
			else if( initClass == null )
			{
				initClass = args[lpc];
				initArgs = new String[args.length - (lpc + 1)];
			}
			else
			{
				initArgs[lpc -
					(args.length - initArgs.length)] =
					args[lpc];
			}
		}
		if (initClass == null)
		{
			Node.println("ERROR: Must supply a class to load into root domain.");
			Node.shutdown();
		}
		
		// Find the package the initial class is in.
		pkgend = initClass.lastIndexOf('.');
		initPackage = null;
		if (pkgend != -1)
			initPackage = initClass.substring(0, pkgend);
		defineSystemPackages(initPackage);
		
		PROFILESTAGE("NodeOSInit");

		try
		{
			startRootDomain(initClass, initArgs);
		}
		catch (DomainStartException fse)
		{
			Node.println("Exception during root domain setup: ");
			Node.println("Cause: ");
			Node.printStackTrace(fse.getCause());
			Node.println("DomainStartException:");
			Node.printStackTrace(fse);
			Node.shutdown();
		}

		/* XXX don't do anything if startRootDomain returns... */
	}

	/**
	 * Start the native root domain.  Not really the "root domain" in
	 * Moab's case, as we start the root domain in md.c before Kaffe's
	 * main even gets going.
	 */
	private static void startRootDomain(String initClass,
					  String[] initArgs)
		throws DomainStartException
	{
		require(initClass);
		require(initArgs);

		Domain.createRootDomain(new Credentials("root cred"),
				    new ThreadPoolSpec(10, 32*1024,
						       ThreadScheduler.
						       basicScheduler()),
				    new MemSpec(),
				    new DomainStartSpec(initClass, initArgs, null));
	}

	/**
	 * Define the set of "system" packages that are loaded into
	 * the system domain.  This set is basically fixed, except for
	 * the implicit package passed on the command line.
	 */
	private static void defineSystemPackages(String initPackage)
	{
		DomainClassSpec fcs = new DomainClassSpec();
		int lpc;

		/* Set up the defaults */
		for( lpc = 0;
		     lpc < DomainClassSpec.defaultSystemPackages.length;
		     lpc++ )
		{
			fcs.addSystemPackage(DomainClassSpec.
					     defaultSystemPackages[lpc]);
		}
		for( lpc = 0;
		     lpc < DomainClassSpec.defaultJavaClasses.length;
		     lpc++ )
		{
			fcs.exportClass(DomainClassSpec.defaultJavaClasses[lpc]);
		}
		for( lpc = 0;
		     lpc < DomainClassSpec.defaultSealedPackages.length;
		     lpc++ )
		{
			fcs.sealPackage(DomainClassSpec.
					defaultSealedPackages[lpc]);
		}
		/* Add the initial package */
		if (initPackage != null)
			fcs.addSystemPackage(initPackage);
		DomainClassSpec.setDefault(fcs);
	}
}

