/*
 * TimeSpec.Java
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * A TimeSpec holds a time value as reported by the NodeOS.
 */
public final class TimeSpec
	extends NATIVE_TIMESPEC
{
	/**
	 * Get the current time.
	 *
	 * @param ts The TimeSpec object to fill out.
	 */
	public static void getTime(TimeSpec ts)
	{
		NATIVE_TIMESPEC.getTime(ts);
	}
	
	/**
	 * Handy unit constants.
	 */
	public static final int NANO = 1000000000;
	public static final int MICRO = 1000000;
	public static final int MILLI = 1000;

	/**
	 * The seconds value.
	 */
	private int seconds;

	/**
	 * The nanoseconds value.
	 */
	private int nanoseconds;

	/**
	 * @param seconds Number of elapsed seconds.
	 * @param nanoseconds Number of elapsed nanoseconds in addition to the
	 * @param nanoseconds seconds value.
	 */
	public TimeSpec(int seconds, int nanoseconds)
	{
		this.seconds = seconds;
		this.nanoseconds = nanoseconds;
	}

	/**
	 * @param ts The TimeSpec object to copy.
	 */
	public TimeSpec(TimeSpec ts)
	{
		this.copy(ts);
	}

	/**
	 * Construct a TimeSpec object with a time value of zero.
	 */
	public TimeSpec()
	{
	}

	/**
	 * @param ts The TimeSpec whose values should be copied into `this'.
	 * @return `this'
	 */
	public TimeSpec copy(TimeSpec ts)
	{
		this.seconds = ts.seconds;
		this.nanoseconds = ts.nanoseconds;
		return this;
	}

	/**
	 * @return The seconds value.
	 */
	public int getSeconds()
	{
		return this.seconds;
	}

	/**
	 * @param s The seconds value.
	 */
	public void setSeconds(int s)
	{
		this.seconds = s;
	}

	/**
	 * @return The nanoseconds value.
	 */
	public int getNanoSeconds()
	{
		return this.nanoseconds;
	}

	/**
	 * @param ns The nanoseconds value.
	 */
	public void setNanoSeconds(int ns)
	{
		this.nanoseconds = ns;
	}

	/**
	 * @return The total number of elapsed milliseconds.
	 */
	public double getTotalMillis()
	{
		double retval;

		retval = ((double)this.seconds) * (double)MILLI;
		retval += ((double)this.nanoseconds /
			   ((double)NANO / (double)MILLI));
		return retval;
	}

	/**
	 * @param ts The TimeSpec whose values should be added to `this'.
	 * @return `this'
	 */
	public TimeSpec add(TimeSpec ts)
	{
		this.seconds += ts.seconds;
		this.nanoseconds += ts.nanoseconds;
		if( this.nanoseconds >= NANO )
		{
			this.seconds++;
			this.nanoseconds -= NANO;
		}
		return this;
	}

	public void add(long ms)
	{
		this.seconds += ms / MILLI;
		this.nanoseconds += (ms % MILLI) * (NANO / MILLI);
	}

	/**
	 * @param ts The TimeSpec whose values should be subtracted from `this'
	 * @return `this'
	 */
	public TimeSpec subtract(TimeSpec ts)
	{
		this.seconds -= ts.seconds;
		this.nanoseconds -= ts.nanoseconds;
		if( this.nanoseconds < 0 )
		{
			this.seconds--;
			this.nanoseconds += NANO;
		}
		return this;
	}

	/**
	 * @param ts The TimeSpec whose values should be compared with `this'.
	 * @return True if `this' is greater than `ts'.
	 */
	public boolean greaterThan(TimeSpec ts)
	{
		boolean retval;

		if( this.seconds > ts.seconds )
			retval = true;
		else if( (this.seconds == ts.seconds) &&
			 (this.nanoseconds > ts.nanoseconds) )
			retval = true;
		else
			retval = false;
		return retval;
	}

	public boolean expired()
	{
		return NATIVE_TIMESPEC.expired(this);
	}
	
	/**
	 * @return The number of bytes needed to encode this TimeSpec.
	 */
	public int getLength()
	{
		return 8;
	}

	/**
	 * @param bh The buffer to marshal this TimeSpec into.
	 * @param offset The offset in the buffer to start marshalling.
	 */
	public int marshal(BufferHandle bh, int offset)
	{
		bh.setInt(offset, this.seconds);
		bh.setInt(offset + 4, this.nanoseconds);
		return 8;
	}

	/**
	 * @param bh The buffer that contains the marshalled TimeSpec.
	 * @param offset The offset in the buffer to start unmarshalling.
	 */
	public int unmarshal(BufferHandle bh, int offset)
	{
		this.seconds = bh.getInt(offset);
		this.nanoseconds = bh.getInt(offset + 4);
		return 8;
	}

	public boolean equals(Object obj)
	{
		boolean retval = false;

		if( obj instanceof TimeSpec )
		{
			TimeSpec ts = (TimeSpec)obj;

			if( (this.seconds == ts.seconds) &&
			    (this.nanoseconds == ts.nanoseconds) )
			{
				retval = true;
			}
		}
		return retval;
	}

	public int hashCode()
	{
		return (int)this.seconds;
	}

	public String toString()
	{
		return "[TimeSpec seconds="
			+ this.seconds
			+ "; nanoseconds="
			+ this.nanoseconds
			+ "]";
	}
}
