/*
 * moab_Domain.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector;

import JVM_SYS_PKG.Team;
import JVM_SYS_PKG.TeamHandle;
import JVM_SYS_PKG.Importable;
import JVM_SYS_PKG.Exportable;
import JVM_SYS_PKG.ExportManager;
import JVM_SYS_PKG.TeamReservation;
import JVM_SYS_PKG.PersistentExport;
import JVM_SYS_PKG.DeadTeamException;
import JVM_SYS_PKG.TeamEjectionException;

import JVM_RESOURCES_PKG.GCPolicy;
import JVM_RESOURCES_PKG.StdioSpec;
import JVM_RESOURCES_PKG.MemorySpec;
import JVM_RESOURCES_PKG.ConfigSpec;
import JVM_RESOURCES_PKG.NameSpaceSpec;
import JVM_RESOURCES_PKG.NameSpaceGroup;
import JVM_RESOURCES_PKG.ResourceSpec;
import JVM_RESOURCES_PKG.ResourceSpecException;

/**
 * Java wrapper for a NodeOS Domain.
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class moab_Domain
	extends Exportable
	implements PersistentExport
{
	CHECK_NATIVE_NODEOS

	/**
	 * A cross-process reference to the ExportedTermHandler in the
	 * parent domain.  The ExportedTermHandler points to the
	 * parent's termination handler function.
	 */
	private final ImportedTermHandler termHandler;

	private final moab_ThreadPool threadPool;
	private final Vector inChannels = new Vector();
	private final Vector outChannels = new Vector();

	private Object obj;

	private static int teamId = 0;

	/**
	 * The "root" domain termination handler.  This is invoked
	 * when the "root" domain is terminated, and causes the
	 * NodeOS to initiate shutdown.
	 * <p>
	 * The Java "root" domain is not the actual root domain.  (The
	 * real root domain is C-only and is created in janosvm/config/i386/nodeos/md.c.)
	 * <p>
	 * XXX NOT USED, as the term handler stuff is all broken right now.  See
	 * kickTermHandler()
	 */
	private static class RootDomainTermHandler
		implements java.lang.Runnable
	{
		public void run()
		{
			// Node.shutdown();
		}
	}

	/**
	 * A class used to keep the child domain termination
	 * callback object alive until the child dies.
	 * (Since potentially the only pointer to the callback
	 * would be in the child's domain.)
	 *
	 * @see ImportedTermHandler
	 */
	private static class ExportedTermHandler
		extends Exportable
	{
		private final Runnable termHandler;

		PACKAGE ExportedTermHandler(Runnable termHandler)
		{
			this.termHandler = termHandler;
		}

		PACKAGE Runnable getHandler()
		{
			return this.termHandler;
		}
	}

	/**
	 * Analogue of ExportedTermHandler, but for the child to hold on
	 * to the termHandler.  By definition, if the child is alive,
	 * the termHandler should be too, and when the child dies
	 * the child's importable will cease to exist, so the
	 * parent's exportable may go away, if no other references
	 * exist.
	 *
	 * @see ExportedTermHandler
	 */
	private static class ImportedTermHandler
		extends Importable
	{
		private ExportedTermHandler remTermHandler;

		public ImportedTermHandler()
		{
		}
		
		protected void bindTo(Exportable ex)
		{
			this.remTermHandler = (ExportedTermHandler)ex;
		}

		protected void unbindFrom()
		{
			this.remTermHandler = null;
		}
		
		public ExportedTermHandler getRemote()
		{
			return this.remTermHandler;
		}
	}

	
	/**
	 * Generate a unique name for a team.  Made by pasting
	 * the toString of the Credentials to a monotonically
	 * increasing number.
	 *
	 * @param cred the Credentials the new domain operates with
	 * @param fss the info for starting the new domain (unused currently)
	 * @return a string name for a new team
	 */
	private static String genTeamName(Credentials cred,
					  DomainStartSpec fss)
	{
		int id;

		synchronized(moab_Domain.class)
		{
			id = teamId++;
		}

		return cred.toString() + "-" + id;
	}

	/**
	 * Create a new domain in a new Team.
	 *
	 * @see Domain#newDomain()
	 */
	PACKAGE static DomainHandle createDomain(final Credentials cred,
					     final ThreadPoolSpec tPoolSpec,
					     final MemSpec memSpec,
					     final DomainStartSpec domainStartSpec,
					     final Runnable termHandler,
					     final DomainClassSpec fcs)
		throws DomainStartException
	{
		ResourceSpec[] resources;
		ExportManager exportedTermManager;
		Domain f;
		
		// Export the TermHandler
		ExportedTermHandler eth = new ExportedTermHandler(termHandler);
		exportedTermManager = Team.current().exportAnonymousObject(eth);
		
		try
		{
			resources = newDomainResources(memSpec, fcs);

			String teamName = genTeamName(cred, domainStartSpec);
			TeamHandle childTeamH = Team.create(teamName,
							   resources);
			DomainHandle childDomainH;
			TeamReservation ta;
			
			childDomainH = new DomainHandle(null);
			ta = childTeamH.pushReservation();
			// Do the actual visit, creating the new domain
			// and starting a thread
			ta.visit();
			{
				ImportedTermHandler ith;
				Team childTeam;
				Domain childDomain;

				ith = new ImportedTermHandler();

				childTeam = Team.current();
				childTeam.setDescription(domainStartSpec.
							 toString());
				
				childDomain = new Domain(cred,
					tPoolSpec,
					memSpec,
					ith,
					fcs);

				// Export the child domain from the childTeam,
				// and point the childDomainH at this child
				childTeam.exportObject("moab_Domain", childDomain).
					addImporter(childDomainH);

				// Have the child import the parent's termHandler
				exportedTermManager.addImporter(ith);

				// Start a new thread wholly in the child
				childDomain.threadStart(new DomainInitializer(domainStartSpec),
					"Initializer");
			}
			ta.leave();
			childTeamH.popReservation();
			return childDomainH;
		}
		catch (ResourceSpecException rse)
		{
			throw new DomainStartException("Resource spec failure",
						     rse);
		}
		catch (DeadTeamException dte)
		{
			throw new DomainStartException("Child team died too soon", dte);
		}
		catch (Throwable th)
		{
			throw new DomainStartException("Exception during bootstrap visit", th);
		}

		// NOT REACHED
	}

	/**
	 * Create root domain.  Different from createDomain because we
	 * use the current Team to host the root domain.
	 *
	 * @see Domain#newDomain()
	 */
	PACKAGE static DomainHandle createRootDomain(final Credentials cred,
						 final ThreadPoolSpec tPoolSpec,
						 final MemSpec memSpec,
						 final DomainStartSpec domainStartSpec)
		throws DomainStartException
	{
		ResourceSpec[] resources;
		Domain f;
		
		try
		{
			resources = newDomainResources(memSpec, null);

			String teamName = genTeamName(cred, domainStartSpec);
			TeamHandle childTeam = Team.create(teamName,
							   resources);
			DomainHandle retval;
			TeamReservation ta;
			Domain child;

			ExportedTermHandler eth;
			ExportManager exportedTermManager;

			eth = new ExportedTermHandler(new RootDomainTermHandler());
			exportedTermManager = Team.current().exportAnonymousObject(eth);

			retval = new DomainHandle(null);
			ta = childTeam.pushReservation();
			// Do the actual visit, creating the new domain
			// and starting a thread
			ta.visit();
			{
				// Create an import for the termHandler to keep it alive.
				ImportedTermHandler ith;
				Team team;

				ith = new ImportedTermHandler();

				team = Team.current();
				
				child = new Domain(cred,
					tPoolSpec,
					memSpec,
					ith,
					DomainClassSpec.getDefault());
				team.setDescription(domainStartSpec.toString());
				
				team.exportObject("moab_Domain", child).
					addImporter(retval);
				
				exportedTermManager.addImporter(ith);

				DEBUGONLY(child.setName("Root Domain"));

				child.threadStart(new DomainInitializer(domainStartSpec),
						  "Root Domain Initializer");
			}
			ta.leave();
			childTeam.popReservation();
			return retval;
		}
		catch (ResourceSpecException rse)
		{
			FATALEXCEPTION(rse, "Resource exception when creating root domain");
		}
		catch (DeadTeamException dte)
		{
			FATALEXCEPTION(dte, "Kernel team is dead.  I think not!");
		}
		catch (Throwable th)
		{
			FATALEXCEPTION(th, "Exception while creating team");
		}

		// NOT REACHED
	}


	private static ResourceSpec[] newDomainResources(MemSpec memSpec,
						       DomainClassSpec fcs)
		throws ResourceSpecException
	{
		StdioSpec stdioSpec;
		MemorySpec vmMemSpec;
		moab_DomainSpec domainSpec;
		ResourceSpec[] resources;
		NameSpaceGroup defaultNames[] =
			moab_Setup.getImportedNameSpaces();
		Vector servers = null;
		NameSpaceGroup nsg[];
		NameSpaceSpec nsSpec;
		int lpc;
		
		// XXX lameo spec, should be 'none' on all counts?
		// XXX make this static
		stdioSpec = new StdioSpec("builtin://stdin",
					  "builtin://stdout",
					  "builtin://stderr");

		vmMemSpec = new MemorySpec(memSpec.getSize(),
			MemoryPolicy.class.getName());

		// XXX this is cheap hack:
		domainSpec = new moab_DomainSpec();

		if( fcs != null )
			servers = fcs.getServerNames();
		nsg = new NameSpaceGroup[defaultNames.length +
					((servers != null) ?
					 servers.size() : 0)];
		for( lpc = 0; lpc < defaultNames.length; lpc++ )
		{
			nsg[lpc] = defaultNames[lpc];
		}
		if( servers != null )
		{
			for( lpc = 0;
			     lpc < servers.size();
			     lpc++ )
			{
				nsg[defaultNames.length + lpc] =
					new NameSpaceGroup(
						Team.current(),
						(String)servers.elementAt(lpc),
						"identity");
			}
		}
		nsSpec = new NameSpaceSpec(nsg);
		resources = new ResourceSpec[] {
			stdioSpec,
			nsSpec,
			vmMemSpec,
			domainSpec,
			// (ResourceSpec)ConfigSpec.current().clone(),
		};

		return resources;
	}

	/**
	 * Initalize a new Domain.  See newDomain().
	 * <p>
	 * Must be run in the context of the domain's team.
	 */
	PACKAGE moab_Domain(moab_Credentials cred,
			  moab_ThreadPoolSpec threadPoolSpec,
			  moab_MemSpec memSpec,
		          Object termHandler)
	{
		// XXX this is weak
		this.termHandler = (ImportedTermHandler)termHandler;
		
		// XXX all parameters are pointers into parent domain's heap.  Beware!
		this.threadPool = new moab_ThreadPool(threadPoolSpec);
		init(cred, this.threadPool, memSpec);
	}

	private native void init(moab_Credentials cred,
				 moab_ThreadPool threadPool,
				 moab_MemSpec memSpec);

	/**
	 * Called from native code (in moab_Domain.c) when a domain's
	 * termination handler is invoked by Moab.
	 * <p>
	 * Note that this will be invoked in the context of the parent
	 * domain.  This is where the real termination of the team
	 * and its Java-level pieces is performed.
	 *
	 * @see #destroy()
	 */
	PACKAGE void native_termHandler() 
	{
		/*
		 * XXX too late, the importables and exportables for
		 * this team have already been toasted, thus no link
		 * back to the parent's termHandler.
		 */
	}

	PACKAGE synchronized void addChannel(moab_InChannel ic)
	{
		this.inChannels.addElement(ic);
	}

	PACKAGE synchronized InChannel getInChannel(int index)
	{
		return (InChannel)this.inChannels.elementAt(index);
	}

	PACKAGE synchronized void remChannel(moab_InChannel ic)
	{
		this.inChannels.removeElement(ic);
	}

	PACKAGE synchronized void addChannel(moab_OutChannel oc)
	{
		this.outChannels.addElement(oc);
	}

	PACKAGE synchronized OutChannel getOutChannel(int index)
	{
		return (OutChannel)this.outChannels.elementAt(index);
	}

	PACKAGE synchronized void remChannel(moab_OutChannel oc)
	{
		this.outChannels.removeElement(oc);
	}

	PACKAGE native void getStats(moab_DomainStats fs);
	
	PACKAGE native void threadStart(java.lang.Runnable r, String name);

	/**
	 * Kick a term handler in the parent.  Wait for the term handler
	 * to comlete.
	 */
	PACKAGE void kickTermHandler()
	{
		ExportedTermHandler remoteHandler;
		
		/*
		 * XXX WE DONT KICK A THREAD IN THE PARENT.  JUST INVOKE DIRECTLY.
		 * WRONG SEMANTICS.
		 */

		require(this.termHandler != null);

		synchronized(this.termHandler)
		{
			remoteHandler = this.termHandler.getRemote();
			require(remoteHandler != null);
			try
			{
				remoteHandler.switchToOwner();
				// XXX this is the bit that should be executed
				// in a callback:
				remoteHandler.getHandler().run();
				remoteHandler.returnFromOwner();
			}
			catch (TeamEjectionException tee)
			{
				Node.println("Shouldn't be here TEE in kickTermHanlder()");
			}
			catch (DeadTeamException dte)
			{
				Node.println("Shouldn't be here DTE in kickTermHanlder()");
			}
		}
	}

	PACKAGE native void destroy0();

	/**
	 *
	 * @see #native_termHandler()
	 */
	PACKAGE void destroy()
	{
		// XXX destroy the inChannels or outChannels?
		// this.inChannels = null;
		// this.outChannels = null;

		// XXX destroy the threadPool (probably a bad idea)
		// this.threadPool.destroy();

		// destroy the Moab domain
		destroy0();

		Team.current().terminate();
		// XXX Remainder of the destruction is done in the native_termHandler()
	}

	PACKAGE Object getObject()
	{
		return this.obj;
	}

	PACKAGE void setObject(Object obj)
	{
		this.obj = obj;
	}

	public String toString()
	{
		return "obj=" +(obj != null ? obj.toString() : "(null)")+ ";"
			+ "inChan-count=" +inChannels.size()+ ";"
			+ "outChan-count=" +outChannels.size();
	}

	PACKAGE static native moab_Domain native_currentDomain();
}
