/*
 * pj_ANEPLayer.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.EOFException;

/**
 * pj_ANEPStreamLayer is a protocol layer implementation that processes packets
 * that are sent over a stream.  Its really meant to be used as a base class
 * which is extended for a specific communication type (TCP/StdIO/etc...).
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE abstract class pj_ANEPStreamLayer
	extends pj_ProtocolLayer
{
	private static final int MAX_PACKET_LENGTH = 16384; // XXX
	
	PACKAGE pj_ANEPStreamLayer()
	{
	}

	PACKAGE int getPacketLength()
	{
		return MAX_PACKET_LENGTH;
	}
	
	public String toString()
	{
		return "pj_ANEPStreamLayer[]";
	}

	/**
	 * StreamReader is an inner class used to read from an InputStream and
	 * produce PacketBuffer's.  The stream is expected to contain only
	 * ANEP packets since the length of the packet encoded in the ANEP
	 * header is used to delimit the beginning and end of packets.
	 */
	protected static abstract class StreamReader
		implements Runnable
	{
		/**
		 * The InputStream to read ANEP packets from.
		 */
		private final InputStream is;

		/**
		 * A reference count for tracking the number of in channels
		 * using this reader.
		 */
		private int users = 1;

		/**
		 * A buffer to read packets into.
		 */
		private final BufferHandle pb;

		/**
		 * Construct a StreamReader with the given values and set
		 * the reference count to one.
		 *
		 * @param is The InputStream to read ANEP packets from.
		 * @param pb A buffer to read packets into.
		 */
		protected StreamReader(InputStream is, BufferHandle pb)
		{
			this.is = is;
			this.pb = pb;
		}

		/**
		 * Increment the internal reference count.
		 */
		protected synchronized void acquire()
		{
			this.users++;
		}

		/**
		 * Decrement the internal reference count.
		 *
		 * @return True if there are no more users.
		 */
		protected synchronized boolean release()
		{
			this.users--;
			return this.users == 0;
		}

		/**
		 * Handle a packet parsed out of the stream
		 */
		protected abstract boolean handlePacket(BufferHandle pb);

		/**
		 * Helper function for always reading the specified amount
		 * of data.
		 *
		 * @param data The buffer to read into.
		 * @param offset The offset to start writing bytes into data.
		 * @param len The number of bytes to read.
		 * @exception Throws an IOException if there is an error while
		 * reading from the stream.
		 */
		private void read(byte data[], int offset, int len)
			throws IOException
		{
			int remaining = len, bytesRead;

			while ( remaining != 0 )
			{
				bytesRead = this.is.read(data,
							 offset,
							 remaining);
				switch ( bytesRead )
				{
				case -1:
					throw new EOFException();
				case 0:
					break;
				default:
					remaining -= bytesRead;
					offset += bytesRead;
					break;
				}
			}
		}

		/**
		 * Read an ANEP packet into the given buffer handle.
		 *
		 * @param bh The buffer to read the packet into.
		 * @exception Throws an IOException if there is an error while
		 * reading from the stream.
		 */
		private void readANEPPacketInto(BufferHandle bh)
			throws pj_InvalidANEPPacketException,
			       IOException
		{
			int anepLength;
			byte[] buf;
			
			require(bh);
			require(bh.getOffset() == 0);
			require(bh.getByteArray());
			require(bh.getByteArray().length == bh.getLength());

			buf = bh.getByteArray();
			/* Read the ANEP header */
			assert(bh.getLength() > 8);
			this.read(buf, 0, 8);

			anepLength = (((buf[6] & 0xFF) << 8)
				      | (buf[7] & 0xFF));

			/* Read the rest of the packet */
			if( bh.getLength() < anepLength )
			{
				throw new pj_InvalidANEPPacketException
					("ANEP packet too large.");
			}
			this.read(buf, 8, anepLength - 8);

			bh.setLength(anepLength);

			ensure(bh.getByteArray() == buf);
		}

		public void run()
		{
			require(this.pb);

			try
			{
				boolean done = false;
				
				/* Keep going while someone cares */
				do
				{
					this.readANEPPacketInto(this.pb);
					this.handlePacket(this.pb);
					synchronized( this )
					{
						done = (this.users == 0);
					}
				} while( !done );
			}
			catch (pj_InvalidANEPPacketException e)
			{
				Node.printStackTrace(e);
			}
			catch (EOFException e)
			{
			}
			catch (IOException e)
			{
				Node.printStackTrace(e);
			}

			try
			{
				this.is.close();
			}
			catch (IOException e)
			{
				Node.printStackTrace(e);
			}
		}

		public String toString()
		{
			return "StreamReader[is="
				+ this.is
				+ "; users="
				+ this.users
				+ "; pb="
				+ this.pb
				+ "]";
		}
	}

	/**
	 * StreamWriter is a helper class used to write ANEP packes to an
	 * OutputStream.
	 */
	protected static class StreamWriter
	{
		/**
		 * The stream to write packets to.
		 */
		private OutputStream os;

		/**
		 * A reference count for tracking the number of in channels
		 * using this writer.
		 */
		private int users = 1;

		/**
		 * Construct a StreamWriter with the given values and set the
		 * reference count to one.
		 *
		 * @param os The stream to write packets to.
		 */
		protected StreamWriter(OutputStream os)
		{
			this.os = os;
		}

		/**
		 * Increment the internal reference count.
		 */
		protected synchronized void acquire()
		{
			this.users++;
		}

		/**
		 * Decrement the internal reference count.
		 *
		 * @return True if there are no more users.
		 */
		protected synchronized boolean release()
		{
			this.users--;
			return this.users == 0;
		}

		/**
		 * @param pb The buffer to write to the stream.
		 * @exception Throws an IOException if there was a problem
		 * writing to the stream.
		 */
		protected synchronized void write(BufferHandle pb)
			throws IOException
		{
			this.os.write(pb.getByteArray(),
				      pb.getOffset(),
				      pb.getLength());
		}
		
		public String toString()
		{
			return "pj_ANEPStreamLayer$StreamWriter[os="
				+ this.os
				+ "; users="
				+ this.users
				+ "]";
		}
	}
}
