/*
 * pj_BufferHandle.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * pj_BufferHandle does not need to do array bounds checks or
 * to synchronize access.  (Done by the wrapper layer.)
 * <p>
 * A pj_BufferHandle can be on one of three queues: a pj_InChannel's
 * free buffer queue, a pj_OutChannel's outgoing queue, or a
 * pj_ThreadPool's dispatched BufferQ (waiting to be handled).
 * <p>
 * XXX I belive all of the synchronized statements in here are redundant
 * with protections necessairly provided by the caller.  I think.
 * <p>
 * XXX cannot synchronize on the pj_BufferHandle object, as it is externally
 * visible and is shared with the dispatcher.
 *
 * @see BufferHandle
 * @see pj_InChannel
 * @see pj_OutChannel
 * @see pj_ThreadPool
 */
PACKAGE class pj_BufferHandle
{
	private static final int MAX_PACKET_LEN = 16 * 1024; // XX XXX XX XX

	/**
	 * The actual data buffer.  Should be allocated in
	 * fixed size chunks.
	 */
	private byte[] data;

	/**
	 * Length of the valid region of the data.  
	 */
	private int length;

	/**
	 * Current offset into the buffer
	 */
	private int offset;

	// --- Break following into a separate (superclass) WorkItem?

	/**
	 * The deliver func associated with incoming packets that
	 * are queued up in a ThreadPool.
	 */
	private DeliverFunc deliverFunc;

	/**
	 * The payload start associated with incoming packets
	 * that are queud up in a ThreadPool.
	 *
	 * @see pj_InChannel#handle()
	 */
	private int payloadStart;

	/**
	 * Reference to the next buffer handle in a list.
	 */
	PACKAGE BufferHandle next;

	PACKAGE static int getMaxPacketLen()
	{
		return MAX_PACKET_LEN;
	}

	protected pj_BufferHandle(pj_BufferHandle other, int offset, int length)
	{
		require(other);
		require(offset >= 0);
		require(length >= 0);
		require(offset <= MAX_PACKET_LEN);
		require(length <= MAX_PACKET_LEN);
		require(offset+length <= MAX_PACKET_LEN);

		this.cloneHandle(other, offset, length);
		
		ensure(this.data != null);
		ensure(this.offset >= 0);
		ensure(this.length >= 0);
	}

	protected pj_BufferHandle(int offset, int length)
	{
		require(offset >= 0);
		require(length >= 0);
		require(offset <= MAX_PACKET_LEN);
		require(length <= MAX_PACKET_LEN);
		require(offset+length <= MAX_PACKET_LEN);

		this.data = new byte[MAX_PACKET_LEN];
		this.offset = offset;
		this.length = length;
		
		ensure(this.data != null);
		ensure(this.offset >= 0);
		ensure(this.length >= 0);
	}

	PACKAGE int getLength()
	{
		return this.length;
	}

	PACKAGE int getOffset()
	{
		return this.offset;
	}

	/**
	 * Set the state associated with incoming buffers.
	 *
	 * @param f the deliver func associated with the InChannel
	 * this BufferHandle was dispatched out of
	 * @param payloadStart an integer associated with the buffer.
	 * internally used to mark the first byte after the filter in
	 * this buffer, but may be anything (from InChannelHandle.push()).
	 */
	PACKAGE void setIncomingState(DeliverFunc f, int payloadStart)
	{
		require(f);
		// Make no restrictions on payloadStart: require(payloadStart >= 0);
		require(this.deliverFunc == null);

		this.deliverFunc = f;
		this.payloadStart = payloadStart;
	}

	// DEBUGONLY
	PACKAGE void clearIncomingState()
	{
		this.deliverFunc = null;
		this.payloadStart = 0;
	}

	PACKAGE DeliverFunc getDeliverFunc()
	{
		return this.deliverFunc;
	}

	PACKAGE int getPayloadStart()
	{
		return this.payloadStart;
	}

	/**
	 * For pj_* layer internal use only.
	 */
	PACKAGE void setLength(int len)
	{
		require(len >= 0);
		require(len <= this.data.length);

		synchronized(this)
		{
			this.length = len;
		}
	}
	

	PACKAGE byte[] getByteArray()
	{
		return this.data;
	}

	/**
	 * System reset of a BufferHandle.  This is
	 * invoked in pj_InChannel.handle().  A system
	 * reset, resets a buffer to a zero-offset and
	 * maximum length (for receipt of new packets).
	 *
	 * @see pj_InChannel#handle(pj_BufferHandle,int,int)
	 */
	PACKAGE void systemReset()
	{
		this.offset = 0;
		this.length = this.data.length;
	}

	/**
	 * Reset this BufferHandle, dropping the buffer
	 * it points to and setting offset and length to zero.
	 */
	PACKAGE void reset()
	{
		synchronized(this)
		{
			this.data = null;
			this.offset = 0;
			this.length = 0;
		}
	}

	PACKAGE void newBuffer(int offset, int length)
	{
		require(offset >= 0);
		require(length >= 0);
		require(length <= MAX_PACKET_LEN);
		
		synchronized(this)
		{
			// just drop this.data on the floor

			this.data = new byte[MAX_PACKET_LEN];
			this.offset = offset;
			this.length = length;
		}
	}

	PACKAGE void cloneHandle(pj_BufferHandle src, int fromOffset, int fromLength)
	{
		byte[] srcData;
		int srcOffset;
		int srcLength;

		/* BufferHandle did all of the bounds checks */

		require(fromOffset >= 0);
		require(fromLength >= 0);

		require(src.deliverFunc == null);
		require(src.payloadStart == 0);
		require(this.deliverFunc == null);
		require(this.payloadStart == 0);


		// Copy consistent information out of the src buffer
		synchronized(src) // synch with reset()
		{
			srcData = src.data;
			srcOffset = src.offset;
			srcLength = src.length;
		}

		assert(srcData.length == MAX_PACKET_LEN);
		assert((srcOffset + fromOffset) < srcData.length);
		assert((srcOffset + fromOffset + fromLength) < srcData.length);
		assert((fromOffset + fromLength) <= srcLength);

		synchronized(this) // synch with reset()
		{
			// Just drop this.data (gotta love gc!)  

			this.data = srcData;
			this.offset = srcOffset + fromOffset;
			this.length = fromLength;
		}
	}

	/**
	 * Swap the contents of this pj_BufferHandle with the given
	 * src pj_BufferHandle.  Only the data, offset and length
	 * fields are copied (the other must be irrelevant).
	 * Only the specified sub-sequence of the src buffer is
	 * "copied" into this buffer handle.
	 *
	 * @param src the buffer to swap with this
	 * @param fromOffset the offset (in addition to src.offset) to set in this
	 * @param fromLength the new length of this buffer
	 */
	PACKAGE void swapWith(pj_BufferHandle src, int fromOffset, int fromLength)
	{
		byte[] srcData, thisData;
		int srcOffset, thisOffset;
		int srcLength, thisLength;

		/*
		 * Technically, we probably don't need to synchronize
		 * these reads and writes out of the buffers, as
		 * swapWith() is only called in a context where
		 * buffers are owned entirely by the system, and
		 * only visible to a single thread.
		 */

		require(fromOffset >= 0);
		require(fromLength >= 0);

		require(src.deliverFunc == null);
		require(src.payloadStart == 0);
		require(this.deliverFunc == null);
		require(this.payloadStart == 0);

		// Copy consistent information out of the src buffer
		srcData = src.data;
		srcOffset = src.offset;
		srcLength = src.length;

		assert(srcData.length == MAX_PACKET_LEN);
		assert(srcOffset >= 0);
		assert((srcOffset + fromOffset + fromLength) <= srcLength);

		// Copy consistent info out of this buffer, and push
		// src data into this.

		thisData = this.data;
		thisOffset = this.offset;
		thisLength = this.length;
		
		this.data = srcData;
		this.offset = srcOffset + fromOffset;
		this.length = fromLength;

		assert(thisData.length == MAX_PACKET_LEN);
		assert(thisOffset >= 0);
		
		src.data = thisData;
		src.offset = thisOffset;
		src.length = thisLength;
	}

	PACKAGE byte getByte(int index)
	{
		require(index >= 0);
		require(index < length);
		return data[offset + index];
	}

	public void setByte(int index, byte val)
	{
		require(index >= 0);
		require(index < length);
		data[offset + index] = val;
	}

	PACKAGE short getShort(int index)
	{
		require(index >= 0);
		require(index + 2 <= length);
		index += offset;
		byte[] buf = data;
		return (short)(((buf[index++] & 0xFF) <<  8) |
			       ((buf[index]   & 0xFF)));
	}

	public void setShort(int index, short val)
	{
		require(index >= 0);
		require(index + 2 <= length);

		index += offset;
		byte[] buf = data;
		data[index++] = (byte) (val >> 8);
		data[index] = (byte) (val);
	}

	PACKAGE int getInt(int index)
	{
		require(index >= 0);
		require(index + 4 <= length);
		index += offset;
		byte[] buf = data;
		return (((buf[index++] & 0xFF) << 24) |
			((buf[index++] & 0xFF) << 16) |
			((buf[index++] & 0xFF) <<  8) |
			((buf[index]   & 0xFF)));
	}

	public void setInt(int index, int val)
	{
		require(index >= 0);
		require(index + 4 <= length);

		index += offset;
		byte[] buf = data;
		data[index++] = (byte) (val >> 24);
		data[index++] = (byte) (val >> 16);
		data[index++] = (byte) (val >> 8);
		data[index] = (byte) (val);
	}

	PACKAGE void getBytes(int srcPosition, byte[] dest,
			      int destPosition, int destLength)
	{
		require(srcPosition >= 0);
		require(srcPosition + destLength <= length);

		System.arraycopy(data, offset + srcPosition,
				 dest, destPosition, destLength);
	}

	PACKAGE void setBytes(int destPosition, byte[] src,
			      int srcPosition, int srcLength)
	{
		require(destPosition >= 0);
		require(destPosition + srcLength <= length);

		System.arraycopy(src, srcPosition,
				 data, offset + destPosition, srcLength);
	}

	PACKAGE void getInts(int srcPosition, int[] dest,
			     int destPosition, int destLength)
	{
		int intct;

		require(srcPosition >= 0);
		require(destPosition >= 0);
		require(srcPosition < this.length);
		require(destPosition < dest.length);
		require(destLength+destPosition <= dest.length);

		// blech
		for (intct = 0; intct < destLength; intct++)
		{
			int byteoffset;
			int val;
			
			byteoffset = this.offset + srcPosition + intct*4;
			// stored little-endian
			val = (this.data[byteoffset] & 0xff)
				| ((this.data[byteoffset+1] & 0xff) << 8)
				| ((this.data[byteoffset+2] & 0xff) << 16)
				| ((this.data[byteoffset+3] & 0xff) << 24);

			dest[destPosition + intct] = val;
		}
	}

	PACKAGE void setInts(int destPosition, int[] src,
			     int srcPosition, int srcLength)
	{
		int intct;

		require(srcPosition >= 0);
		require(destPosition >= 0);
		require(destPosition < this.length);
		require(srcPosition < src.length);
		require(srcPosition+srcLength <= src.length);
		require(destPosition+srcLength <= this.length);

		// blech
		for (intct = 0; intct < srcLength; intct++)
		{
			int byteoffset;
			int val;
			
			byteoffset = destPosition + intct*4;
			val = src[intct];

			// stored little-endian
			data[byteoffset] = (byte)(val & 0xff);
			val = val >> 8;
			data[byteoffset+1] = (byte)(val & 0xff);
			val = val >> 8;
			data[byteoffset+2] = (byte)(val & 0xff);
			val = val >> 8;
			data[byteoffset+3] = (byte)(val & 0xff);
		}
	}
	
	PACKAGE short computeChecksum(int start, int count)
	{
		int offset;
		int sum = 0;
		
		require(start >= 0);
		require(start < this.length);
		require(count >= 0);
		require(start+count <= this.length);

		offset = this.offset + start;
		while (count > 1)
		{
			short val;
			val = (short)((data[offset] |
				      (data[offset+1] << 8)) & 0xffff);
			sum  += val;
			offset += 2;
			count-= 2;
		}
		
		if (count > 0)
			sum += data[offset];

		while ((sum >> 16) != 0)
			sum = (sum & 0xffff) + (sum >> 16);

		return (short)(~sum);
	}
	

	PACKAGE boolean compareBytes(int offset, byte[] sequence,
				     int seqOffset, int seqLen)
	{
		require(offset >= 0);
		require(offset+seqLen < this.length);
		require(seqOffset >= 0);
		require(seqOffset+seqLen < sequence.length);
		
		offset = this.offset + offset;
		for (int i = 0; i < seqLen; i++)
		{
			if (this.data[offset+i] != sequence[seqOffset+i])
				return false;
		}

		return true;
	}

	/**
	 * Print this packet into the given buffer.  The "sbuf" is a
	 * a concession to the native implementation (it need not know
	 * about StringBuffers this way.)
	 */
	PACKAGE void dbgPacketSummary(char[] sbuf, int offset, int len)
	{
		int bufOffset = 0;

		for(int i = offset; i < this.length; i++)
		{
			if(i >= len)
			{
				// sbuf.append("...");
				sbuf[bufOffset++] = '.';
				sbuf[bufOffset++] = '.';
				sbuf[bufOffset++] = '.';
				break;
			}

			// I want two hex digits per byte.
			if((i % 16) == 0)
			{
				//sbuf.append("  ");
				sbuf[bufOffset++] = ' ';
				sbuf[bufOffset++] = ' ';
			}
			byte val = data[i];
			byte hi = (byte)((int)(val&0xFF) >> 4);

			//sbuf.append(Integer.toHexString(hi&0xF));
			sbuf[bufOffset++] = Integer.toHexString(hi&0xF).charAt(0);
			//sbuf.append(Integer.toHexString(val&0xF));
			sbuf[bufOffset++] = Integer.toHexString(val&0xF).charAt(0);
			//sbuf.append(" ");
			sbuf[bufOffset++] = ' ';
		}
	}
	
	public String toString()
	{
		return "pj_BufferHandle[data=<" 
			+ this.data
	                + ",len=" + this.data.length
			+ ">" 
			+ "; off=" + offset
			+ "; len=" + length
			+ "]";
	}
}
