/*
 * pj_Debug.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * The package-internal debugging infrastructure.
 * XXX Currently this is public for my convenience.
 * <p>
 * To avoid dependencies on classes that are not automagically
 * loaded into all spaces, I have to jump through several hoops to
 * make sure that non-critical classes are not used.  (In particular,
 * I have avoided the '"string" + String' syntax to avoid implicitly
 * referencing java/lang/StringBuffer.
 * <p>
 * The majority of routines in this file deal with printing directly
 * to stdout/stderr through VM facilities.  See the corresponding
 * native layer for the details on where that output is directed.
 * <p>
 * There are three basic facilities provided by this class, first
 * the <code>assert()</code> methods which really aren't very useful
 * without a pre-processor pass.
 * Second, the primitive print functions and object print functions,
 * because we all love printf-debugging.
 * Third, the flag-based trace primitive for dynamically adjusting
 * the printing of various parts of the Java system.
 * <p>
 * This class is not final.  However the only subclass Debug is final.
 * 
 * @see Debug
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class pj_Debug
{
	CHECK_NATIVE_NODEOS

#ifdef DEBUG
	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints a trailing newline.
	 *
	 * @param str The string to print
	 */
	public static void println(String str)
	{
		java.lang.System.err.println(str);
	}
	

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Does not print a trailing newline.
	 *
	 * @param str The string to print
	 */
	public static void print(String str)
	{
		java.lang.System.err.print(str);
	}

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints the given integer as a string (in base 10).
	 *
	 * @param i the integer to print
	 */
	public static void printInt(int i)
	{
		java.lang.System.err.print(i);
	}

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints the given integer as a string (in hex).
	 *
	 * @param i the integer to print
	 */
	public static void printIntHex(int i)
	{
		// XXX not implemented
		java.lang.System.err.print("0xXX");
	}

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints the given char array as a sequence of characters.
	 *
	 * @param array the char array to print
	 * @param start the first element of the char array to print
	 * @param end the last element of the char array to print
	 */
	public static void printCharArray(char[] array, int start, int end)
	{
		// XXX not implemented
		print("char[] XX");
	}

	/**
	 * A direct exit.  Do not pass Go; do not collect two hundred dollars.
	 *
	 * TODO: Rather dangerous to leave around.
	 */ 
	public static void exit()
	{
		try
		{
			java.lang.System.err.println("Debug.exit() invoked.");
			// XXX by " + Domain.currentDomain().getName());
			java.lang.System.exit(11);
		}
		catch (Throwable th)
		{
			java.lang.System.err.println("Error thrown in exit(): " +th.getMessage());
		}
		java.lang.System.exit(11);
	}
	
	/**
	 * Print the stack trace of the given Throwable.
	 */
	public static void printStackTrace(Throwable t)
	{
		beginDebugOutput();
		t.printStackTrace(java.lang.System.err);
		endDebugOutput();
	}

	/**
	 * Print the stack trace of the given Throwable
	 * Assumes we're already in an 'beginDebugOutput',
	 * 'endDebugOutput' block.
	 */
	public static void printStackTraceLocked(Throwable t)
	{
		t.printStackTrace(java.lang.System.err);
	}

	public static void beginDebugOutput()
	{
		// XXX
		java.lang.Thread.currentThread().setPriority(10);
	}
	
	public static void endDebugOutput()
	{
		// XXX
		java.lang.Thread.currentThread().setPriority(5);
	}

	public static boolean checkLocked(Object obj)
	{
		// XXX cannot check in Pure Java
		return true;
	}
	
#endif /*DEBUG*/
}
