/*
 * pj_DemultiplexKey.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector;

/**
 * A key for matching incoming packets.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_DemultiplexKey
{
	/**
	 * A segment of a demultiplex key.
	 */
	private static abstract class KeySegment
	{
		public KeySegment()
		{
		}

		/**
		 * Attempt to match this segment in the given payload.
		 */
		public abstract int match(byte payload[],
					  int payloadStart,
					  int payloadLength);
	}

	/**
	 * A segment that matches a series of bytes.
	 */
	protected static class ByteMask
		extends KeySegment
	{
		/**
		 * Number of bytes to match
		 */
		public final int length;
		/**
		 * The actual bytes to match
		 */
		public final byte sequence[];
		/**
		 * Bitmask of which bytes to match
		 */
		public final byte mask[];

		public ByteMask(int length,
				byte sequence[], byte mask[])
		{
			require((sequence == null)
				|| (length <= sequence.length));
			require((mask == null)
				|| (sequence.length <= mask.length));

			this.length = length;
			this.sequence = sequence;
			this.mask = mask;
		}
		
		public int match(byte payload[],
				 int payloadStart,
				 int payloadLength)
		{
			int retval;

			// Skipping `offset' bytes is done by the caller.

			/* Check for underrun */
			if( this.length <= (payloadLength - payloadStart) )
			{
				int lpc;

				/* Match the bytes */
				retval = payloadStart + this.length;

				if (this.sequence == null)
				{
				}
				else if (mask == null)
				{
					for (lpc = 0;
					     (lpc < this.length);
					     lpc++)
					{
						if (sequence[lpc] !=
						    payload[payloadStart + lpc])
						{
							return -1;
						}
					}
				}
				else
				{
					for (lpc = 0; lpc < this.length; lpc++)
					{
						if ((sequence[lpc]&mask[lpc]) !=
						    (payload[payloadStart + lpc]&mask[lpc]))
						{
							return -1;
						}
					}
				}
			}
			else
			{
				retval = -1;
			}
			return retval;
		}

		public String toString()
		{
			StringBuffer r = new StringBuffer(120);
			
			r.append("pj_DemultiplexKey.ByteMask[length=")
				.append(this.length)
				.append("; sequence='");
			
			for(int i = 0; i < length; i++)
			{
				if( sequence == null )
				{
					r.append("/XX");
				}
				else
				{
					int val = sequence[i];
					
					r.append(Integer.toHexString(val & 0xF0).charAt(0))
						.append(Integer.toHexString(val & 0x0F).charAt(0));
					
					if (mask != null)
					{
						// Append a "/XX" modifier for masked bits
						if ((mask[i]&0x0FF) != 0x0FF)
						{
							int mval = mask[i];
							
							r.append("/")
								.append(Integer.toHexString((mval>>8)&0xF).charAt(0))
								.append(Integer.toHexString(mval&0xF).charAt(0));
						}
					}
					
					if (i < (length-1))
						r.append(" ");
				}
			}
			r.append("'")
				.append("; mask=")
				.append(this.mask)
				.append("]");

			return r.toString();
		}
	}

	/**
	 * A key segment that matches an ANEP header.
	 *
	 * 0               *   1           *       2       *           3
	 * 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 * |   Version     |     Flags     |            Type ID            |
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 * |      ANEP Header Length       |     ANEP Packet Length        |
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 * ~                            Options                            ~
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 * |                               .                               |
	 * ~                            Payload                            ~
	 * |                               .                               |
	 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 */
	private static class ANEPHeader
		extends KeySegment
	{
		/**
		 * A key for matching the first part of the header
		 */
		private final ByteMask headerMask;

		/**
		 * Bitmask for matching the first part of the header, we don't
		 * want to match the flags field.
		 */
		private static final byte mask[] = {(byte)0xFF, 0x0, (byte)0xFF, (byte)0xFF};

		/**
		 * Similar to above, but we don't match the version either
		 */
		private static final byte allMask[] = {0x0, 0x0, (byte)0x0FF, (byte)0x0FF};

		/**
		 * Construct an ANEPHeader that matches a specific version and
		 * protocol ID.
		 */
		public ANEPHeader(byte version,
				  boolean allVersions, short protocolID)
		{
			byte[] sequence;
			sequence = new byte[4];
			sequence[0] = version;
			sequence[2] = (byte)((protocolID >> 8) & 0xFF);
			sequence[3] = (byte)((protocolID) & 0xFF);
			this.headerMask = new ByteMask(4, sequence,
						       allVersions ?
						       allMask : mask);
		}

		/**
		 * "Match" an ANEP header.  Matches the fixed part of the
		 * header (stored in this.headerMask), and then skips over
		 * the variable length part of the ANEP header.
		 */
		public int match(byte payload[],
				 int payloadStart,
				 int payloadLength)
		{
			int retval;

			// Skipping `offset' bytes is done by the caller.

			if( ((retval = this.headerMask.match(payload,
							     payloadStart,
							     payloadLength))
			     > 0) &&
			    (payloadLength >= (retval + 4)) )
			{
				int headerLen =
					(((payload[payloadStart + 4] & 0xFF) << 8)
					 | (payload[payloadStart + 5] & 0xFF))
					* 4;

				if( headerLen < 8 ) // minimum allowed value.
				{
					TRACE(Trace.NODEOS_DISPATCH, "ANEPHeader Malformed packet: headerlen="+headerLen+ " < 8.");
					retval = -1;
				}
				else if( headerLen > payloadLength )
				{
					TRACE(Trace.NODEOS_DISPATCH, "ANEPHeader Malformed packet: headerlen="+headerLen+ "; payloadLength=" +payloadLength+ ".");
					retval = -1;
				}
				else
				{
					retval = payloadStart + headerLen;
					TRACE(Trace.NODEOS_DISPATCH, "ANEP header match: offset=" + retval);
				}
			}
			else
			{
				TRACE(Trace.NODEOS_DISPATCH, "ANEP header mismatch.");
				retval = -1;
			}
			return retval;
		}

		public String toString()
		{
			return "pj_DemultiplexKey.ANEPHeader[headerMask="
				+ this.headerMask
				+ "]";
		}
	}

	/**
	 * The set of segments that make up this key.  The are matched based
	 * on their ordering.
	 */
	private final Vector segments = new Vector();
    
	private KeySegment keySegments[];

	/**
	 * The approximate length of data that this key will match.
	 */
	private int totalLength;

	PACKAGE pj_DemultiplexKey(pj_Domain domain)
	{
		require(domain);
		require(domain == pj_Domain.native_currentDomain());
	}

	/**
	 * Add a segment that matches the given sequence of bytes.
	 *
	 * @param length The length of the bytes to match
	 * @param sequence The sequence of bytes to match
	 * @param mask If non-null, its a bit mask that indicates which bytes
	 * in `sequence' should be matched.
	 */
	PACKAGE synchronized void add(int length, byte sequence[], byte mask[])
	{
		this.segments.addElement(new ByteMask(length,
						      sequence, mask));
		this.totalLength += length;
	}

	PACKAGE void addByte(byte b)
	{
		byte[] buf = new byte[1];
		buf[0] = b;
		add(1, buf, null);
	}

	PACKAGE void addShort(short s)
	{
		byte[] buf = new byte[2];
		buf[0] = (byte)(s & 0xff);
		s = (short)(s >> 8);
		buf[1] = (byte)(s & 0xff);
		add(2, buf, null);
	}

	PACKAGE void addInt(int i)
	{
		byte[] buf = new byte[4];
		buf[0] = (byte)(i & 0xff);
		i = i >> 8;
		buf[1] = (byte)(i & 0xff);
		i = i >> 8;
		buf[2] = (byte)(i & 0xff);
		i = i >> 8;
		buf[3] = (byte)(i & 0xff);
		add(4, buf, null);
	}

	PACKAGE void addLong(long l)
	{
		byte[] buf = new byte[8];
		buf[0] = (byte)(l & 0xff);
		l = l >> 8;
		buf[1] = (byte)(l & 0xff);
		l = l >> 8;
		buf[2] = (byte)(l & 0xff);
		l = l >> 8;
		buf[3] = (byte)(l & 0xff);
		l = l >> 8;
		buf[4] = (byte)(l & 0xff);
		l = l >> 8;
		buf[5] = (byte)(l & 0xff);
		l = l >> 8;
		buf[6] = (byte)(l & 0xff);
		l = l >> 8;
		buf[7] = (byte)(l & 0xff);
		add(8, buf, null);
	}

	/**
	 * Add a segment to match an Ethernet header
	 */
	PACKAGE void addETH(int flags,
			    byte src[], byte dst[], short type)
	{
		/* XXX ... */
		throw new InternalError("Ethernet header matching isn't supported");
	}

	/**
	 * Add a segment to match an IPv4 header
	 */
	PACKAGE void addIPv4(int flags,
			     int source, int srcMask,
			     int dest, int dstMask,
			     byte protocol)
	{
		/* XXX ... */
		throw new InternalError("IPv4 header matching isn't done yet.");
	}

	/**
	 * Add a segment to match a UDP header.
	 */
	PACKAGE void addUDP(int srcPort, int destPort)
	{
		/* XXX ... */
		throw new InternalError("UDP header matching isn't done yet.");
	}

	/**
	 * Add a segment to match a TCP header.
	 */
	PACKAGE void addTCP(int srcPort, int destPort)
	{
		/* XXX ... */
		throw new InternalError("TCP header matching isn't done yet.");
	}
	
	/**
	 * Add a segment that matches an ANEP header
	 *
	 * @param version The ANEP version number to match
	 * @param allVersions Match any version of ANEP
	 * @param protocolID The protocolID to match
	 */
	PACKAGE void addANEP(byte version,
			      boolean allVersions, short protocolID)
	{
		this.segments.addElement(new ANEPHeader(version,
							allVersions,
							protocolID));
		/* XXX `32' is a good guess at the size of the header */
		this.totalLength += 32;
	}

	/**
	 * Get the approximate length of the bytes that this key
	 * will match.  This is approximate.  The actual number of
	 * skipped bytes can only be determined for a specific packet.
	 */
	PACKAGE int length()
	{
		return this.totalLength;
	}

	PACKAGE synchronized void seal()
	{
		this.keySegments = new KeySegment[this.segments.size()];
		this.segments.copyInto(this.keySegments);
	}

	/**
	 * Test a packet to see if this key matches.
	 *
	 * @param payload The packet data
	 * @param payloadStart The place to start matching in `payload'
	 * @param payloadLength The length of data in payload, including bytes
	 * up to payloadStart.
	 *
	 * @return -1 if there was an error otherwise the index of the first
	 * byte after the end of the key.
	 */
	PACKAGE int matches(byte payload[],
			    int payloadStart,
			    int payloadLength)
	{
		int lpc, size, currPayload = payloadStart, retval = -1;
		KeySegment ks;

		if( this.keySegments == null )
			this.seal();
		for( lpc = 0;
		     (lpc < this.keySegments.length) && (currPayload >= 0);
		     lpc++ )
		{
			ks = (KeySegment)this.segments.elementAt(lpc);
			/* Try to match this segment to the current payload */
			currPayload = ks.match(payload,
					       currPayload,
					       payloadLength);
		}
		if( currPayload >= 0 )
			retval = currPayload;
		return retval;
	}

	public String toString()
	{
		StringBuffer retval = new StringBuffer(200);
		int lpc, size;

		retval.append("pj_DemultiplexKey[totalLength=")
			.append(Integer.toString(this.totalLength))
			.append("; segments={");

		size = this.segments.size();
		for( lpc = 0; lpc < size; lpc++ )
		{
			retval.append(this.segments.elementAt(lpc).toString());
			if (lpc < size-1)
				retval.append("+");
		}
		retval.append("}]");
		return retval.toString();
	}
}
