/*
 * pj_Domain.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX

/**
 * Java implementation of a NodeOS Domain.
 * A domain is the closest approximation to a process in Janos.
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_Domain
{
	/**
	 * The credentials provided when this pj_Domain was created.  This
	 * is the bucket from which the pj_Domain gets to extract resources
	 * and access Node facilities.
	 */
	private final pj_Credentials cred;

	/**
	 * The threadPool for this domain.
	 */
	private final pj_ThreadPool threadPool;

	/**
	 * The Memory limit for this domain.
	 */
	private final pj_MemSpec memSpec;

	/**
	 * Channels created in this pj_Domain.  This list only exists
	 * so that the channels can be cleanly destroyed when the
	 * domain is terminated.
	 */
	private final Vector channels = new Vector(5);

	/**
	 * A per-domain Object that user code can set and get.
	 * (for associating things with the domain).
	 */
	private Object obj;

	/**
	 * XXX: need to examine the global locking structure for
	 * deadlock, etc
	 * <p>
	 * XXX: Probably should move to locks on each element of the
	 * state and not a global state lock.
	 */
	PACKAGE final Object lock = new Object();

	private final Runnable termHandler;
	private pj_Domain parentDomain = null;
	
	/**
	 * Create a new domain "under" the given parent.
	 *
	 * @see #createDomain()
	 * @see #createRootDomain()
	 */
	private static DomainHandle createDomainUnder(Credentials cred,
						  ThreadPoolSpec tPoolSpec,
						  MemSpec memSpec,
						  DomainStartSpec domainStartSpec,
						  Runnable termHandler,
						  pj_Domain parentDomain,
						  DomainClassSpec fcs)
		throws DomainStartException
	{
		Domain f;

		require((parentDomain == null)
			|| (parentDomain == Domain.currentDomain().getNativeDomain()));

		try
		{
			/* Switch to heap of new domain */
			f = new Domain(cred,
				     tPoolSpec,
				     memSpec,
				     termHandler,
				     fcs);
			/* Switch back to heap of parent domain */
		}
		catch (Throwable th)
		{
			throw new DomainStartException(
				"Exception while creating new domain", th);
		}
		
		/* PJ-specific initialization */
		f.getNativeDomain().parentDomain = parentDomain; 

		/* XXX implicit thread start in a constructor is bad. */
		f.threadStart(new DomainInitializer(domainStartSpec),
			      "Initializer");
		
		return new DomainHandle(f);
	}

	private static final MemoryPolicy fakeMemPolicy = new MemoryPolicy(1000);

	PACKAGE MemoryPolicy getMemoryPolicy()
	{
		return fakeMemPolicy;
	}
	
	/**
	 * Odd method for this level (its just using the public interfaces).
	 * This is here because the moab_Domain.createDomain() is quite funky.
	 *
	 * @see Domain#newDomain()
	 */
	PACKAGE static DomainHandle createDomain(Credentials cred,
					     ThreadPoolSpec tPoolSpec,
					     MemSpec memSpec,
					     DomainStartSpec domainStartSpec,
					     Runnable termHandler,
		                             DomainClassSpec fcs)
		throws DomainStartException
	{
		return createDomainUnder(cred, tPoolSpec, memSpec,
				       domainStartSpec, termHandler,
				       Domain.currentDomain().getNativeDomain(),
			               fcs);
	}

	/**
	 * Creating root domain is no different from creating any other
	 * domain in pj version.  Much different in Moab version.
	 *
	 * @see Domain#newDomain()
	 */
	PACKAGE static DomainHandle createRootDomain(Credentials cred,
						 ThreadPoolSpec tPoolSpec,
						 MemSpec memSpec,
						 DomainStartSpec domainStartSpec)
		throws DomainStartException
	{
		DomainHandle fh;
		
		fh = createDomainUnder(cred, tPoolSpec, memSpec,
				     domainStartSpec,
				     /* Root domain has no term handler: */
				     null,
				     /* Root domain has no parent: */
				     null,
				     /* Use default DomainClassSpec: */
				     DomainClassSpec.getDefault());

		// XXX this is a hack:
		((pj_DomainHandle)fh).setName("Root Domain");

		return fh;
	}

	/**
	 * Initalize a new Domain.  See newDomain().
	 */
	PACKAGE pj_Domain(pj_Credentials cred,
			pj_ThreadPoolSpec threadPoolSpec,
			pj_MemSpec memSpec,
		        Object termHandler)
	{
		this.cred = cred;
		this.memSpec = memSpec;

		assert((termHandler == null) || (termHandler instanceof Runnable));
		this.termHandler = (Runnable)termHandler;

		this.threadPool = new pj_ThreadPool(this, threadPoolSpec);
	}

	/**
	 * Create an explicit thread.  This is a new thread which is
	 * not subject to whatever concurrency limit this Domain's
	 * ThreadPool has.  But, it will be scheduled by the
	 * ThreadPool's scheduler.
	 * 
	 * @allocations: This function will allocate a new Thread object.
	 * @param r the Runnable to start the target thread on.
	 * @param name the String name of the new thread.
	 */
	PACKAGE void threadStart(java.lang.Runnable r, String name)
	{
		this.threadPool.threadStart(r, name);
	}

	private boolean destroyed = false;

	PACKAGE final void kickTermHandler()
	{
		// Run the termination handler in the parent domain (if this isn't the root)
		if (this.parentDomain != null)
		{
			pj_Thread th;

			TRACE(Trace.NODEOS_DESTROY, "Running parent's term handler: " +this.termHandler);
			assert(this.termHandler);
			this.termHandler.run();
		}
		else
		{
			// then this is the the root domain
			assert(this.termHandler == null);

			Node.shutdown();
		}
	}

	PACKAGE synchronized void destroy()
	{
		// For debugging, make sure this isn't invoked more than once
		assert(!destroyed);
		this.destroyed = true;

		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		
		for (int i = 0; i < channels.size(); i++)
		{
			// Yes, this is stupid.  It is just temporary.
			// XXX make all the pj_channels extend
			// abstract class pj_AnyChannel

			Object c = channels.elementAt(i);
			if (c instanceof pj_InChannel)
			{
				((pj_InChannel)c).destroy();
			}
			else if (c instanceof pj_OutChannel)
			{
				((pj_OutChannel)c).destroy();
			}
			else if (c instanceof pj_CutThroughChannel)
			{
				((pj_CutThroughChannel)c).destroy();
			}
			else
			{
				panic("Non-channel object in channel list.");
			}
		}

		TRACE(Trace.NODEOS_DESTROY, "Destroying thread pool");
		threadPool.destroy();
	}
	
	PACKAGE void addChannel(pj_InChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE void addChannel(pj_OutChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE void addChannel(pj_CutThroughChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE void remChannel(pj_InChannel c)
	{
		channels.removeElement(c);
	}

	PACKAGE void remChannel(pj_OutChannel c)
	{
		channels.removeElement(c);
	}

	PACKAGE void remChannel(pj_CutThroughChannel c)
	{
		channels.removeElement(c);
	}

	PACKAGE Object getObject()
	{
		return obj;
	}

	PACKAGE void setObject(Object obj)
	{
		this.obj = obj;
	}

	PACKAGE pj_ThreadPool getThreadPool()
	{
		return threadPool;
	}

	PACKAGE void getStats(pj_DomainStats fs)
	{
		// XXX no stats in pj, yet
	}

	/**
	 * Return a fragment of an id string.  Domain.toString() will wrap this
	 */
	public String toString()
	{
		if (false) // longwinded
			return  "cred=" + cred
				+ "; obj=" + obj
				+ "; objid=0x" + Integer.toHexString(this.hashCode());
		else // short and sweet
			return "objid=0x" + Integer.toHexString(this.hashCode());
	}

	/**
	 * Get the current domain.  The current Domain is the
	 * domain associated with the current Thread.
	 *
	 * @see Thread#getDomain()
	 */
	PACKAGE static pj_Domain native_currentDomain()
	{
		return pj_Thread.self().getDomain();
	}
}

