/*
 * pj_InChannel.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

/**
 * Pure Java implementation of a NodeOS InChannel.
 *
 * @see InChannel
 * @see OutChannel
 * @see pj_OutChannel
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_InChannel
{
	/**
	 * The domain this receive channel is associated with.
	 *
	 * XXX what is this used for?
	 */
	PACKAGE final pj_Domain domain;

	/**
	 * The demultiplex key associated with this channel.
	 */
	PACKAGE final pj_DemultiplexKey classifier;

	/**
	 * Need the handler so the pj_InChannel can remove itself
	 * when it is destroyed.
	 * <p>
	 * XXX is this necessary?
	 */
	private pj_InChannelDispatcher handler;
	
	/**
	 * The callback for delivery.
	 */
	private final DeliverFunc deliverFunc;

	/**
	 * XXX Fill me in
	 */
	private BufferHandle freeBufferList;
	private final Object freeBufferLock = new Object();

	private final pj_ThreadPool threadPool;

	PACKAGE final ChanStats stats = new ChanStats();
	
	PACKAGE pj_InChannel(pj_Domain domain,
			     int threadCount,
			     pj_DemultiplexKey classifier,
			     String protocolSpec,
			     String addrSpec,
			     DeliverFunc deliverFunc)
		throws IOException,
		PacketAddressFormatException,
		ProtocolSpecFormatException
	{
		pj_ProtocolLayer pl;
		ThreadPoolSpec tps;
		ProtocolSpec ps;
		String layername;
		
		this.domain = domain;
		tps = new ThreadPoolSpec(threadCount, 32 * 1024, null);
		this.threadPool = new pj_ThreadPool(domain, tps);
		this.classifier = classifier;
		this.classifier.seal();
		this.deliverFunc = deliverFunc;

		/* Find the matching protocol layer */
		ps = new ProtocolSpec(protocolSpec);
		layername = ps.getComponent(ps.componentCount() - 1);
		pl = pj_ProtocolLayer.findLayer(layername);
		if( pl != null )
		{
			pj_ProtocolLayer.PacketAddressMask source;
			pj_InChannelDispatcher icd;

			/* Get the source address mask */
			source = pl.createAddressMask();
			source.parseAddressMask(addrSpec);
			icd = pj_InChannelDispatcher.
				getDispatcher(pl, source, this);
			this.handler = icd;
		}
		else
		{
			throw new ProtocolSpecFormatException("unknown protocolSpec " +protocolSpec);
		}
		
		domain.addChannel(this);
	}
	
	/**
	 * Dispatch an incoming packet which was determined to match
	 * this InChannel's packet filter.  The InChannel must have a
	 * buffer available to swap with the system buffer.  If there
	 * is a buffer available, it is swapped with the incoming buffer
	 * and queued up on the appropriate thread pool.
	 * <p>
	 * This is operating in "interrupt context", that is, this is
	 * all system time up to this point.
	 * 
	 * @param systembh the incoming buffer
	 * @param packetStart the "start" of the packet as seen by the receiver.
	 * @param payloadStart the start of the payload (i.e., skipping matched header bytes).
	 * @return returns true if the swap happened.
	 */
	PACKAGE boolean handle(BufferHandle systembh,
			       int packetStart,
			       int payloadStart)
	{
		boolean didSwitch;
		BufferHandle freebh;

		require(systembh);
		require(packetStart >= 0);
		// Make no restrictions on payload start

		/*
		 * Fetch a free buffer from the queue on this
		 * channel. Returns null if none are available.
		 */
		synchronized( this.freeBufferLock )
		{
			if( (freebh = this.freeBufferList) != null )
			{
				this.freeBufferList = freebh.next;
			}
		}
		
		didSwitch = false;
		if (freebh != null)
		{
			freebh.next = null;
			this.stats.inPackets += 1;
			this.stats.inBytes += systembh.getLength();
			
			// Swap the buffers freebh and systembh point to.
			// BufferHandles do *not* migrate around the system.
			freebh.swapWith(systembh, packetStart,
					systembh.getLength() - packetStart);
			freebh.setIncomingState(this.deliverFunc, payloadStart);
			this.threadPool.addWork(freebh);
			didSwitch = true;
		}
		else
		{
			this.stats.dropBytes += systembh.getLength();
			this.stats.dropPackets += 1;
			this.stats.dropNoBufs += 1;
		}

		// in any event, reset the (new or old) system buffer
		systembh.systemReset();

		return didSwitch;
	}

	/**
	 * Add a buffer to the set of buffers available for buffering
	 * incoming packets to this pj_InChannel.
	 * <p>
	 * This is run as a LIFO queue to maximize locality.
	 */
	PACKAGE void native_addBuffer(BufferHandle bufh)
	{
		synchronized( this.freeBufferLock )
		{
			bufh.next = this.freeBufferList;
			this.freeBufferList = bufh;
		}
	}

	/**
	 * Remove a buffer from the queue of free buffers
	 * on this pj_InChannel.  Returns null if there are
	 * no free buffers on this pj_InChannel.
	 * <p>
	 * This is run as a LIFO queue to maximize locality.
	 */
	PACKAGE BufferHandle native_remBuffer()
	{
		BufferHandle retval;
		
		synchronized( this.freeBufferLock )
		{
			if( (retval = this.freeBufferList) != null )
			{
				this.freeBufferList = retval.next;
			}
		}
		retval.next = null;
		return retval;
	}

	PACKAGE boolean handoff(BufferHandle bh)
	{
		boolean retval = false;
		int payloadStart;

		require(bh);

		payloadStart = this.classifier.matches(bh.getByteArray(),
						       0, bh.getLength());
		if (payloadStart != -1)
		{
			retval = this.handle(bh, 0, payloadStart);
		}
		return retval;
	}

	PACKAGE void getStats(pj_ChanStats cs)
	{
		this.stats.copyInto(cs);
	}
	
	public String toString()
	{
		return "pj_InChannel[domain="
			+ domain
			+ "; handler="
			+ handler
			+ "; classifier="
			+ classifier
			+ "; threadPool="
			+ threadPool
			+ "; deliverFunc="
			+ deliverFunc
			+ "]";
	}

	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);

		this.handler.remChannel(this);
		this.threadPool.destroy();
		this.domain.remChannel(this);
	}
}
