/*
 * pj_InChannelDispatcher.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.util.Vector;
import java.io.IOException;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Class that is used to dispatch packets from protocol layers to 
 * the appropriate InChannel.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class pj_InChannelDispatcher
	extends pj_ProtocolLayer.PacketHandler
{
	/**
	 * The initial size of the InChannel array.
	 */
	private static final int INITIAL_CHANNEL_SIZE = 4;

	/**
	 * The size to add/subtract when resizing the channels array.
	 */
	private static final int CHANNEL_DIFF = 4;
	
	/**
	 * The set of dispatchers for the system.
	 * <p>
	 * XXX global state?
	 */
	private static final Hashtable dispatchers = new Hashtable();

	/**
	 * Get the dispatcher that matches the given source mask.
	 *
	 * @param pl The protocol layer were packets are coming from
	 * @param source The source address mask packets are matched too
	 */
	PACKAGE static pj_InChannelDispatcher
	getDispatcher(pj_ProtocolLayer pl,
		      pj_ProtocolLayer.PacketAddressMask pam,
		      pj_InChannel ic)
		throws IOException
	{
		pj_InChannelDispatcher dispatcher;

		synchronized (pj_InChannelDispatcher.dispatchers)
		{
			/*
			 * Search our hash table first, and then mask one if
			 * needed
			 */
			dispatcher = (pj_InChannelDispatcher)
				pj_InChannelDispatcher.dispatchers.get(pam);

			if (dispatcher == null)
			{
				dispatcher = new pj_InChannelDispatcher(pl);
				dispatcher.mask = pam;
				pj_InChannelDispatcher.dispatchers.
					put(pam, dispatcher);
				pl.attachHandler(dispatcher);
			}
			dispatcher.addChannel(ic);
		}
		return dispatcher;
	}

	/**
	 * The protocol layer that this dispatcher is attached to
	 */
	private final pj_ProtocolLayer protocolLayer;
	
	/**
	 * The set of channels that this handler can dispatch too.
	 * We only really care about their demultiplex keys since
	 * the rest of the demultiplex is done elsewhere.
	 */
	private pj_InChannel channels[];

	/**
	 * The number of slots actually used in the channels array.
	 */
	private int channelsSize;

	/**
	 * Construct a pj_InChannelDispatcher with the given values.
	 *
	 * @param pl The protocol layer this dispatcher is being used with.
	 */
	private pj_InChannelDispatcher(pj_ProtocolLayer pl)
	{
		this.protocolLayer = pl;
		this.channels = new pj_InChannel[INITIAL_CHANNEL_SIZE];
	}

	/**
	 * Resize the channels array and possibly insert or remote an element.
	 *
	 * @param diff The difference between the new and old sizes.
	 * @param insert An element to insert, can be null.
	 * @param at The slot to insert or remove an element.
	 */
	private void resizeList(int diff, pj_InChannel insert, int at)
	{
		pj_InChannel ics[];
		int newSize;

		newSize = this.channelsSize + diff;
		if( (newSize > this.channels.length) ||
		    ((this.channels.length - newSize) >= CHANNEL_DIFF) )
		{
			int newLength;
			
			if( diff < 0 )
			{
				newLength =
					this.channels.length - CHANNEL_DIFF;
			}
			else
			{
				newLength =
					this.channels.length + CHANNEL_DIFF;
			}
			ics = new pj_InChannel[newLength];
		}
		else
		{
			ics = this.channels;
		}
		if( insert != null )
		{
			System.arraycopy(this.channels, 0,
					 ics, 0,
					 at);
			System.arraycopy(this.channels, at,
					 ics, at + 1,
					 this.channelsSize - at);
			ics[at] = insert;
		}
		else if( at >= 0 )
		{
			System.arraycopy(this.channels, 0,
					 ics, 0,
					 at);
			System.arraycopy(this.channels, at + 1,
					 ics, at,
					 this.channelsSize - at - 1);
		}
		else
		{
			System.arraycopy(this.channels, 0,
					 ics, 0,
					 newSize);
		}
		this.channels = ics;
		this.channelsSize = newSize;

		ensure(this.channelsSize >= 0);
	}

	/**
	 * Add a channel to this dispatcher.  The channel is placed
	 * in the selection line by how large its key is, the larger
	 * the key the earlier it will be checked.
	 */
	PACKAGE synchronized void addChannel(pj_InChannel newChan)
	{
		int lpc;

		for( lpc = 0; lpc < this.channelsSize; lpc++ )
		{
			pj_InChannel elem = this.channels[lpc];
			
			/* Insert in biggest to smallest order */
			if( newChan.classifier.length() >=
			    elem.classifier.length() )
			{
				break;
			}
		}
		TRACE(Trace.NODEOS_DISPATCH, "Adding packet classifier " + newChan.classifier);
		this.resizeList(1, newChan, lpc);
		this.notifyAll();
	}

	/**
	 * Remove a channel from this dispatcher.
	 */
	PACKAGE void remChannel(pj_InChannel deadChan)
	{
		synchronized( pj_InChannelDispatcher.dispatchers )
		{
			synchronized( this )
			{
				int lpc;
				
				for( lpc = 0; lpc < this.channelsSize; lpc++ )
				{
					if( this.channels[lpc] == deadChan )
						break;
				}
				if( lpc < this.channelsSize )
				{
					this.resizeList(-1, null, lpc);
					if( this.channelsSize <= 0 )
					{
						this.notifyAll();
						/*
						 * No more channels, detach
						 * ourselves from the protocol
						 * layer.
						 */
						this.protocolLayer.
							detachHandler(this);
						pj_InChannelDispatcher.
							dispatchers.
							remove(this.mask);
					}
				}
			}
		}
	}

	/**
	 * Handle a packet that came in through the transport layer.
	 * Returns true if the packet was handled.
	 * <p>
	 * The buffer is handled only if an InChannel can be found and
	 * that InChannel has buffers available to swap with.
	 * 
	 * @context Operating in "interrupt context".  (no real interrupts, mind you). 
	 * @param pb The system's receive pj_BufferHandle
	 * @return the input pb, perhaps with a different underlying buffer
	 */
	PACKAGE boolean handle(BufferHandle pb)
	{
		pj_InChannel matchChan = null;
		int payloadStart = 0;
		int lpc, size;

		/**
		 * Walk all the channels attached to this handler
		 * and find a match.
		 */
		synchronized( this )
		{
			for( lpc = 0; lpc < this.channelsSize; lpc++ )
			{
				pj_InChannel elem = this.channels[lpc];

				payloadStart = elem.classifier.
					matches(pb.getByteArray(), 0,
						pb.getLength());
				if( payloadStart != -1 )
				{
					/* Found a match */
					matchChan = elem;
					break;
				}
			}
		}
		if( matchChan != null )
		{
			TRACE(Trace.NODEOS_INCOMING, "demultiplexed to: " +matchChan.classifier);

			/*
			 * Ask the channel to take the matching buffer
			 * and return a replacement.  Channel will
			 * return false if it has nothing to swap for
			 * it.
			 */
			if (matchChan.handle(pb, 0, payloadStart))
				return true;

			/*
			 * The channel can't afford the packet.
			 */
			TRACE(Trace.NODEOS_INCOMING, "  Channel can't afford packet.");
		}
		else
		{
			TRACE(Trace.NODEOS_INCOMING, "No matching channel for pkt: data=" +Debug.dbgPacketSummary(pb));
			pb.systemReset();
		}

		return false;
	}

	public String toString()
	{
		return "pj_InChannelDispatcher["
			+ super.toString()
			+ "; protocolLayer="
			+ this.protocolLayer
			+ "; channels="
			+ this.channels
			+ "; channelsSize="
			+ this.channelsSize
			+ "]";
	}
}
