/*
 * pj_InetAddressMask.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.net.InetAddress;
import java.net.UnknownHostException;

import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 * Class that can be used to select InetAddress based on a mask with
 * wildcards.
 * <p>
 * PART OF THE PURE JAVA NODEOS IMPLEMENTATION.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class pj_InetAddressMask
{
	/**
	 * Address value
	 */
	private int address;

	/**
	 * If not null, then nameMask is a host name with or without
	 * wildcards.
	 * <p>
	 * XXX nameMask overrides the address masks?
	 */
	private String nameMask;

	/**
	 * Because the nameMask is only set in the constructor, the
	 * tokenization of it should be cached for comparison later
	 */
	private String nameTokens[];

	/**
	 * Bitmask telling us which parts of address are valid, if the bit is
	 * on then we match the corresponding 8 bits in address.  If the bit
	 * is off then anything matches.
	 */
	private int addressMask;

	/**
	 * If its an exact match then we just store the regular InetAddress
	 * here.
	 */
	private InetAddress reduced;

	/**
	 * @return An InetAddress if this mask is exact or NULL if the mask
	 * matches more than one address.
	 */
	PACKAGE InetAddress reduce()
	{
		return this.reduced;
	}

	public boolean equals(Object o)
	{
		boolean retval = false;
		
		if( o instanceof pj_InetAddressMask )
		{
			pj_InetAddressMask iam = (pj_InetAddressMask)o;

			if( (iam.nameMask == null)
			    && (this.nameMask == null) )
			{
				/*
				 * No name mask, check the address integer and
				 * bitmask
				 */
				if ((iam.address == this.address)
				    && (iam.addressMask == this.addressMask))
					retval = true;
			}
			else if( (iam.nameMask != null)
				 && (this.nameMask != null) )
			{
				/* Name mask, just compare the strings */
				retval = this.nameMask.equals(iam.nameMask);
			}
		}
		return retval;
	}

	/**
	 * Return the hashCode for the name string, if there is one,
	 * otherwise just use the address integer
	 */
	public int hashCode()
	{
		int retval;
		
		if( this.nameMask != null )
			retval = this.nameMask.hashCode();
		else
			retval = this.address;
		return retval;
	}

	/**
	 * Return a value that can be used when prioritizing
	 * pj_InetAddressMasks.
	 * <p>
	 * Used by the UDPLayer to distinguish overlapping packet
	 * handlers.
	 */
	PACKAGE int priority()
	{
		int retval;

		/*
		 * Simple algorithm for now, just return one if it isn't a mask
		 * otherwise, we return zero.
		 */
		if( this.addressMask == ~0 )
			retval = 1;
		else
			retval = 0;
		return retval;
	}
	
	public String toString()
	{
		String retval;

		/* Check if its a host mask or an IP number */
		if( nameMask != null )
		{
			retval = nameMask;
		}
		else
		{
			int tmpAddr, lpc;

			retval = "";
			tmpAddr = this.address;
			for( lpc = 0; lpc < 4; lpc++ )
			{
				if( lpc > 0 )
					retval += ".";
				/* Check if the bitmask is set. */
				if( (this.addressMask & (1L << (31 - lpc)))
				    != 0 )
					retval += (tmpAddr >> 24) & 0xFF;
				else
					retval += "*";
				tmpAddr <<= 8;
			}
		}
		return "pj_InetAddressMask["
			+ retval
			+ "]";
	}

	/**
	 * Construct an address mask from the given string
	 *
	 * @param addr An IP number or DNS name, possibly containing '*'s
	 * representing wildcards.
	 */
	PACKAGE pj_InetAddressMask(String addr)
		throws UnknownHostException, NumberFormatException
	{
		StringTokenizer st = new StringTokenizer(addr, ".");
		boolean byName = false;
		int slot = 31; // Start at the end of the mask

		this.address = 0;
		this.addressMask = ~0;
		try
		{
			String s;

			/* Walk name components/IP numbers */
			while( true )
			{
				s = st.nextToken();
				/*
				 * Push the address number to get ready for the
				 * next number
				 */
				this.address <<= 8;
				if( s.equals("*") )
				{
					/*
					 * Turn off the bit for this segment.
					 * XXX We don't handle host names
					 * with a lot of qualifiers
					 */
					this.addressMask &= ~(1L << slot);
				}
				else if( Character.isDigit(addr.charAt(0)) )
				{
					/*
					 * Its an IP digit, parse it.
					 */
					this.address |= Integer.parseInt(s);
				}
				else
				{
					/*
					 * We've hit some characters, must
					 * be a host name mask
					 */
					byName = true;
				}
				slot--;
			}
		}
		catch(NoSuchElementException e)
		{
		}
		if( byName || addr.equals("*") )
		{
			/* Its a host name mask, or we just match everything */
			if( this.addressMask == ~0 )
			{
				InetAddress ia = InetAddress.getByName(addr);
				byte addressBytes[];

				/*
				 * Its actually a fully specified host name,
				 * so we just resolve the name to the real
				 * address.
				 */
				addressBytes = ia.getAddress();
				this.address =
					((addressBytes[0] & 0xFF) << 24) |
					((addressBytes[1] & 0xFF) << 16) |
					((addressBytes[2] & 0xFF) <<  8) |
					((addressBytes[3] & 0xFF)      );
				this.reduced = ia;
			}
			else if( !addr.equals("*") )
			{
				int lpc;
				
				/* Save the mask string */
				this.nameMask = addr;
				st = new StringTokenizer(this.nameMask, ".");
				this.nameTokens = new String[st.countTokens()];
				for( lpc = 0;
				     lpc < this.nameTokens.length;
				     lpc++ )
				{
					this.nameTokens[lpc] =
						st.nextToken();
				}
			}
			else
			{
				this.addressMask = 0;
			}
		}
		else if( this.addressMask == ~0 )
		{
			this.reduced = InetAddress.getByName(addr);
		}
	}

	/**
	 * Test for a match between the given inet address and the mask.
	 *
	 * @param ia The InetAddress to test.
	 */
	PACKAGE boolean match(InetAddress ia)
	{
		boolean retval = true;
		
		if( this.nameMask != null )
		{
			StringTokenizer st =
				new StringTokenizer(ia.getHostName(), ".");
			String iaComp, thisComp = "";
			int lpc = 0;
			int tok = 0;

			/* Host string mask */
			try
			{
				while( retval )
				{
					iaComp = st.nextToken();
					thisComp = nameTokens[tok++];
					/*
					 * Test the bitmask for the segment
					 * XXX We don't handle matching several
					 * segments for a single `*'.
					 */
					if( (this.addressMask &
					     (1L << (31 - lpc)))
					    != 0 )
					{
						/* Check for equal names */
						if( !iaComp.equals(thisComp) )
						{
							retval = false;
						}
					}
					lpc++;
				}
			}
			catch(NoSuchElementException e)
			{
			}
			/*
			 * If the mask wasn't supposed to match everything
			 * and there are still more tokens to match, but the
			 * last segment wasn't a `*', then its not a good
			 * match.
			 */
			if( !thisComp.equals("*") &&
			    (((this.addressMask & (1L << (31 - lpc))) != 0) &&
			     st.hasMoreTokens()) ||
			    tok<nameTokens.length )
			{
				retval = false;
			}
		}
		else if( this.reduced != null )
		{
			retval = this.reduced.equals(ia);
		}
		else
		{
			byte addressBytes[];
			int tmpAddr, lpc;

			/* IP address mask */
			addressBytes = ia.getAddress();
			tmpAddr = this.address;
			for( lpc = 0; (lpc < 4) && retval; lpc++ )
			{
				/* Test the bitmask for the current segment */
				if( (this.addressMask &
				     (1L << (31 - lpc))) != 0 )
				{
					/*
					 * Bit is set, make sure values are the
					 * same
					 */
					if( (tmpAddr >> 24) !=
					    addressBytes[lpc] )
					{
						retval = false;
					}
				}
				tmpAddr <<= 8;
			}
		}
		return retval;
	}
}
