/*
 * pj_OutChannel.Java
 *
 * Copyright (c) 1999, 2000, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

import java.net.DatagramSocket;
import java.net.DatagramPacket;

IMPORT_DEBUG

/**
 * Java wrapper for an OutChannel.
 *
 * @see CutThroughChannel
 * @see InChannel
 * @see OutChannel
 */
PACKAGE class pj_OutChannel
	extends pj_ProtocolLayer.PacketGenerator
{
	/**
	 * The list of buffers to send asynchronously.
	 */
	private BufferHandle outBufferList;

	/**
	 * The lock for outBufferList.
	 */
	private final Object outBufferLock = new Object();
	
	PACKAGE final pj_Domain domain;
	PACKAGE pj_ProtocolLayer protocolLayer;

	PACKAGE final ChanStats stats = new ChanStats();

	PACKAGE pj_OutChannel(pj_Domain domain,
			      NetworkSpec ns,
			      String protocolSpec,
			      String addressSpec)
		throws IOException,
		       PacketAddressFormatException,
		       ProtocolSpecFormatException
	{
		pj_ProtocolLayer pl;
		ProtocolSpec ps;
		
		require(domain);
		require(protocolSpec);
		require(addressSpec);

		this.domain = domain;

		// XXX ns is ignored
		
		/* Find the matching protocol layer */
		ps = new ProtocolSpec(protocolSpec);
		if( (pl = pj_ProtocolLayer.
		     findLayer(ps.getComponent(ps.componentCount() - 1)))
		    != null )
		{
			AddressSpec as;

			/*
			 * Get an address object corresponding to the
			 * destination
			 */
			as = pl.createAddress(addressSpec);
			this.destination = as;
			pl.attachGenerator(this);
			this.protocolLayer = pl;
		}
		else
		{
			throw new IOException("protocolSpec="+protocolSpec+
					      " addressSpec="+addressSpec+
					      " ps="+ps);
		}

		domain.addChannel(this);
	}

	/**
	 * Packet generator function defined by ProtocolLayer.PacketGenerator.
	 */
	PACKAGE BufferHandle generate()
	{
		BufferHandle retval;
		
		synchronized( this.outBufferLock )
		{
			if( (retval = this.outBufferList) != null )
			{
				this.outBufferList = retval.next;
			}
		}
		return retval;
	}

	/**
	 * Queue a packet for sending whenever convenient.
	 *
	 * XXX needs callback function
	 */
	PACKAGE void queueAsyncPacket(BufferHandle pb)
	{
		require(pb);

		synchronized( this.outBufferLock )
		{
			pb.next = this.outBufferList;
			this.outBufferList = pb;
		}
		this.protocolLayer.generatorReady(this); 
	}

	/**
	 * Send a single buffer.  Synchronous (blocks until
	 * send is "done".
	 */
	PACKAGE boolean send(BufferHandle bh)
	{
		boolean retval = true;
		
		require(bh);

		TRACE(Trace.NODEOS_OUTGOING, "sending: destination=" +this+ " data=" +Debug.dbgPacketSummary(bh));

		try
		{
			this.protocolLayer.synchronousSend(this, bh);
			this.stats.outPackets += 1;
			this.stats.outBytes += bh.getLength();
		}
		catch(IOException e)
		{
			DEBUGONLY(Debug.println("IOException while trying to send in UDP layer: " +e.getMessage()));
			this.stats.dropPackets += 1;
			this.stats.dropBytes += bh.getLength();
			retval = false;
		}
		return retval;
	}

	/**
	 * Send a single buffer.  Synchronous.  Puts the
	 * buffer on the specified inChan after the send.
	 */
	PACKAGE void sendAndRecycle(BufferHandle bh, pj_InChannel inChan)
	{
		require(bh);
		require(inChan);

		TRACE(Trace.NODEOS_OUTGOING, "sending(w/recycle): destination=" +this+ " data=" +Debug.dbgPacketSummary(bh));
		send(bh);

		inChan.native_addBuffer(bh);
	}
	
	/**
	 * Queue a packet for sending on this OutChannel.  callback
	 * will be invoked when the call is complete.
	 * <p>
	 * XXX doesn't work yet.
	 * <p>
	 * XXX need to reconcile with synchronous sends.
	 */
	PACKAGE void sendAsynch(BufferHandle bh, DeliverFunc callback)
	{
		require(bh);

		TRACE(Trace.NODEOS_OUTGOING, "sending(async): destination=" +this+ " data=" +Debug.dbgPacketSummary(bh));

		queueAsyncPacket(bh);

		NOTIMPLEMENTED("sendAsynch()");
	}


	/**
	 * Send an array of buffer handles.
	 */
	PACKAGE void sendv(BufferHandle[] bhl)
	{
		require(bhl); // XXX
		require(bhl.length > 0); // XXX

		if (bhl.length == 1) 
		{
			this.send(bhl[0]);
		}
		else
		{
			NOTIMPLEMENTED("sendv() of a >1 length array of buffers");
		}
	}

	PACKAGE void getStats(pj_ChanStats cs)
	{
		this.stats.copyInto(cs);
	}

	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		this.protocolLayer.detachGenerator(this);
		this.domain.remChannel(this);
	}

	public String toString()
	{
		return "pj_OutChannel:"
			+ super.toString()
			+ "[domain="
			+ this.domain
			+ "; protocolLayer="
			+ this.protocolLayer
			+ "]";
	}
}
