/*
 * pj_PacketHandlerList.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector;
import java.net.DatagramPacket;
import java.net.Socket;

/**
 * pj_PortHandleList is a helper class for protocol layers that is
 * used to map a number of PortAddress-based PacketHandlers to an
 * individual socket.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class pj_PacketHandlerList
{
	/**
	 * The initial size of the mask array.
	 */
	private static final int INITIAL_MASKS_SIZE = 4;

	/**
	 * The size to add/subtract when resizing the mask array.
	 */
	private static final int MASKS_DIFF = 4;
	
	private pj_PortAddressMask masks[] =
		new pj_PortAddressMask[INITIAL_MASKS_SIZE];
	private int masksSize;

	PACKAGE pj_PacketHandlerList()
	{
	}
	
	/**
	 * Resize the channels array and possibly insert or remote an element.
	 *
	 * @param diff The difference between the new and old sizes.
	 * @param insert An element to insert, can be null.
	 * @param at The slot to insert or remove an element.
	 */
	private void resizeList(int diff, pj_PortAddressMask insert, int at)
	{
		pj_PortAddressMask pams[];
		int newSize;

		newSize = this.masksSize + diff;
		if( (newSize > this.masks.length) ||
		    ((this.masks.length - newSize) >= MASKS_DIFF) )
		{
			int newLength;
			
			if( diff < 0 )
			{
				newLength =
					this.masks.length - MASKS_DIFF;
			}
			else
			{
				newLength =
					this.masks.length + MASKS_DIFF;
			}
			pams = new pj_PortAddressMask[newLength];
		}
		else
		{
			pams = this.masks;
		}
		if( insert != null )
		{
			System.arraycopy(this.masks, 0,
					 pams, 0,
					 at);
			System.arraycopy(this.masks, at,
					 pams, at + 1,
					 this.masksSize - at);
			pams[at] = insert;
		}
		else if( at >= 0 )
		{
			System.arraycopy(this.masks, 0,
					 pams, 0,
					 at);
			System.arraycopy(this.masks, at + 1,
					 pams, at,
					 this.masksSize - at - 1);
		}
		else
		{
			System.arraycopy(this.masks, 0,
					 pams, 0,
					 newSize);
		}
		this.masks = pams;
		this.masksSize = newSize;

		ensure(this.masksSize >= 0);
	}

	/**
	 * Find a packet handler that matches the given socket, we
	 * just check the source masks of each handler.
	 */
	PACKAGE synchronized
		pj_ProtocolLayer.PacketHandler matchMask(Socket socket)
	{
		pj_PortAddressMask pam;
		int lpc;
		
		for ( lpc = 0; lpc < this.masksSize; lpc++ )
		{
			pam = this.masks[lpc];
			if( pam.matchSocket(socket) )
				return pam.ph;
		}
		return null;
	}

	/**
	 * Find a packet handler that matches the given packet, we
	 * just check the source masks of each handler.
	 */
	PACKAGE synchronized
		pj_ProtocolLayer.PacketHandler matchMask(DatagramPacket dp)
	{
		pj_PortAddressMask pam;
		int lpc;
		
		for ( lpc = 0; lpc < this.masksSize; lpc++ )
		{
			pam = this.masks[lpc];
			if( pam.matchPacket(dp) )
				return pam.ph;
		}
		return null;
	}

	/**
	 * Add a packet handler to this socket.
	 */
	PACKAGE synchronized void addHandler(pj_ProtocolLayer.PacketHandler ph)
	{
		pj_PortAddressMask pam;
		int lpc, size;

		pam = (pj_PortAddressMask)ph.mask;
		pam.ph = ph;
		for ( lpc = 0; lpc < this.masksSize; lpc++ )
		{
			pj_PortAddressMask elem = this.masks[lpc];
			
			if ( pam.priority() >= elem.priority() )
				break;
		}
		this.resizeList(1, pam, lpc);
	}
	
	/**
	 * Remove a packet handler from this socket.
	 */
	PACKAGE synchronized
		boolean remHandler(pj_ProtocolLayer.PacketHandler ph)
	{
		pj_PortAddressMask pam;
		int lpc;
		
		pam = (pj_PortAddressMask)ph.mask;
		for( lpc = 0; lpc < this.masksSize; lpc++ )
		{
			if( this.masks[lpc] == pam )
				break;
		}
		if( lpc < this.masksSize )
		{
			this.resizeList(-1, null, lpc);
			
		}
		return (this.masksSize == 0);
	}

	public String toString()
	{
		return "pj_PacketHandlerList[masks="
			+ this.masks
			+ "; masksSize="
			+ this.masksSize
			+ "]";
	}
}
