/*
 * pj_PortAddressMask.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.net.Socket;
import java.net.InetAddress;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 * Mask for PortAddresses.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_PortAddressMask
	extends pj_ProtocolLayer.PacketAddressMask
{
	PACKAGE pj_InetAddressMask srcMask;
	PACKAGE int srcPort = -1;
	PACKAGE int destPort = -1;
	PACKAGE pj_InetAddressMask destMask;
	
	/**
	 * Parses UDP address mask strings.  The format is:
	 *  "<src address>:<dest port>:<dest address>:<src port>
	 */
	PACKAGE void parseAddressMask(String addr)
		throws PacketAddressFormatException
	{
		StringTokenizer st = new StringTokenizer(addr, ":");
		String comp;
		
		try
		{
			/* Source address mask */
			this.srcMask = new pj_InetAddressMask(st.nextToken());

			/* Destination port */
			comp = st.nextToken();
			if( !comp.equals("*") )
			{
				this.destPort = Integer.parseInt(comp);
				if( (this.destPort < 0) ||
				    (this.destPort > 65536) )
				{
					throw new PacketAddressFormatException
						("Destination port out of "
						 + "range");
				}
			}

			/* Destination address */
			comp = st.nextToken();
			if( comp.equals("$source") )
			{
				/* XXX Doesn't really matter atm */
				this.destMask = new pj_InetAddressMask("*");
			}
			else
			{
				this.destMask = new pj_InetAddressMask(comp);
			}

			/* Source port */
			comp = st.nextToken();
			if( !comp.equals("*") )
			{
				this.srcPort = Integer.parseInt(comp);
				if( (this.srcPort < 0) ||
				    (this.srcPort > 65536) )
				{
					throw new PacketAddressFormatException
						("Source port out of range");
				}
			}

			ensure((this.srcPort == -1) || (this.srcPort >= 0));
			ensure(this.srcPort < 65536);
			ensure((this.destPort == -1) || (this.destPort >= 0));
			ensure(this.destPort < 65536);
		}
		catch(NumberFormatException e)
		{
			throw new PacketAddressFormatException
				("Invalid port number: " + e.getMessage());
		}
		catch(UnknownHostException e)
		{
			throw new PacketAddressFormatException
				("Bad host (" +destMask+ ") in spec: " +addr);
		}
		catch(NoSuchElementException e)
		{
		}
	}

	/**
	 * Check for a match between our mask and the datagram packet
	 */
	PACKAGE boolean matchPacket(DatagramPacket dp)
	{
		boolean retval = false;

		if ( this.srcMask.match(dp.getAddress()) &&
		     ((this.srcPort == -1) ||
		      (dp.getPort() == this.srcPort)) )
		{
			retval = true;
		}
		return retval;
	}

	PACKAGE boolean matchSocket(Socket sock)
	{
		boolean retval = false;

		if ( this.srcMask.match(sock.getInetAddress()) &&
		     ((this.srcPort == -1) ||
		      (sock.getPort() == this.srcPort)) )
		{
			retval = true;
		}
		return retval;
	}


	PACKAGE int priority()
	{
		int retval;

		retval = this.destMask.priority() +
			this.srcMask.priority();
		if( this.srcPort != -1 )
			retval++;
		if( this.destPort != -1 )
			retval++;
		return retval;
	}
	
	public int hashCode()
	{
		return ((this.srcMask != null) ?
			this.srcMask.hashCode() : 0) +
			this.srcPort +
			this.destPort +
			((this.destMask != null) ?
			 this.destMask.hashCode() : 0);
	}
	
	public boolean equals(Object o)
	{
		boolean retval = false;
		
		if( o instanceof pj_PortAddressMask )
		{
			pj_PortAddressMask uam = (pj_PortAddressMask)o;

			if( ((uam.srcMask != null) ?
			     uam.srcMask.equals(this.srcMask) :
			     (this.srcMask == null)) &&
			    ((uam.destMask != null) ?
			     uam.destMask.equals(this.destMask) :
			     (this.destMask == null)) &&
			    (uam.srcPort == this.srcPort) &&
			    (uam.destPort == this.destPort) )
			{
				retval = true;
			}
		}
		return( retval );
	}
	
	public String toString()
	{
		return "pj_PortAddressMask[srcAddrMask="
			+ this.srcMask
			+ "; srcPort="
			+ this.srcPort
			+ "; destPort="
			+ this.destPort
			+ "; destAddrMask="
			+ this.destMask
			+ "]";
	}
}

