/*
 * pj_Thread.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;

/**
 * A "Native" Thread.
 * <p>
 * A package-private wrapper for the "native" thread
 * layer.
 * <p>
 * This is a simple wrapper for java.lang.Thread.  Its
 * mostly a separate class so I can cleanly subclass it
 * without exposing all of the standard java.lang.Thread
 * methods.
 *
 * @see pj_ThreadPool
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_Thread
{
	/**
	 * Per-thread state.
	 * <p>
	 * This is part of the NodeOS spec.  Probably doesn't need to
	 * be part of the Java-level interface (users of this should
	 * just subclass Janos Thread).
	 */
	private Object obj;
	
	/**
	 * The ThreadPool this thread is a part of.
	 */
	private final pj_ThreadPool threadPool;

	private final ThreadWrapper jlThread;

	private final Hashtable threadLocals = new Hashtable();

	/**
	 * Hide java.lang.Thread in a member class so that subclasses
	 * of pj_Thread don't inherit the public java.lang.Thread
	 * methods.
	 */
	private class ThreadWrapper
		extends java.lang.Thread
	{
		/**
		 * The Runnable to use when thread is started.
		 */
		private final java.lang.Runnable r;

		PACKAGE ThreadWrapper(Runnable r, String name)
		{
			super(name);

			require(r);
			this.r = r;
			this.setDaemon(false);
		}

		public void run()
		{
			try
			{
				assert(r != null);
				r.run();
			}
			catch(Throwable t)
			{
				System.out.println("Exception in thread " +this+ " (runnable = " +r+ ")");
				System.out.println("Thread blew up with:" +t.getMessage());
				t.printStackTrace();
			}
		}

		/**
		 * Return container class.  Used to convert from
		 * java.lang.Thread.currentThread() to a pj_Thread().
		 */
		PACKAGE pj_Thread pj_Thread()
		{
			return pj_Thread.this;
		}
	}

	/**
	 * InfoHook class that reports on all of the native threads
	 * in the system.
	 */
	private static class pj_ThreadInfo
		extends InfoHook
	{
		PACKAGE pj_ThreadInfo()
		{
			super("threads");
		}
		
		public String[] getInfo()
		{
			String[] retval;
			java.lang.Thread[] allThreads;
			ThreadGroup r, p;

			// find root ThreadGroup
			r = pj_Thread.currentThread().jlThread.getThreadGroup();
			p = r;
			while (p != null)
			{
				r = p;
				p = r.getParent();
			}
			assert(r != null);

			allThreads = new java.lang.Thread[r.activeCount() + 50];
			int ct = r.enumerate(allThreads, true);
			
			retval = new String[ct];
			for (int i = 0; i < ct; i++)
			{
				retval[i] = allThreads[i].toString()
					+ " name="
					+ allThreads[i].getName();
			}

			return retval;
		}
	}

	static
	{
		Status.addInfoHook(new pj_ThreadInfo());
	}

	/**
	 * Create a new native thread.  This constructor is
	 * used to only create "explicit" threads.
	 * <p>
	 * The given Runnable will be used when the
	 * thread is started (instead of the run method on this).
	 *
	 * @allocations: A new ThreadWrapper
	 */
	PACKAGE pj_Thread(pj_ThreadPool threadPool, Runnable runnable, String name)
	{
		require(threadPool);
		require(runnable);
		
		this.jlThread = new ThreadWrapper(runnable, name);
		this.threadPool = threadPool;

		DEBUGONLY(this.jlThread.setPriority(5););

		TRACE(Trace.NODEOS_THREADS, " Created new thread: " +this);
	}

	PACKAGE static void yieldTo(pj_Thread th)
	{
		java.lang.Thread.yield();
	}

	PACKAGE static void sleep(long millis, int nanos)
		throws InterruptedException
	{
		java.lang.Thread.sleep(millis, nanos);
	}

	PACKAGE static Object getObject()
	{
		return self().obj;
	}

	PACKAGE static Object setObject(Object obj)
	{
		pj_Thread self = self();
		Object old;

		// XXX publically visible lock, but only hoses user
		synchronized(self) 
		{
			old = self.obj;
			self.obj = obj;
		}
		return old;
	}

	PACKAGE static Hashtable getThreadLocals()
	{
		pj_Thread self = self();

		return self.threadLocals;
	}


	/**
	 * Package-private run method that all *implicit* threads
	 * start in.
	 */
	PACKAGE void start()
	{
		this.jlThread.start();
	}

	PACKAGE pj_Domain getDomain()
	{
		return this.threadPool.native_getDomain();
	}

	/**
	 * Return the Thread object associated with the currently
	 * executing thread of control.
	 *
	 * @see currentThread()
	 * @return a Thread object associated with the current thread.
	 */
	PACKAGE static pj_Thread self() 
	{
		return currentThread();
	}

	/**
	 *
	 * @see #self()
	 */
	PACKAGE static pj_Thread currentThread()
	{
		java.lang.Thread jth;
		jth = java.lang.Thread.currentThread();
#ifdef DEBUG
		if (!(jth instanceof ThreadWrapper))
		{
			Node.printStackTrace(null);
			Node.fatalError("ERROR: current thread is not a pj_Thread.ThreadWrapper!");
		}
#endif
		return ((ThreadWrapper)jth).pj_Thread();
	}

	public void setName(String name)
	{
		this.jlThread.setName(name);
	}

	public String getName()
	{
		return this.jlThread.getName();
	}

	public String toString()
	{
		String name = this.getName();
		return "pj_Thread[n=" +name+ "]";
	}
}
