/*
 * pj_ThreadPool.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * NodeOS ThreadPool implementation in Java.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class pj_ThreadPool
	implements Runnable
{
	private static final int DEFAULT_WORK_SIZE = 16;
	
	/**
	 * The domain this ThreadPool is associated with.
	 */
	private final pj_Domain domain;

	private BufferHandle queue;
	private BufferHandle last;
	
	private boolean destroyed;

	private final String unassignedName = "Unassigned worker";

	/**
	 * Create a new pj_ThreadPool.
	 * <p>
	 * The variousSchedulerParameters and scheduler parameters
	 * hardly even qualify as hacks.
	 */
	PACKAGE pj_ThreadPool(pj_Domain domain, pj_ThreadPoolSpec tpspec)
	{
		require(domain);
		require(tpspec);

		this.destroyed = false;
		this.domain = domain;

		for (int i = 0; i < tpspec.maxThreads(); i++)
		{
			pj_Thread th;
			th = new pj_Thread(this, this, unassignedName);
			th.start();
		}
	}

	/**
	 * Create an explicit thread and return that thread handle.
	 * This is a new thread which is not subject to whatever
	 * concurrency limit the pj_ThreadPool has.  But, it will be
	 * scheduled by the pj_ThreadPool's scheduler.
	 * 
	 * @allocations: This function will allocate a new Thread object.
	 * @param r the Runnable to start the target thread on.
	 */
	PACKAGE void threadStart(java.lang.Runnable runnable, String name)
	{
		require(runnable);

		if (destroyed())  // XXX throw an exception?
			return;

		pj_Thread th = new pj_Thread(this, runnable, name);
		th.start();

		return;
	}

	/**
	 * Return the pj_Domain this ThreadPool is associated with.
	 */
	PACKAGE pj_Domain native_getDomain()
	{
		return this.domain;
	}

	/**
	 * Destroy this pj_ThreadPool.
	 * <p>
	 * XXX this is a hack.
	 */
	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		synchronized( this )
		{
			destroyed = true;
			this.notifyAll();
		}
		TRACE(Trace.NODEOS_DESTROY, "  Marked as destroyed (" +this+ ")");
	}

	PACKAGE boolean destroyed()
	{
		return destroyed;
	}

	/**
	 * Add a Buffer to the pj_ThreadPool's Buffer list.
	 * <p>
	 * The buffer must not be on some other list (either
	 * an InChannel list, OutChannel list, or some bufferQueue).
	 */
	PACKAGE synchronized void addWork(BufferHandle bh)
	{
		require(bh);
		require(bh.getDeliverFunc());
		require(bh.next == null);
		
		if( this.last == null )
			this.queue = bh;
		else
			this.last.next = bh;
		this.last = bh;
		this.notify();
	}

	PACKAGE synchronized BufferHandle dequeueWorkItem()
	{
		BufferHandle retval;
		
		while( !this.destroyed && (this.queue == null) )
		{
			try
			{
				this.wait();
			}
			catch(InterruptedException e)
			{
			}
		}
		retval = this.queue;
		this.queue = this.queue.next;
		retval.next = null;
		if( this.queue == null )
			this.last = null;
		return retval;
	}
	
	/**
	 * Do the work on this pj_ThreadPool.
	 * <p>
	 * This function won't return until the ThreadPool is destroyed.
	 */
	public void run()
	{
		try
		{
			this.realRun();
		}
		catch(NullPointerException e)
		{
			if( !this.destroyed )
			{
				Node.printStackTrace(e);
				Node.fatalError("ThreadPool thread died");
			}
		}
		catch(Throwable th)
		{
			Node.printStackTrace(th);
			Node.fatalError("ThreadPool thread died");
		}
	}

	private void realRun()
	{
		BufferHandle bh;

		TRACE(Trace.NODEOS_THREADS, " Worker thread started work loop.");

		while( !this.destroyed )
		{
			bh = this.dequeueWorkItem();

			// Do the item outside the context of the bufferQLock!
			if (bh != null)
			{
				DeliverFunc receiver;
				int payloadStart;

				TRACE(Trace.NODEOS_THREADS, " Worker thread started on " +bh);
				DEBUGONLY(pj_Thread.self().setName("Working on " +bh));

				receiver = bh.getDeliverFunc();
				assert(receiver);
				payloadStart = bh.getPayloadStart();

				DEBUGONLY(bh.clearIncomingState());

				try
				{
					bh.userOwns();
					receiver.receive(bh, payloadStart);
				}
				catch(Throwable th)
				{
					Node.println("Uncaught exception in deliverFunc (" +receiver+ "):");
					Node.println(th.getMessage());
					Node.printStackTrace(th);
				}

				DEBUGONLY(pj_Thread.self().setName(unassignedName));

				// Cannot do anything with the
				// pj_BufferHandle at this point, it
				// might already be scheduled for
				// re-use.
				bh = null;
			}
		}
		
		TRACE(Trace.NODEOS_DESTROY, " Worker Thread " +pj_Thread.self()+ " destroyed.");
	}
}
