/*
 * pj_UDPLayer.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

import java.util.Hashtable;

import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.SocketException;

/**
 * UDPLayer is a protocol layer implementation that runs over UDP.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_UDPLayer
	extends pj_ProtocolLayer
{
	/**
	 * Datgram wrapper for sending packets.  The stupid empty
	 * buffer is replaced with the pj_BufferHandle's buffer right
	 * before the send.
	 */
	private final DatagramPacket dp = new DatagramPacket(new byte[1], 0);

	/**
	 * Output UDP socket for sending packets.
	 */
	private final DatagramSocket ds;
	
	PACKAGE pj_UDPLayer()
		throws SocketException
	{
		this.ds = new DatagramSocket();
	}

	//public String toString()
	//{
	//	return "UDPLayer[ds=" +ds+ "]";
	//}

	/**
	 * The table of SocketReaders that we have open, organized by port
	 * number.
	 */
	private final Hashtable sockets = new Hashtable();

	PACKAGE String getID()
	{
		return "udp";
	}

	PACKAGE int getPacketLength()
	{
		return 16384;
	}

	private pj_UDPSocketReader getReader(InetAddress src,
					     InetAddress dst,
					     int port)
		throws SocketException,
		       IOException
	{
		pj_UDPSocketReader retval;

		synchronized( this.sockets )
		{
			java.lang.Thread srThread;
			Integer iport;
			
			iport = new Integer(port);
			if( (retval = (pj_UDPSocketReader)
			     this.sockets.get(iport)) == null )
			{
				retval = new pj_UDPSocketReader(
					src,
					dst,
					iport.intValue(),
					BufferHandle.create());
				srThread = new java.lang.Thread(retval);
				srThread.setName(
					"SocketReader for pj_UDPLayer port "
					+ port);
				srThread.start();
				retval.waitFor();
				
				this.sockets.put(iport, retval);
			}
		}
		return retval;
	}

	private DatagramSocket getSocket(AddressSpecPortAddr aspa)
	{
		DatagramSocket retval = null;
		int port;

		port = aspa.sourcePort();
		if( port == AddressSpecPortAddr.WILDCARD_PORT )
		{
			retval = this.ds;
		}
		else if( (aspa.sourceAddr() ==
			  AddressSpecPortAddr.SOURCE_ADDRESS) ||
			 (aspa.sourceAddr() ==
			  AddressSpecPortAddr.WILDCARD_ADDRESS) )
		{
			pj_UDPSocketReader sr;

			if( (sr = (pj_UDPSocketReader)this.sockets.
			     get(new Integer(port))) != null )
			{
				retval = sr.ds;
			}
		}
		return retval;
	}

	PACKAGE AddressSpec createAddress(String addr)
		throws PacketAddressFormatException,
		       IOException
	{
		return new AddressSpecPortAddr(addr);
	}

	PACKAGE PacketAddressMask createAddressMask()
	{
		return new pj_PortAddressMask();
	}

	PACKAGE void attachHandler(PacketHandler ph)
		throws IOException
	{
		pj_PortAddressMask pam = (pj_PortAddressMask)ph.mask;
		InetAddress ia = null;
		int port;

		if( pam.destMask != null )
			ia = pam.destMask.reduce();
		port = pam.destPort;
		synchronized(this.sockets)
		{
			/* XXX Fix Me */
			this.getReader(ia, ia, port).masks.addHandler(ph);
		}
	}

	PACKAGE void detachHandler(PacketHandler ph)
	{
		int port = ((pj_PortAddressMask)ph.mask).destPort;
		Integer i = new Integer(port);
		pj_UDPSocketReader sr;

		synchronized(this.sockets)
		{
			sr = (pj_UDPSocketReader)this.sockets.get(i);
			if( (sr != null) &&
			    sr.masks.remHandler(ph) &&
			    (sr.writers == 0) )
			{
				sr.close();
				this.sockets.remove(i);
			}
		}
	}

	PACKAGE void attachGenerator(PacketGenerator pg)
		throws IOException
	{
		AddressSpecPortAddr aspa;

		aspa = (AddressSpecPortAddr)pg.destination;
		if( (aspa.sourcePort() != AddressSpecPortAddr.WILDCARD_PORT) &&
		    ((aspa.sourceAddr() ==
		      AddressSpecPortAddr.SOURCE_ADDRESS) ||
		     (aspa.sourceAddr() ==
		      AddressSpecPortAddr.WILDCARD_ADDRESS)) )
		{
			pj_UDPSocketReader sr;

			sr = this.getReader(aspa.sourceAddr(),
					    aspa.destAddr(),
					    aspa.sourcePort());
			sr.addWriter();
		}
	}

	/**
	 * Synchronous send a packet, blocking until the send
	 * completes.
	 */
	PACKAGE void synchronousSend(PacketGenerator pg, BufferHandle bh)
		throws IOException
	{
		AddressSpecPortAddr ua;
		boolean retval = true;
		
		// XXX should clear out the asynchronous queue first!

		// XXX assert the reference count on bh is correct

		/* Get the destination and setup the datagram packet */
		ua = (AddressSpecPortAddr)pg.destination;

		synchronized(this.dp)
		{
			DatagramSocket sock;
			
			this.dp.setAddress(ua.destAddr());
			this.dp.setPort(ua.destPort());

			// XXX need a synchronized(bh) retrevial of byte[] and length?
			this.dp.setData(bh.getByteArray());
			this.dp.setLength(bh.getLength());

			if( (sock = this.getSocket(ua)) != null )
			{
				sock.send(this.dp);
			}
			else
			{
				throw new IOException("No socket found?");
			}
		}
	}
	
	/**
	 * Signal to UDP layer that asynchronous packets are
	 * available.  We just do a synchronous send.  :)
	 */

	PACKAGE void generatorReady(PacketGenerator pg)
	{
		BufferHandle pb;

		try
		{
			while( (pb = pg.generate()) != null )
			{
				synchronousSend(pg, pb);
			}
		}
		catch(IOException e)
		{
		}
	}

	PACKAGE void detachGenerator(PacketGenerator pg)
	{
		AddressSpecPortAddr aspa;

		aspa = (AddressSpecPortAddr)pg.destination;
		if( (aspa.sourcePort() != AddressSpecPortAddr.WILDCARD_PORT) &&
		    ((aspa.sourceAddr() ==
		      AddressSpecPortAddr.SOURCE_ADDRESS) ||
		     (aspa.sourceAddr() ==
		      AddressSpecPortAddr.WILDCARD_ADDRESS)) )
		{
			synchronized(this.sockets)
			{
				pj_UDPSocketReader sr;
				Integer i;

				i = new Integer(aspa.sourcePort());
				sr = (pj_UDPSocketReader)this.sockets.get(i);
				if( (sr != null) && sr.remWriter() )
				{
					this.sockets.remove(i);
				}
			}
		}
	}
}
