/*
 * pj_UDPSocketReader.Java
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.SocketException;
import java.net.MulticastSocket;

/**
 * Class used to read and dispatch packets coming in on a UDP socket.
 */
PACKAGE class pj_UDPSocketReader
	implements Runnable
{
	/**
	 * Dummy socket we use to wake up readers that are stuck in a
	 * receive.
	 */
	private static DatagramSocket dummySocket;

	static {
		try
		{
			dummySocket = new DatagramSocket();
		}
		catch(SocketException e)
		{
			Node.printStackTrace(e);
			Node.fatalError("Can't open pj_UDPSocketReader dummy "
					+ "port");
		}
	}
	
	/**
	 * Dummy packet used to wake the reader thread when it needs to die.
	 */
	private static final DatagramPacket dummyPacket =
		new DatagramPacket(new byte[1], 1);

	/**
	 * A set of call backs that claim to handle the packets that we
	 * receive.
	 */
	PACKAGE final pj_PacketHandlerList masks;

	/**
	 * The socket we're listening on.
	 */
	PACKAGE final DatagramSocket ds;

	private static final int STATE_NEWBORN = 0;
	private static final int STATE_RUNNING = 1;
	private static final int STATE_DYING = 2;
	private static final int STATE_DEAD = 3;
	private static final int STATE_BURIED = 4;
	
	/**
	 * Indicates whether we should continue listening for packets.
	 */
	private int state = STATE_NEWBORN;

	/**
	 * The number of people using this reader's socket to send packets.
	 */
	PACKAGE int writers = 0;
	
	/**
	 * Initial packet buffer that first packet is recevied into.
	 */
	private final BufferHandle pb;

	/**
	 * Construct a new SocketReader that listens on the given port.
	 *
	 * @param port The UDP port number to listen on.
	 * @param pb The BufferHandle object that datagrams should be read
	 * @param pb into.
	 * @exception Throws a SocketException if there was a problem with
	 * @exception creating the port.
	 */
	PACKAGE pj_UDPSocketReader(InetAddress src,
				   InetAddress dst,
				   int port,
				   BufferHandle pb)
		throws SocketException,
		       IOException
	{
		require((port == -1) || (port >= 0));
		require(port < 65535);
		require(pb);

		this.masks = new pj_PacketHandlerList();
		if( (dst != null) && dst.isMulticastAddress() )
		{
			MulticastSocket ms = new MulticastSocket(port);

			ms.joinGroup(dst);
			ms.setTTL((byte)1);
			this.ds = ms;
		}
		else if( port == -1 )
		{
			try
			{
				this.ds = new DatagramSocket();
			}
			catch(SocketException e)
			{
				Node.println("Unable to bind to address: "
					     + src
					     + "; port: "
					     + port);
				throw e;
			}
		}
		else
		{
			try
			{
				this.ds = new DatagramSocket(port, src);
			}
			catch(SocketException e)
			{
				Node.println("Unable to bind to address: "
					     + src
					     + "; port: "
					     + port);
				throw e;
			}
		}
		this.pb = pb;
	}

	PACKAGE synchronized void waitFor()
	{
		while( this.state == STATE_NEWBORN )
		{
			try
			{
				this.wait();
			}
			catch(InterruptedException e)
			{
			}
		}
	}

	/**
	 * Increment the writer user count.
	 *
	 * @return True if the socket is still available, false otherwise.
	 */
	PACKAGE synchronized boolean addWriter()
	{
		boolean retval = false;
		
		if( this.state == STATE_RUNNING )
		{
			this.writers++;
			retval = true;
		}
		return retval;
	}

	/**
	 * Decrement the writer user count for this socket.
	 */
	PACKAGE synchronized boolean remWriter()
	{
		boolean retval = false;
		
		this.writers--;
		if( (this.writers == 0) && (this.state == STATE_DYING) )
		{
			retval = true;
		}
		return retval;
	}

	private void sendDummy()
	{
		try
		{
			synchronized(dummySocket)
			{
				dummyPacket.setAddress(this.ds.
						       getLocalAddress());
				dummyPacket.setPort(this.ds.getLocalPort());
				dummySocket.send(dummyPacket);
			}
		}
		catch(IOException e)
		{
			Node.printStackTrace(e);
			Node.print("Bah!  Can't send dummy packet "
				   + "to wake up dying port");
		}
	}

	public synchronized void close()
	{
		switch( this.state )
		{
		case STATE_NEWBORN:
		case STATE_RUNNING:
			if( this.writers == 0 )
			{
				this.state = STATE_DEAD;
				while( this.state != STATE_BURIED )
				{
					this.sendDummy();
					try
					{
						this.wait(1000);
						if( this.state != STATE_BURIED )
							Node.println("Close is failing...");
					}
					catch(InterruptedException e)
					{
					}
				}
			}
			break;
		default:
			break;
		}
	}

	public void run()
	{
		pj_ProtocolLayer.PacketHandler ph;
		final DatagramPacket dp;

		require(this.state == STATE_NEWBORN);
		require(this.pb);
		require(this.pb.getByteArray().length > 0);

		synchronized(this)
		{
			this.state = STATE_RUNNING;
			this.notifyAll();
		}
		/* Put the initial buffer in DatagramPacket */
		dp = new DatagramPacket(pb.getByteArray(),
					pb.getByteArray().length);

		/* Loop until we're asked to die */
		while (this.state != STATE_DEAD)
		{
			try
			{
				/* Wait for a packet. */
				this.ds.receive(dp);

				/*
				 * Set effective length of pb to received data
				 */
				pb.setLength(dp.getLength());
				TRACE(Trace.NODEOS_INCOMING, "in!");
				TRACE(Trace.NODEOS_INCOMING, "received(len=" + dp.getLength() + "): " + Debug.dbgPacketSummary(pb));
				if( (ph = this.masks.matchMask(dp)) != null )
				{
					/* Swap up for new packet buffer */
					ph.handle(pb);

					/*
					 * pb may or may not point to a new
					 * buffer
					 */
					dp.setData(pb.getByteArray());
					dp.setLength(pb.getByteArray().
						     length);
				}
				else
				{
					TRACE(Trace.NODEOS_INCOMING, "Couldn't match packet to any masks..." + Debug.dbgPacketSummary(pb));
				}
			}
			catch(IOException e)
			{
				Node.println("IOException while "
					     + "reading from socket:"
					     + this);
				Node.printStackTrace(e);
			}
			catch(Throwable th)
			{
				Node.printStackTrace(th);
				Node.fatalError("In pj_UDPSocketReader "
						+ "thread "
						+ this);
			}
		}
		TRACE(Trace.NODEOS_INCOMING, "recv port (on " +this.ds.getLocalPort()+ "): exiting");
		this.ds.close();
		synchronized(this)
		{
			this.state = STATE_BURIED;
			this.notifyAll();
		}
	}

	public String toString()
	{
		return "pj_UDPSocketReader[masks="
			+ this.masks
			+ "; ds="
			+ this.ds
			+ "; state="
			+ this.state
			+ "; writers="
			+ this.writers
			+ "; pb="
			+ this.pb
			+ "]";
	}
}
