/*
 * IPHeader.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_PROTOCOLS_PKG.ip;

import JANOS_UTIL_PKG.PacketSegment;
import JANOS_UTIL_PKG.InvalidPacketException;

import JANOS_NODEOS_PKG.Node;
import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.DemultiplexKey;

/**
 * The IPHeader is a PacketSegment for the standard IP header.
 */
public class IPHeader
	implements PacketSegment
{
	/**
	 * Type of service constants.
	 */
	public static final byte TOS_LOWDELAY		= 0x10;
	public static final byte TOS_THROUGHPUT		= 0x08;
	public static final byte TOS_RELIABILITY	= 0x04;
	public static final byte TOS_MINCOST		= 0x02;

	/**
	 * Protocol types.
	 */
	public static final byte PROTO_IP		=  0;
	public static final byte PROTO_ICMP		=  1;
	public static final byte PROTO_TCP		=  6;
	public static final byte PROTO_UDP		= 17;

	/**
	 * Maximum possible TTL.
	 */
	public static final byte MAX_TTL		= (byte)255;

	/**
	 * Offsets into an IP header in a buffer.
	 */
	public static final int OFFSET_VHL		=  0;
	public static final int OFFSET_TOS		=  1;
	public static final int OFFSET_LEN		=  2;
	public static final int OFFSET_ID		=  4;
	public static final int OFFSET_OFF		=  6;
	public static final int OFFSET_TTL		=  8;
	public static final int OFFSET_PROTOCOL		=  9;
	public static final int OFFSET_SUM		= 10;
	public static final int OFFSET_SRC		= 12;
	public static final int OFFSET_DST		= 16;

	/**
	 * IP header data.
	 */
	private byte version;
	private byte headerLength;
	private byte typeOfService;
	private short identification;
	private short fragOffset;
	private byte ttl;
	private byte protocol;
	private int src;
	private int dest;

	/**
	 * Length of the header with no options.
	 */
	public static final int HEADER_LENGTH_NO_OPTIONS = 20;

	/**
	 * Construct an IPHeader with the given values.
	 */
	public IPHeader(byte typeOfService,
			short identification,
			short fragOffset,
			byte ttl,
			byte protocol,
			int src,
			int dest)
	{
		this.version = 4;
		this.headerLength = HEADER_LENGTH_NO_OPTIONS;
		this.typeOfService = typeOfService;
		this.identification = identification;
		this.fragOffset = fragOffset;
		this.ttl = ttl;
		this.protocol = protocol;
		this.src = src;
		this.dest = dest;
	}
	
	public IPHeader()
	{
		this.version = 4;
		this.headerLength = HEADER_LENGTH_NO_OPTIONS;
	}

	public int getLength()
	{
		return this.headerLength;
	}

	public void newID()
	{
		this.identification++;
	}

	public int getSource()
	{
		return this.src;
	}

	public void setSource(int src)
	{
		this.src = src;
	}

	public int getDestination()
	{
		return this.dest;
	}

	public static int getDestination(BufferHandle bh, int offset)
	{
		return bh.getInt(offset + OFFSET_DST);
	}

	public void setDestination(int dest)
	{
		this.dest = dest;
	}

	public void setTTL(byte ttl)
	{
		this.ttl = ttl;
	}

	public byte getTTL()
	{
		return this.ttl;
	}
	
	public void consumeTTL()
	{
		this.ttl--;
	}

	public static boolean consumeTTL(BufferHandle bh, int offset)
	{
		boolean retval = true;
		byte ttl;

		/* XXX Ick! */
		ttl = bh.getByte(offset + OFFSET_TTL);
		bh.setByte(offset + OFFSET_TTL, (byte)(ttl - 1));
		if( ttl == 1 )
		{
			retval = false;
		}
		else
		{
			int cs = bh.getShort(offset + OFFSET_SUM) & 0xffff;

			cs += 0x0100;
			cs += (cs >> 16);
			bh.setShort(offset + OFFSET_SUM, (short)cs);
		}
		return retval;
	}
	
	public int marshal(BufferHandle bh, int offset)
	{
		short checksum;
		
		bh.setByte(offset + OFFSET_VHL,
			   (byte)((this.version << 4)
				  | (this.headerLength >> 2)));
		bh.setByte(offset + OFFSET_TOS, this.typeOfService);
		bh.setShort(offset + OFFSET_LEN, (short)bh.getLength());
		bh.setShort(offset + OFFSET_ID, this.identification);
		bh.setShort(offset + OFFSET_OFF, this.fragOffset);
		bh.setByte(offset + OFFSET_TTL, this.ttl);
		bh.setByte(offset + OFFSET_PROTOCOL, this.protocol);
		bh.setShort(offset + OFFSET_SUM, (short)0);
		bh.setInt(offset + OFFSET_SRC, this.src);
		bh.setInt(offset + OFFSET_DST, this.dest);
		checksum = bh.computeChecksum(offset, this.headerLength);
		bh.setShort(offset + OFFSET_SUM, checksum);
		return offset + this.headerLength;
	}

	public int unmarshal(BufferHandle bh, int offset)
		throws InvalidPacketException
	{
		short totalLength;
		byte tmp;

		tmp = bh.getByte(offset + OFFSET_VHL);
		this.version = (byte)((tmp >> 4) & 0xff);
		if( this.version != 4 )
			throw new InvalidPacketException(
				"Can only handle ipv4");
		this.headerLength = (byte)((tmp & 0x0f) << 2);
		if( bh.computeChecksum(offset, this.headerLength) != 0 )
			throw new InvalidPacketException("Bad IP Checksum");
		this.typeOfService = bh.getByte(offset + OFFSET_TOS);
		totalLength = bh.getShort(offset + OFFSET_LEN);
		if( totalLength != bh.getLength() )
			throw new InvalidPacketException("Invalid length");
		this.identification = bh.getShort(offset + OFFSET_ID);
		this.fragOffset = bh.getShort(offset + OFFSET_OFF);
		this.ttl = bh.getByte(offset + OFFSET_TTL);
		this.protocol = bh.getByte(offset + OFFSET_PROTOCOL);
		this.src = bh.getInt(offset + OFFSET_SRC);
		this.dest = bh.getInt(offset + OFFSET_DST);
		return offset + this.headerLength;
	}

	public int hashCode()
	{
		return this.src;
	}
	
	public boolean equals(Object obj)
	{
		boolean retval = false;

		if( obj instanceof IPHeader )
		{
			IPHeader ip = (IPHeader)obj;

			if( this.version == ip.version
			    && this.headerLength == ip.headerLength
			    && this.typeOfService == ip.typeOfService
			    && this.identification == ip.identification
			    && this.fragOffset == ip.fragOffset
			    && this.ttl == ip.ttl
			    && this.protocol == ip.protocol
			    && this.src == ip.src
			    && this.dest == ip.dest )
			{
				retval = true;
			}
		}
		return retval;
	}
	
	public String toString()
	{
		return "[IPHeader version="
			+ this.version
			+ "; headerLength="
			+ this.headerLength
			+ "; typeOfService="
			+ this.typeOfService
			+ "; identification="
			+ this.identification
			+ "; fragOffset="
			+ this.fragOffset
			+ "; ttl="
			+ (this.ttl & 0xff)
			+ "; protocol="
			+ this.protocol
			+ "; src="
			+ Address.toString(this.src)
			+ "; dst="
			+ Address.toString(this.dest)
			+ "]";
	}
}
