/*
 * UDPHeader.Java
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_PROTOCOLS_PKG.udp;

import JANOS_UTIL_PKG.PacketSegment;
import JANOS_UTIL_PKG.InvalidPacketException;

import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.DemultiplexKey;

/**
 * The UDPHeader is a PacketSegment for the standard UDP header.
 */
public class UDPHeader
	implements PacketSegment
{
	/**
	 * Offsets into a UDP header in a buffer.
	 */
	public static final int OFFSET_SOURCE_PORT	= 0;
	public static final int OFFSET_DESTINATION_PORT	= 2;
	public static final int OFFSET_LENGTH		= 4;
	public static final int OFFSET_CHECKSUM		= 6;

	/**
	 * UDP header data.
	 */
	private short sourcePort;
	private short destinationPort;
	private short length;

	/**
	 * Construct a UDPHeader with the given values.
	 *
	 * @param sourcePort The source port of this packet.
	 * @param destinationPort The destination port of this packet.
	 * @param bodyLength The length of the packet body.
	 */
	public UDPHeader(short sourcePort,
			 short destinationPort,
			 short bodyLength)
	{
		this.sourcePort = sourcePort;
		this.destinationPort = destinationPort;
		this.length = (short)(bodyLength + 8);
	}

	/**
	 * Construct an empty UDPHeader.
	 */
	public UDPHeader()
	{
		this.sourcePort = 0;
		this.destinationPort = 0;
		this.length = 8;
	}

	/**
	 * @return The source port.
	 */
	public short getSourcePort()
	{
		return this.sourcePort;
	}

	/**
	 * @param port The new source port value.
	 */
	public void setSourcePort(short port)
	{
		this.sourcePort = port;
	}

	/**
	 * @return The destination port.
	 */
	public short getDestinationPort()
	{
		return this.destinationPort;
	}

	/**
	 * @param port The new destination port value.
	 */
	public void setDestinationPort(short port)
	{
		this.destinationPort = port;
	}

	/**
	 * @return The length of a UDPHeader.
	 */
	public int getLength()
	{
		return 8;
	}

	/**
	 * @return The length of the body.
	 */
	public int getBodyLength()
	{
		return this.length - 8;
	}

	/**
	 * @return The total length of the UDP portion of the packet.
	 */
	public int getTotalLength()
	{
		return this.length;
	}

	public int marshal(BufferHandle bh, int offset)
	{
		short checksum;
		
		bh.setShort(offset + OFFSET_SOURCE_PORT, this.sourcePort);
		bh.setShort(offset + OFFSET_DESTINATION_PORT,
			    this.destinationPort);
		bh.setShort(offset + OFFSET_LENGTH, this.length);
		bh.setShort(offset + OFFSET_CHECKSUM, (short)0);
		//checksum = bh.computeChecksum(offset, bh.getLength() - offset);
		//bh.setShort(offset + OFFSET_CHECKSUM, checksum);
		return offset + 8;
	}

	public int unmarshal(BufferHandle bh, int offset)
		throws InvalidPacketException
	{
		this.sourcePort = bh.getShort(offset + OFFSET_SOURCE_PORT);
		this.destinationPort = bh.getShort(offset
						   + OFFSET_DESTINATION_PORT);
		this.length = bh.getShort(offset + OFFSET_LENGTH);
		/* Is this too strict? */
		if( (offset + this.length) != bh.getLength() )
			throw new InvalidPacketException("Bogus UDP length");
		//if( bh.computeChecksum(offset, this.length) != 0 )
		//	throw new InvalidPacketException("Bad UDP checksum");
		return offset + 8;
	}

	public int hashCode()
	{
		return ((this.sourcePort << 16) & 0xffff0000)
			| (this.destinationPort & 0x0000ffff);
	}

	public boolean equals(Object obj)
	{
		boolean retval = false;

		if( obj instanceof UDPHeader )
		{
			UDPHeader udp = (UDPHeader)obj;

			if( (this.sourcePort == udp.sourcePort)
			    && (this.destinationPort == udp.destinationPort)
			    && (this.length == udp.length) )
			{
				retval = true;
			}
		}
		return retval;
	}

	public String toString()
	{
		return "[UDPHeader sourcePort="
			+ this.sourcePort
			+ "; destinationPort="
			+ this.destinationPort
			+ "; length="
			+ this.length
			+ "]";
	}
}
