/*
 * RouteEntry.Java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_ROUTE_PKG;

import JANOS_NODEOS_PKG.Address;

/**
 * A simple class representing a route in the table.
 */
public class RouteEntry
	implements Cloneable
{
	public static final int FULL_MASK = ~0;
	
	/**
	 * The destination address.
	 */
	private int address;

	/**
	 * The applicable bits of the destination address.
	 */
	private int addressMask;

	/**
	 * The object representing the next hop to the destination address.
	 */
	private final Object nextHop;

	/**
	 * Copy constructor
	 *
	 * @param re The RouteEntry to copy.
	 */
	public RouteEntry(RouteEntry re)
	{
		this(re.address, re.addressMask, re.nextHop);
	}

	/**
	 * Construct a RouteEntry with the given values.
	 *
	 * @param address The destination address.
	 * @param addressMask The applicable bits of the destination address.
	 * @param nextHop The object representing the next hop to the
	 * @param nextHop destination address.
	 */
	public RouteEntry(int address, int addressMask, Object nextHop)
	{
		this.address = address & addressMask;
		this.addressMask = addressMask;
		this.nextHop = nextHop;
	}

	/**
	 * Construct a RouteEntry with the given values and 0xffffffff for the
	 * address mask.
	 *
	 * @param address The destination address.
	 * @param nextHop The object representing the next hop to the
	 * @param nextHop destination address.
	 */
	public RouteEntry(int address, Object nextHop)
	{
		this(address, FULL_MASK, nextHop);
	}

	/**
	 * @return The destination address.
	 */
	public int getAddress()
	{
		return this.address;
	}

	/**
	 * @param am The new value of the address mask.
	 */
	void setAddressMask(int am)
	{
		this.addressMask = am;
		this.address = this.address & am;
	}

	/**
	 * @return The address mask.
	 */
	public int getAddressMask()
	{
		return this.addressMask;
	}

	/**
	 * @return The object representing the next hop to the destination.
	 */
	public Object getNextHop()
	{
		return this.nextHop;
	}

	/**
	 * @param re The RouteEntry to compare too.
	 * @return The longest mask of bits starting from the left that the
	 * @return addresses match.
	 */
	int longestMatchingPrefix(RouteEntry re)
	{
		boolean done = false;
		int lpc, retval = 0;
		
		for( lpc = 31; !done && (lpc >= 0); lpc-- )
		{
			int mask = (1 << lpc);
			
			if( (this.address & mask) == (re.address & mask) )
			{
				retval |= mask;
			}
			else
			{
				done = true;
			}
		}
		return retval;
	}

	/**
	 * @param destinationAddress The address to compare to.
	 * @return True if this RouteEntry matches the given address.
	 */
	public int matches(int destinationAddress)
	{
		int retval = 0;

		if( (destinationAddress & this.addressMask) == this.address )
		{
			retval = this.addressMask;
		}
		return retval;
	}

	public int hashCode()
	{
		return this.address;
	}

	public boolean equals(Object obj)
	{
		boolean retval = false;

		if( obj instanceof RouteEntry )
		{
			RouteEntry re = (RouteEntry)obj;

			if( (this.address == re.address) &&
			    (this.addressMask == re.addressMask) &&
			    this.nextHop.equals(re.nextHop) )
			{
				retval = true;
			}
		}
		return retval;
	}

	public Object clone()
	{
		try
		{
			Object retval = super.clone();

			return retval;
		}
		catch(CloneNotSupportedException e)
		{
			throw new InternalError("RouteEntry isn't Cloneable?");
		}
	}

	public String toString()
	{
		return "RouteEntry[address="
			+ Address.toString(this.address)
			+ "; addressMask=0x"
			+ Integer.toHexString(this.addressMask)
			+ "; nextHop="
			+ this.nextHop
			+ "]";
	}
}
