/*
 * RouteTrie.Java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_ROUTE_PKG;

import java.util.Vector;

/**
 * Class that maintains the tree of RouteEntry's.
 */
final class RouteTrie
	extends RouteEntry
{
	/**
	 * The default size of the entry array.
	 */
	private static final int DEFAULT_SIZE = 10;

	/**
	 * The default value used to increment the size of the entry array.
	 */
	private static final int DEFAULT_INCREMENT = 10;

	/**
	 * The array of child RouteEntry/RouteTrie's.
	 */
	private RouteEntry entries[] = new RouteEntry[DEFAULT_SIZE];

	/**
	 * The actual number of used elements in this.entries.
	 */
	private int usedEntries = 0;

	/**
	 * Construct a RouteTrie with the given values.
	 *
	 * @param address The destination address.
	 * @param addressMask The applicable bits of the destination address.
	 * @param nextHop The object representing the next hop to the
	 * @param nextHop destination address.
	 */
	RouteTrie(int address, int addressMask, Object nextHop)
	{
		super(address, addressMask, nextHop);
	}

	/**
	 * @param res An array to fill with the table's RouteEntry's.
	 * @return The number of elements used in the `res' array.
	 */
	int getEntries(RouteEntry res[], int used)
	{
		int lpc, retval = 0;

		for( lpc = 0;
		     (used < res.length) && (lpc < this.usedEntries);
		     lpc++ )
		{
			RouteEntry re = this.entries[lpc];

			if( re instanceof RouteTrie )
			{
				RouteTrie rt;

				rt = (RouteTrie)re;
				used += rt.getEntries(res, used);
			}
			else
			{
				res[used] = re;
				used += 1;
				retval += 1;
			}
		}
		return retval;
	}
	
	/**
	 * @param re The entry to add to the table.
	 */
	int addEntry(RouteEntry re)
	{
		int lmp, bmp = this.getAddressMask(), bestSlot = -1, lpc;
		int retval = 1;

		for( lpc = 0; lpc < this.usedEntries; lpc++ )
		{
			lmp = this.entries[lpc].longestMatchingPrefix(re);
			if( lmp < bmp )
			{
				bmp = lmp;
				bestSlot = lpc;
			}
		}
		if( bestSlot == -1 )
		{
			if( this.usedEntries == this.entries.length )
			{
				RouteEntry oldEntries[];
				
				oldEntries = this.entries;
				this.entries = new RouteEntry[
					oldEntries.length + DEFAULT_INCREMENT];
				System.arraycopy(oldEntries,
						 0,
						 this.entries,
						 0,
						 oldEntries.length);
			}
			this.entries[this.usedEntries] = re;
			this.usedEntries += 1;
		}
		else
		{
			RouteTrie rt;
			
			if( this.entries[bestSlot] instanceof RouteTrie )
			{
				rt = (RouteTrie)this.entries[bestSlot];
				rt.setAddressMask(bmp);
				rt.addEntry(re);
			}
			else if( re.getNextHop().
				 equals(this.entries[bestSlot].getNextHop()) )
			{
				this.entries[bestSlot].setAddressMask(bmp);
				retval = 0;
			}
			else
			{
				RouteEntry old;

				old = this.entries[bestSlot];
				rt = new RouteTrie(old.getAddress(),
						   bmp,
						   null);
				rt.addEntry(old);
				rt.addEntry(re);
				this.entries[bestSlot] = rt;
			}
		}
		return( retval );
	}

	/**
	 * XXX Needs to take addressMask.
	 *
	 * @param address The address of the entries to remove.
	 * @return True if the entry was found and removed.
	 */
	boolean removeEntry(int address)
	{
		boolean retval = false;
		int lpc, fillCount = 0;

		for( lpc = 0; (lpc < this.usedEntries); lpc++ )
		{
			RouteEntry re = this.entries[lpc];

			if( fillCount > 0 )
			{
				this.entries[lpc - fillCount] = re;
			}
			if( re.matches(address) != 0 )
			{
				if( re instanceof RouteTrie )
				{
					RouteTrie rt = (RouteTrie)re;
					
					retval = rt.removeEntry(address);
					switch( rt.usedEntries )
					{
					case 0:
						this.usedEntries -= 1;
						fillCount += 1;
						break;
					case 1:
						this.entries[lpc - fillCount] =
							rt.entries[0];
						break;
					default:
						break;
					}
				}
				else
				{
					this.usedEntries -= 1;
					fillCount += 1;
					retval = true;
				}
			}
		}
		return retval;
	}

	/**
	 * @param hop The hop to remove from the table.
	 * @return The number of entries removed from the table.
	 */
	int removeHop(Object hop)
	{
		int retval = 0, lpc, fillCount = 0;

		for( lpc = 0; lpc < this.usedEntries; lpc++ )
		{
			RouteEntry re = this.entries[lpc];

			if( fillCount > 0 )
			{
				this.entries[lpc - fillCount] = re;
			}
			if( re instanceof RouteTrie )
			{
				RouteTrie rt = (RouteTrie)re;
				
				retval += rt.removeHop(hop);
				switch( rt.usedEntries )
				{
				case 0:
					this.usedEntries -= 1;
					fillCount += 1;
					break;
				case 1:
					this.entries[lpc - fillCount] =
						rt.entries[0];
					break;
				default:
					break;
				}
			}
			else if( re.getNextHop() == hop )
			{
				retval += 1;
				this.usedEntries -= 1;
				fillCount += 1;
			}
		}
		return retval;
	}

	/**
	 * @param address The destination address to find.
	 * @return The RouteEntry matching the given address or null.
	 */
	RouteEntry lookup(int address)
	{
		int lpc, match, best = Integer.MIN_VALUE;
		RouteEntry retval = null;

		for( lpc = 0; lpc < this.usedEntries; lpc++ )
		{
			RouteEntry re = this.entries[lpc];

			if( ((match = re.matches(address)) < 0) &&
			    (match >= best) )
			{
				if( re instanceof RouteTrie )
				{
					RouteTrie rt = (RouteTrie)re;
					
					retval = rt.lookup(address);
				}
				else
				{
					retval = re;
				}
			}
		}
		return retval;
	}
	
	public String toString()
	{
		String retval = "RouteTrie["
			+ super.toString()
			+ "; usedEntries="
			+ this.usedEntries
			+ "; entries={";
		int lpc;

		for( lpc = 0; lpc < this.usedEntries; lpc++ )
		{
			retval += " " + this.entries[lpc];
		}
		retval += " }";
		return retval;
	}
}
