/*
 * ICMPPacketHandler.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_SERVICES_PKG.icmp;

import java.io.IOException;

import JANOS_NODEOS_PKG.Domain;
import JANOS_NODEOS_PKG.Node;
import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.InChannel;
import JANOS_NODEOS_PKG.OutChannel;
import JANOS_NODEOS_PKG.DeliverFunc;
import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.DemultiplexKey;
import JANOS_NODEOS_PKG.ProtocolSpecFormatException;
import JANOS_NODEOS_PKG.PacketAddressFormatException;

import JANOS_PROTOCOLS_PKG.ip.IPHeader;
import JANOS_PROTOCOLS_PKG.icmp.ICMPHeader;

import JANOS_UTIL_PKG.InvalidPacketException;

PACKAGE class ICMPPacketHandler
	implements DeliverFunc
{
	private String protocolSpec;
	private int address;
	private ICMPDomain domain;
	private IPHeader ip = new IPHeader();
	private ICMPHeader icmp = new ICMPHeader();
	private InChannel ic;

	PACKAGE ICMPPacketHandler(ICMPDomain domain,
				  int address,
				  String protocolSpec)
		throws ProtocolSpecFormatException,
		       PacketAddressFormatException,
		       IOException
	{
		DemultiplexKey dk;

		this.domain = domain;
		this.address = address;
		this.protocolSpec = protocolSpec;
		dk = new DemultiplexKey(domain.getDomain());
		/* ICMP Packets */
		IPHeader.matchProtocol(dk, 0, IPHeader.PROTO_ICMP);
		/* ICMP Packets addressed to this node */
		IPHeader.matchDestination(dk, 0, this.address);
		Node.println("Processing ICMP Packets for "
			     + Address.toString(address)
			     + " on "
			     + protocolSpec);
		this.ic = InChannel.create(Domain.currentDomain(),
					   dk,
					   protocolSpec,
					   "0:0x800",
					   this);
		this.ic.addBuffers(10);
	}

	public void handlePacket(BufferHandle bh, int payloadOffset)
		throws InvalidPacketException
	{
		int ipOffset, icmpOffset;
		
		ipOffset = this.ip.unmarshal(bh, 0);
		icmpOffset = this.icmp.unmarshal(bh, ipOffset);
		switch( icmp.getType() )
		{
		case ICMPHeader.ECHO:
			{
				OutChannel oc;
				int tmp;

				tmp = this.ip.getSource();
				this.ip.setSource(this.ip.getDestination());
				this.ip.setDestination(tmp);
				this.ip.consumeTTL();
				this.icmp.setType(ICMPHeader.ECHOREPLY);
				this.icmp.marshal(bh, ipOffset);
				this.ip.marshal(bh, 0);
				if( (oc = this.domain.
				     lookupRoute(this.protocolSpec, tmp)) !=
				    null )
				{
					BufferHandle bh_copy;

					bh_copy = BufferHandle.create(bh);
					oc.send(bh_copy);
				}
				else
				{
					Node.println("No route to host "
						     + Address.toString(tmp));
				}
			}
			break;
		default:
			break;
		}
	}
	
	public synchronized void receive(BufferHandle bh, int payloadOffset)
	{
		try
		{
			//Node.println("Received packet...");
			this.handlePacket(bh, payloadOffset);
		}
		catch(InvalidPacketException e)
		{
			Node.println("Bogus ICMP packet received...\n" + e);
			Node.printStackTrace(e);
		}
		catch(Throwable th)
		{
			Node.println("Exception during packet processing\n"
				     + th);
			Node.printStackTrace(th);
		}
		this.ic.addBuffer(bh);
	}
}
