/*
 * Test Domain API a bit
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_TESTS_PKG;

import JANOS_NODEOS_PKG.Thread;
import JANOS_NODEOS_PKG.Domain;
import JANOS_NODEOS_PKG.MemSpec;
import JANOS_NODEOS_PKG.Node;
import JANOS_NODEOS_PKG.DomainRun;
import JVM_COMMSPACE_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.Credentials;
import JANOS_NODEOS_PKG.ThreadPoolSpec;
import JANOS_NODEOS_PKG.DomainStartSpec;
import JANOS_NODEOS_PKG.DomainHandle;
import JANOS_NODEOS_PKG.DomainStartException;
import JANOS_UTIL_PKG.DomainBase;

import java.util.Random;

public class DomainTest
	extends DomainBase
{
	private static class TestThread
		implements Runnable
	{
		static int ids = 1;
		int id;
		
		public TestThread()
		{
			this.id = ids++;
		}
		
		public void run()
		{
			Node.println("Thread " +id+ ") running in domain " +Domain.currentDomain());
		}
	}

	private static class TermHandler
		implements Runnable
	{
		private final boolean doExit;

		public TermHandler(boolean doExit)
		{
			this.doExit = doExit;
		}

		public void run()
		{
			Node.println("Domain termhandler run");
			if (this.doExit)
			{
				Node.println("Domain termhandler shutting down VM.");
				Node.shutdown();
			}
		}
	}

	public static class DomainStartup
		extends DomainRun
	{
		public void run(Domain domain,
			String[] args,
			CommSpaceHandle csh)
		{
			Node.println("Domain " +args[0]+ " started");
		}
	}
	

	private static String names[] = 
	{
		"ADomainName",
		"",
		null,
		"Another @#$!@% name"
	};

	private static Object objects[] =
	{
		new String("<a string object>"),
		new Throwable(),
		null,
		"Foo"
	};
	
	private void nameCheck(Domain f)
	{
		String name = f.getName();
		for (int i = 0; i < names.length; i++)
		{
			Node.println("  Setting name to: " +names[i]);
			f.setName(names[i]);
			Node.println("  Name is: " +f.getName());
		}
		f.setName(name);
	}
	
	private void objCheck(Domain f)
	{
		for (int i = 0; i < objects.length; i++)
		{
			Node.println("  Setting object to: " +objects[i]);
			f.setObject(objects[i]);
			Node.println("  Object is: " +f.getObject());
		}
	}

	private static void sleep(int millis)
	{
		sleep(millis, 0);
	}

	private static void sleep(int millis, int nanos)
	{
		try
		{
			Thread.sleep(millis, nanos);
		}
		catch (InterruptedException ie)
		{
			Node.println("Sleep interrupted.  Oops.");
		}
	}
	

	public void testCurrDomain()
	{
		Domain currDomain = Domain.currentDomain();

		Node.println("Testing domain:" +currDomain);
		Node.println("Domain name is " +currDomain.getName());

		Node.println("Name tests:");
		nameCheck(currDomain);
		Node.println("Object tests:");
		objCheck(currDomain);
		Node.println("Thread start tests:");
		currDomain.threadStart(new TestThread(),
			"Test Thread 1");

		sleep(500); // half second
	}

	private void forkChild(Domain currDomain, Random rgen)
		throws DomainStartException
	{
		DomainHandle child;

		child = currDomain.newDomain(new Credentials("cf1"),
			new ThreadPoolSpec(10, 0, null),
			new MemSpec(),
			new DomainStartSpec(DomainStartup.class.getName(),
				new String[] { "true" },
				null
				),
			new TermHandler(false));
		
		// 0 to 500 mills
		int millis = 250 + rgen.nextInt() % 250;
		
		sleep(millis);
		
		child.destroy();
	}

	public void start()
	{
		Domain currDomain = Domain.currentDomain();
		DomainHandle child;
		Random rgen = new Random(1973); // repeatable series of tests.

		testCurrDomain();

		for (int i = 0; i < 10; i++)
		{
			try
			{
				forkChild(currDomain, rgen);
			}
			catch(DomainStartException e)
			{
				Node.println("Couldn't start up child #"
					     + i);
				Node.printStackTrace(e);
			}
		}
		
		sleep(2500); // 2.5 seconds

		Node.println("Root Domain done, shutting down.");
		Node.shutdown();
	}
}

