/*
 * UdpTest.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package JANOS_TESTS_PKG;

import JANOS_NODEOS_PKG.DomainRun;
import JANOS_NODEOS_PKG.Domain;
import JANOS_NODEOS_PKG.MemSpec;
import JANOS_NODEOS_PKG.Node;
import JANOS_NODEOS_PKG.ChanStats;
import JANOS_NODEOS_PKG.OutChannel;
import JANOS_NODEOS_PKG.InChannel;
import JANOS_NODEOS_PKG.DeliverFunc;
import JANOS_NODEOS_PKG.DemultiplexKey;
import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.AddressFactory;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;
import JANOS_UTIL_PKG.DomainBase;
import JANOS_UTIL_PKG.SimpleDemultiplexKey;

import java.io.IOException;

public class UdpTest
	extends DomainBase
{
	private static final boolean VERBOSE = false;

	public UdpTest()
	{
	}

	static final int port1 = 33001;
	static final int port2 = 33002;
	static final byte[] key1 = {5, 4, 3, 2};
	static final byte[] key2 = {5, 4, 3, 1};
	static final int count = 100;

	public void start()
	{
		Node.println("Started UdpTest");

		try 
		{
			start1();
		}
		catch(Throwable e)
		{
			e.printStackTrace();
			throw new RuntimeException("test failed");
		}

		// Node is shutdown by receiver of udp packets.
	}

	public void start1()
		throws Exception
	{
		// Get address factory because we want to create
		// nodes address with specific "physical" mappings.

		AddressFactory af = new AddressFactory();
		Address addr1 = af.newAddress(1,
					      "if/ipv4/udp",
					      "*:" + port1 + ":localhost");
		Address addr2 = af.newAddress(2,
					      "if/ipv4/udp",
					      "*:" + port2 + ":localhost");

		OutChannel out1 = makeOutChannel(addr2);
		OutChannel out2 = makeOutChannel(addr1);
		InChannel in1 = makeInChannel(addr1, out1, key1, key2);
		InChannel in2 = makeInChannel(addr2, out2, key2, key1);
		in1.addBuffers(10);
		in2.addBuffers(10);

		BufferHandle b = BufferHandle.create(0, key2.length + 4);
		b.setInt(0, 0);
		b.setBytes(4, key2, 0, key2.length);

		out1.send(b);
	}

	private OutChannel makeOutChannel(Address outAddr)
		throws Exception
	{
		return OutChannel.create(domain,
					 outAddr.getProtocolSpec(),
					 outAddr.getAddressSpec());
	}

	private class UdpTest_Deliver
		implements DeliverFunc
	{
		PACKAGE InChannel myInChan;

		private final Address inAddr;
		private final OutChannel replyChannel;
		private final byte[] replyKey;

		public UdpTest_Deliver(Address inAddr,
				       OutChannel replyChannel,
				       byte[] replyKey)
		{
			this.inAddr = inAddr;
			this.replyChannel = replyChannel;
			this.replyKey = replyKey;
		}
		
		
		// XXX this only works if packets are
		// reliable and ordered
		
		public void receive(BufferHandle packet, int _)
		{
			try
			{
				receive1(packet);
			}
			catch(IOException e)
			{
				// XXX msg
				e.printStackTrace();
			}
		}
		
		public void receive1(BufferHandle packet)
			throws IOException
		{
			int i = packet.getInt(0);
			
			if (VERBOSE)
				Node.println("Got a packet. i=" +i);

			if(i <= count)
			{
				Node.println("port "
					     + inAddr.getProtocolSpec()
					     + ";"
					     + inAddr.getAddressSpec()
					     + " recv:" + i);
				packet.setInt(0, i+1);
				packet.setBytes(4, replyKey, 0, replyKey.length);

				if (VERBOSE)
					Node.println("Sending a packet. i=" +(i+1));

				replyChannel.send(packet);
				myInChan.addBuffer(packet);
			}
			
			if(i >= count)
			{
				ChanStats cs = new ChanStats();
				
				myInChan.getStats(cs);
				if( cs.getInPackets() != ((i / 2) + 1) )
				{
					Node.println("In Stats don't jive; i="
						     + i
						     + "; cs="
						     + cs);
				}
				replyChannel.getStats(cs);
				if( cs.getOutPackets() != ((i / 2) + 1) )
				{
					Node.println("Out Stats don't jive; i="
						     + i
						     + "; cs="
						     + cs);
				}
				
				if (VERBOSE)
					Node.println("Shutting the test down.");
				UdpTest.this.domain.destroy();
			}
		}
	}
	
	private InChannel makeInChannel(Address inAddr,
					OutChannel replyChannel,
					byte[] recvKey,
					byte[] replyKey)
		throws Exception
	{
		byte[] data = new byte[replyKey.length + 4];
		InChannel retval;
		System.arraycopy(replyKey, 0, data, 4, replyKey.length);

		UdpTest_Deliver deliver = new UdpTest_Deliver(inAddr, replyChannel, replyKey);

		retval = InChannel.create(domain,
					  new SimpleDemultiplexKey(this.domain,
								   4,
								   recvKey),
					  inAddr.getProtocolSpec(),
					  inAddr.getAddressSpec(),
					  deliver);

		deliver.myInChan = retval;

		return retval;
	}
}

