/*
 * NetHogDomain.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_TESTS_PKG.demo;

import JANOS_NODEOS_PKG.Domain;
import JANOS_NODEOS_PKG.Node;
import JANOS_NODEOS_PKG.MemSpec;
import JANOS_NODEOS_PKG.ThreadPoolSpec;
import JVM_COMMSPACE_PKG.CommSpaceElement;

import JANOS_PROTOCOLS_PKG.ip.IPHeader;
import JANOS_PROTOCOLS_PKG.udp.UDPHeader;

import JANOS_NODEOS_PKG.OutChannel;
import JANOS_NODEOS_PKG.NetworkSpec;
import JANOS_NODEOS_PKG.BufferHandle;

/**
 */
class NetHogDomain
	extends SubDomain
{
	private int THREAD_COUNT = 1;
	private int SLEEP_TIME = 1;
	private int RATE_DIVISOR = 20;
	private int BANDWIDTH_DIVISOR = 100;
	private int BUFFER_COUNT = 1500;
	
	private short PKT_SIZE = 1500;
	private static final int ETHER_HEADER_SIZE = 16;

	private String DSTIF_PROTOCOL = "if0";
	private String DSTIF_ADDRESS  = "0xbbbbbbbbbbbb:0x0f0f";
	
	private boolean dying = false;
	private boolean spew = false;
	
	NetHogDomain()
	{
		super(null);

		try
		{
			this.configure();
		}
		catch(Exception e)
		{
			Node.println("NetHog configuration failed: " + e);
			Node.printStackTrace(e);
		}
	}

	private void configure()
		throws Exception
	{
		String str;
		
		str = Domain.lookupConfig("NetHogConfig", "thread-count");
		if( str != null )
			THREAD_COUNT = Integer.parseInt(str);
		str = Domain.lookupConfig("NetHogConfig", "sleep-time");
		if( str != null )
			SLEEP_TIME = Integer.parseInt(str);
		str = Domain.lookupConfig("NetHogConfig", "rate-divisor");
		if( str != null )
			RATE_DIVISOR = Integer.parseInt(str);
		str = Domain.lookupConfig("NetHogConfig", "packet-size");
		if( str != null )
			PKT_SIZE = Short.parseShort(str);
		str = Domain.lookupConfig("NetHogConfig", "band-divisor");
		if( str != null )
			BANDWIDTH_DIVISOR = Integer.parseInt(str);
		str = Domain.lookupConfig("NetHogConfig", "output-protocol");
		if( str != null )
			DSTIF_PROTOCOL = str;
		str = Domain.lookupConfig("NetHogConfig", "output-address");
		if( str != null )
			DSTIF_ADDRESS = str;
	}

	private void printConfiguration()
	{
		Node.println("NetHogDomain configuration:");
		Node.println("  Threads\t" + THREAD_COUNT);
		Node.println("  Sleep\t\t" + SLEEP_TIME);
		Node.println("  Rate divisor\t" + RATE_DIVISOR);
		Node.println("  Packet size\t" + PKT_SIZE);
		Node.println("  Buffer count\t" + BUFFER_COUNT);
		Node.println("  Band divisor\t" + BANDWIDTH_DIVISOR);
		Node.println("  Output\t"
			     + DSTIF_PROTOCOL
			     + " "
			     + DSTIF_ADDRESS);
	}

	private OutChannel createOutChannel()
	{
		OutChannel retval;

		try
		{
			retval = OutChannel.create(
				this.domain,
				new NetworkSpec(BANDWIDTH_DIVISOR,
						NetworkSpec.DIVISOR),
				DSTIF_PROTOCOL,
				DSTIF_ADDRESS);
		}
		catch(Exception e)
		{
			Node.println("Unable to create OutChannel...");
			retval = null;
		}
		return retval;
	}

	private class Spewer
		implements Runnable
	{
		private OutChannel outchan;
		private BufferHandle[] buffers;

		public Spewer(OutChannel oc, int bufCt, int bufSize)
		{
			this.outchan = oc;
			this.buffers = new BufferHandle[bufCt];
			for (int i = 0; i < bufCt; i++)
			{
				buffers[i] = BufferHandle.create(0, bufSize);
				/* XXX fill in payload data */
			}
		}
		
		private void sleep(int millis)
		{
			try
			{
				Thread.sleep(millis);
			}
			catch (InterruptedException ie)
			{
				// ignore it;
			}
		}

		public void run()
		{
			int i = 0;
			final int max = buffers.length;
			final BufferHandle[] buffers = this.buffers;
			final OutChannel oc = this.outchan;

			while( !NetHogDomain.this.dying )
			{
				if( NetHogDomain.this.spew )
				{
					int tries = 0;
					
					/*
					 * If the send fails, block for a minimum
					 * timeslice, and try again.
					 */
					while (!oc.send(buffers[i]))
					{
						assert(tries < 5);
						sleep(5);
						tries++;
					}

					// next buffer
					i++;
					
					if (i == max)
						i = 0;
				}
				else
				{
					/*
					 * Be polite and block if
					 * we're not being used.
					 */
					synchronized(NetHogDomain.this)
					{
						while (!NetHogDomain.this.spew)
						{
							try
							{
								NetHogDomain.this.wait();
							}
							catch (InterruptedException ie)
							{
								// xxx
							}
						}
					}
				}
			}
		}
	}

	private void spawnWorkerBees()
	{
		int lpc;

		for( lpc = 0; lpc < THREAD_COUNT; lpc++ )
		{
			this.domain.threadStart(
				new Spewer(this.createOutChannel(),
					   BUFFER_COUNT,
					   PKT_SIZE - ETHER_HEADER_SIZE),
				"Spewer-" + lpc);
		}
	}
	
	void doRun(CommSpaceElement[] args)
	{
		try
		{
			this.configure();
			this.printConfiguration();
		}
		catch(Exception e)
		{
			Node.println("NetHog configuration failed: " + e);
			Node.printStackTrace(e);
		}

		this.spawnWorkerBees();
	}
	
	void doStartMisbehaving()
	{
		Node.println("NetHog being bad.");
		synchronized(this)
		{
			this.spew = true;
			this.notify();
		}
		
		return;
	}
	
	void doStopMisbehaving()
	{
		Node.println("NetHog not being bad.");
		synchronized(this)
		{
			this.spew = false;
			this.notify();
		}
		return;
	}
	
	PACKAGE void kick(Domain pdomain)
	{
		int memSize = THREAD_COUNT * (20 * 1024) + (2 * 1024 * 1024);
		int rate = ThreadPoolSpec.getMaxRate() / RATE_DIVISOR; // XXX
		
		super.kick(pdomain,
			   new ThreadPoolSpec(THREAD_COUNT + 1,
					      20 * 1024,
					      rate,
					      null),
			   new MemSpec(memSize));
	}
}

