/*
 * SubDomain.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_TESTS_PKG.demo;

import JANOS_NODEOS_PKG.Domain;
import JANOS_NODEOS_PKG.Node;
import JANOS_NODEOS_PKG.DomainRun;
import JANOS_NODEOS_PKG.MemSpec;
import JANOS_NODEOS_PKG.DomainStats;
import JANOS_NODEOS_PKG.ChanStats;
import JANOS_NODEOS_PKG.DomainHandle;
import JANOS_NODEOS_PKG.Credentials;
import JANOS_NODEOS_PKG.DomainStartSpec;
import JANOS_NODEOS_PKG.ThreadPoolSpec;
import JVM_COMMSPACE_PKG.CommSpaceHandle;
import JVM_COMMSPACE_PKG.CommSpaceElement;

/**
 * XXX this class has two halves: controller and controllee.
 * They probably should be separated into separate classes.
 */
abstract class SubDomain
	extends DomainRun
{
	protected Domain domain;
	protected DomainHandle child;

	/**
	 * [0] is the command channel (for the controller to
	 * send commmands to the controllee).
	 * [1..] are the initial arguments.
	 */
	private CommSpaceHandle csh;

	private final int COMMAND_CHANNEL = 0;
	private final int ARGS_CHANNEL = 1; 

	private final Integer COMMAND_EXIT = new Integer(11);
	private final Integer COMMAND_START_MISBEHAVING = new Integer(12);
	private final Integer COMMAND_STOP_MISBEHAVING = new Integer(13);

	SubDomain()
	{
		this.csh = null;
	}

	SubDomain(CommSpaceElement[] args)
	{
		int len;
		if (args != null)
			len = 1 + args.length;
		else
			len = 1;

		csh = CommSpaceHandle.newCommSpace(len);

		if (args != null)
		{
			for (int i = 0; i < args.length; i++)
			{
				csh.write(ARGS_CHANNEL + i, args[i]);
			}
		}
	}

	protected void run(Domain domain,
			   String ignored_args[],
			   CommSpaceHandle csh)
	{
		CommSpaceElement[] args = null;

		assert(ignored_args.length == 0);
		this.domain = domain;
		this.csh = csh;
		
		if (csh.size() > 1)
		{
			args = new CommSpaceElement[csh.size()-1];
			for(int i = 0; i < args.length; i++)
			{
				args[i] = csh.read(ARGS_CHANNEL+i);
			}
		}

		startCommandHandler();

		doRun(args);
	}

	private void startCommandHandler()
	{
		final SubDomain dsf = this;
		Thread t = new Thread() {
			public void run() {
				dsf.commandHandler();
			}
		};
		t.start();
	}

	private void commandHandler()
	{
		try
		{
			while (true)
			{
				try
				{
					CommSpaceElement el;
					el = this.csh.waitToTake(COMMAND_CHANNEL);
					handleCommand(el.getInteger());
				}
				catch(Throwable th)
				{
					Thread.sleep(1000);
				}
			}
		}
		catch (InterruptedException ie)
		{
			Node.println("Command handler for " +this.getClass().getName()+ " blew up: " +ie);
			this.domain.destroy();
		}
	}

	private void handleCommand(Integer command)
	{
		if (command.equals(COMMAND_EXIT))
			// xxx
			this.domain.destroy();
		else if (command.equals(COMMAND_START_MISBEHAVING))
			doStartMisbehaving();
		else if (command.equals(COMMAND_STOP_MISBEHAVING))
			doStopMisbehaving();
	}

	abstract void doRun(CommSpaceElement[] args);
	abstract void doStartMisbehaving();
	abstract void doStopMisbehaving();
		
	// CONTROLLER INTERFACES

	private void sendCommand(Integer cmd)
	{
		this.csh.write(COMMAND_CHANNEL, new CommSpaceElement(cmd));
	}

	void startMisbehaving()
	{
		sendCommand(COMMAND_START_MISBEHAVING);
	}
	
	void stopMisbehaving()
	{
		sendCommand(COMMAND_STOP_MISBEHAVING);
	}

	PACKAGE DomainHandle getHandle()
	{
		return this.child;
	}

	private ChanStats cs = new ChanStats();
	private DomainStats fs = new DomainStats();

	public void printStats()
	{
		DomainHandle fh = this.child;
		int lpc;

		fh.getStats(fs);
		Node.println("Domain `"
			     + this.getClass().getName()
			     + "' - "
			     + fs);
		for( lpc = 0; fh.getInChannelStats(lpc, cs); lpc++ )
		{
			Node.println("  InChannel "
				     + lpc
				     + ": "
				     + cs);
		}
		for( lpc = 0; fh.getOutChannelStats(lpc, cs); lpc++ )
		{
			Node.println("  OutChannel "
				     + lpc
				     + ": "
				     + cs);
		}
	}

	PACKAGE void kick(Domain pdomain, ThreadPoolSpec tps, MemSpec ms)
	{
		String cName = this.getClass().getName();
		
		this.child = pdomain.newDomain(
			new Credentials("Flux Demo Domain:  " + cName),
			tps,
			ms,
			new DomainStartSpec(cName,
					  new String[0],/*XXX this is bogus */
					  this.csh),
			/*termhandler*/null);
	}
	
	PACKAGE void kick(Domain pdomain, ThreadPoolSpec tps)
	{
		this.kick(pdomain, tps, new MemSpec());
	}

	PACKAGE void kick(Domain pdomain, MemSpec ms)
	{
		this.kick(pdomain, new ThreadPoolSpec(1, 32 * 1024, null), ms);
	}
	
	PACKAGE void kick(Domain pdomain)
	{
		this.kick(pdomain,
			  new ThreadPoolSpec(1, 32 * 1024, null),
			  new MemSpec());
	}
}

