/*
 * BufferHandleTester.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package JANOS_TESTS_NODEOS_PKG;

import JANOS_TESTS_PKG.*;
import JANOS_NODEOS_PKG.*;

/**
 * XXX Add more tests for interaction between reset() and
 * other operations.
 */
public class BufferHandleTester
	extends Tester
{
	/**
	 * Turn on and recompile to get a trace to a problem.
	 */
	private static boolean VERBOSE = false;

	private static void verbose(String msg)
	{
		if (VERBOSE)
			Node.println(msg);
	}

	/**
	 * Generate some random byte array for testing
	 *
	 * @param length The length of the byte array to construct
	 * @return The a byte array of the given length with random element
	 * values
	 */
	private byte[] genByteArray(int length)
	{
		byte retval[] = new byte[length];
		int lpc;

		for( lpc = 0; lpc < length; lpc++ )
		{
			retval[lpc] = (byte)(Math.random() * 10);
		}
		return retval;
	}

	/**
	 * Compare a byte array against a subsection of another one
	 *
	 * @param ba1 The subarray to test for
	 * @param offset The offset into the larger byte array
	 * @param ba2 The super byte array
	 * @return True if the subarray matched the section in the larger array
	 */
	private boolean cmpByteArray(byte ba1[], int offset, byte ba2[])
	{
		boolean retval = true;
		int lpc;
		
		for( lpc = 0; (lpc < ba1.length) && retval; lpc++ )
		{
			if( ba1[lpc] != ba2[offset + lpc] )
			{
				retval = false;
			}
		}
		return retval;
	}

	public void testUnit() throws TestFailedException
	{
		BufferHandle bh[] = new BufferHandle[2];
		BufferHandle bhCurr;
		int lpc, count;

		this.enterSection(ENV_NONE, "Constructors");
		bh[0] = BufferHandle.create(0,1);

		int badOffsets[] =
		{
			-1,
			1,
			Integer.MAX_VALUE
		};
		
		verbose(" bad offset tests");
		for( lpc = 0; lpc < badOffsets.length; lpc++ )
		{
			try
			{
				verbose("  bh[" +badOffsets[lpc]+ "]");
				BufferHandle.create(bh[0],
						    badOffsets[lpc],
						    1);
				throw new TestFailedException("offset "
							      +badOffsets[lpc]+
							      " didn't fail on buffer allocate.",
							      this);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
				; /* Nothing */
			}
		}

		int badLengths[] =
		{
			-1,
			2,
			Integer.MAX_VALUE
		};

		verbose(" bad length tests");
		for( lpc = 0; lpc < badLengths.length; lpc++ )
		{
			try
			{
				verbose("  bh[0.."+badLengths[lpc]+")");
				BufferHandle.create(bh[0],
						    0,
						    badLengths[lpc]);
				throw new TestFailedException(this);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
				if( badLengths[lpc] < 0 )
					throw new TestFailedException(" length of " +badLengths[lpc],
								      this);
			}
			catch(NegativeArraySizeException e)
			{
				if( badLengths[lpc] >= 0 )
					throw new TestFailedException(" length of " +badLengths[lpc],
								      this);
			}
		}
		
		this.enterSection(ENV_NONE, "accessor");
		verbose("check getLength()");
		this.test("getLength() == 1", bh[0].getLength() == 1);
		verbose("check getByte(0)");
		this.test("getByte(0) == an int", bh[0].getByte(0) <= 0 || bh[0].getByte(0) >= 0);

		this.enterSection(ENV_NONE, "buffer operations");

		final int bufferSize = 1024;
		final byte byteArray[] = this.genByteArray(bufferSize);
		byte byteDest[];

		/* Get new buffer and reference it from another handle */
		bh[0].newBuffer(0, bufferSize);
		bh[1] = BufferHandle.create(bh[0]);
		verbose("check bh[0] length is correct");
		this.test("bh[0] length is " +bh[0].getLength()+ " expecting " +bufferSize,
			  bh[0].getLength() == bufferSize);
		verbose("check bh[1] length is correct");
		this.test("bh[1] length is " +bh[1].getLength()+ " expecting " +bufferSize,
			  bh[1].getLength() == bufferSize);

		/* Set/get the bytes */
		bh[0].setBytes(0, byteArray, 0, byteArray.length);
		byteDest = new byte[byteArray.length];
		bh[0].getBytes(0, byteDest, 0, byteDest.length);
		this.test(this.cmpByteArray(byteDest, 0, byteArray));
		bh[1].getBytes(0, byteDest, 0, byteDest.length);
		this.test(this.cmpByteArray(byteDest, 0, byteArray));
		byteDest = new byte[byteArray.length / 2];
		bh[0].getBytes(byteArray.length / 4, byteDest,
			       0, byteDest.length);
		this.test(this.cmpByteArray(byteDest,
					    byteArray.length / 4,
					    byteArray));
		bh[1].getBytes(byteArray.length / 4, byteDest,
			       0, byteDest.length);
		this.test(this.cmpByteArray(byteDest,
					    byteArray.length / 4,
					    byteArray));

		verbose("byte order tests");
		try
		{
			/* Make sure ordering is correct */
			bh[0].setByte(0, (byte)1);
			bh[0].setByte(1, (byte)2);
			bh[0].setByte(2, (byte)3);
			bh[0].setByte(3, (byte)4);
			this.test(bh[0].getByte(0) == 1);
			this.test(bh[0].getByte(1) == 2);
			this.test(bh[0].getByte(2) == 3);
			this.test(bh[0].getByte(3) == 4);
			this.test(bh[0].getInt(0) == 0x01020304);
			bh[0].setInt(0, 0x04030201);
			this.test(bh[0].getInt(0) == 0x04030201);
			this.test(bh[0].getByte(0) == 4);
			this.test(bh[0].getByte(1) == 3);
			this.test(bh[0].getByte(2) == 2);
			this.test(bh[0].getByte(3) == 1);
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			throw new TestFailedException(this);
		}

		int illegalOffsets[] =
		{
			-1,
			bufferSize,
			java.lang.Integer.MAX_VALUE,
			java.lang.Integer.MIN_VALUE
		};

		/* Check for illegal offsets */
		verbose(" illegal offset tests");
		for( lpc = 0; lpc < illegalOffsets.length; lpc++ )
		{
			try
			{
				verbose("  byte offset");
				bh[0].setByte(illegalOffsets[lpc], (byte)0);
				throw new TestFailedException(this);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
				; /* Nothing */
			}
			try
			{
				verbose("  int offset");
				bh[0].setInt(illegalOffsets[lpc], 0);
				throw new TestFailedException(this);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
				; /* Nothing */
			}
		}
		try
		{
			verbose("  int offset - 3");
			bh[0].setInt(bufferSize - 3, 0);
			throw new TestFailedException(this);
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			; /* Nothing */
		}

		verbose("  bh[0].reset()");
		bh[0].reset();
		this.test(bh[0].getLength() == 0);
		try
		{
			bh[0].setByte(0, (byte)0);
			throw new TestFailedException(this);
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			; /* Nothing */
		}
		catch(Exception e)
		{
			// XXX this is an unexpected exception
		}

		bh[0].cloneHandle(bh[1]);
		this.test(bh[1].getLength() == bufferSize);
	}
}
