/*
 * InChannelTester.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package JANOS_TESTS_NODEOS_PKG;

import java.net.*;
import java.io.IOException;
import JANOS_TESTS_PKG.*;
import JANOS_NODEOS_PKG.*;
import JANOS_UTIL_PKG.SimpleDemultiplexKey;

public class InChannelTester extends Tester
{
	/**
	 * Keys for testing
	 */
	private final byte key[][] =
	{
		{ 0, 1, 2, 4 },
		{ 8, 16, 32, 64, 127, -128, -64 },
		{ 0 },
		{ 8, 16 },
		{ 5, 4, 3, 2, 1 }
	};
	private final InetAddress localhost;
	private final int port = 30000;
	private DatagramSocket socket;
	private final int packets = 100;
	private int packetsSent[] = new int[key.length];
	private int packetsLost = 0;
	private int packetsDelivered[] = new int[key.length];

	public InChannelTester()
		throws IOException
	{
		this.localhost = InetAddress.getLocalHost();
		this.socket = new DatagramSocket();
	}
	
	private void sendPacket(byte data[])
	{
		try
		{
			DatagramPacket dp = new DatagramPacket(data,
							       data.length,
							       this.localhost,
							       this.port);
			
			this.socket.send(dp);
		}
		catch(IOException e)
		{
			this.packetsLost++;
		}
	}

	private class Receiver implements DeliverFunc
	{
		private final int key;
		PACKAGE InChannel ic;
		
		public Receiver(int key)
		{
			this.key = key;
		}

		public void receive(BufferHandle packet, int payloadOffset)
		{
			InChannelTester.this.packetsDelivered[key]++;

			// Recycle the buffer
			ic.addBuffer(packet);
		}
	}
	
	public void testUnit() throws TestFailedException
	{
		Domain domain = Domain.currentDomain();
		SimpleDemultiplexKey sdk;
		InChannelHandle ich[];
		InChannel ics[];
		Receiver df;
		int maxKeySize;
		int lpc;

		this.enterSection(ENV_NONE, "Constructor");
		{
			DemultiplexKey dk;

			dk = new DemultiplexKey(domain);
			df = new Receiver(0);
			try
			{
				InChannel.create(domain,
						 dk,
						 "if/ipv4/udp",
						 "-1",
						 df);
				throw new TestFailedException(this);
			}
			catch(PacketAddressFormatException e)
			{
			}
			catch(Exception e)
			{
				throw new TestFailedException(this);
			}
			try
			{
				InChannel.create(domain,
						 dk,
						 "if/ipv4/udp",
						 "foo",
						 df);
				throw new TestFailedException(this);
			}
			catch(PacketAddressFormatException e)
			{
			}
			catch(Exception e)
			{
				Node.printStackTrace(e);
				throw new TestFailedException(this);
			}
		}
		this.enterSection(ENV_NONE, "Packet Handling");
		ich = new InChannelHandle[key.length];
		ics = new InChannel[key.length];
		maxKeySize = 0;
		/* Make some new channels and receiver threads */
		try
		{
			for( lpc = 0; lpc < key.length; lpc++ )
			{
				InChannel ic;
				
				maxKeySize = Math.max(maxKeySize, key[lpc].length);
				sdk = new SimpleDemultiplexKey(Domain.currentDomain(),
							       0,
							       key[lpc]);
				df = new Receiver(lpc);
				ic = InChannel.create(
					domain,
					sdk,
					"if/ipv4/udp",
					"*:" + port + ":" + this.localhost.getHostName(),
					df);
				ic.addBuffers(10);
				df.ic = ic;
				ics[lpc] = ic;
				ich[lpc] = new InChannelHandle(ic);
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
			throw new TestFailedException(this);
		}
		/* Send some packets through and make sure they're received */
		try
		{
			BufferHandle bh;
			int packetsLeft;
			int currKey;

			bh = BufferHandle.create();
			bh.newBuffer(0,maxKeySize);
			for(packetsLeft = packets;
			    packetsLeft > 0;
			    packetsLeft--)
			{
				currKey = packetsLeft % key.length;
				this.sendPacket(this.key[currKey]);
				this.packetsSent[currKey]++;
				bh.setBytes(0, this.key[currKey],
					    0, this.key[currKey].length);
				this.test(ich[currKey].
					  handoff(BufferHandle.create(bh)));
				this.packetsSent[currKey]++;
				java.lang.Thread.sleep(50);
			}
		}
		catch(InterruptedException e)
		{
		}

		ChanStats cs = new ChanStats();
		
		for( lpc = 0; lpc < key.length; lpc++ )
		{
			ics[lpc].getStats(cs);
			if( cs.getInPackets() != this.packetsDelivered[lpc] )
			{
				Node.println("Stats don't jive; delivered="
					     + this.packetsDelivered[lpc]
					     + "; cs="
					     + cs);
			}
			if( this.packetsSent[lpc] !=
			    this.packetsDelivered[lpc] )
			{
				Node.println("Warning: Some packets weren't "
					     + "delivered for key number "
					     + lpc
					     + " although this may just be"
					     + " due to a drop by the OS");
				Node.println("  Packets sent: "
					     + this.packetsSent[lpc]
					     + "  delivered: "
					     + this.packetsDelivered[lpc]);
			}
		}
	}
}
