/*
 * LinkedList.Java
 *
 * Copyright (c) 1999, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_UTIL_PKG;

/**
 * A simple doubly linked list implementation, it exports the node object
 * directly so adding and removing is fast
 */
public class LinkedList
{
	/**
	 * A node for use in doubly linked lists.
	 */
	public static class Node
	{
		/**
		 * The next node in the list
		 */
		private Node succ;
		
		/**
		 * The previous node in the list
		 */
		private Node pred;
		
		/**
		 * Get the next node in the list
		 */
		public final Node getSucc()
		{
			return this.succ;
		}
		
		/**
		 * Get the previous node in the list
		 */
		public final Node getPred()
		{
			return this.pred;
		}
		
		/**
		 * Remove the node from the list
		 */
		public void remove()
		{
			this.pred.succ = this.succ;
			this.succ.pred = this.pred;
			this.succ = null;
			this.pred = null;
		}
		
		/**
		 * Insert a node after this one
		 */
		public void append(Node node)
		{
			this.succ.pred = node;
			node.succ = this.succ;
			node.pred = this;
			this.succ = node;
		}

		/**
		 * Insert a node before this one
		 */
		public void prepend(Node node)
		{
			this.pred.succ = node;
			node.succ = this;
			node.pred = this.pred;
			this.pred = node;
		}
	}

	/**
	 * Points to a dummy node which points to the head of the list
	 */
	private Node head;

	/**
	 * Points to a dummy node which points to the tail of the list
	 */
	private Node tail;

	/**
	 * Create a linked list
	 */
	public LinkedList()
	{
		this.head = new Node();
		this.tail = new Node();
		this.head.succ = this.tail;
		this.tail.pred = this.head;
	}

	/**
	 * Returns true if the list is empty
	 */
	public final boolean isEmpty()
	{
		boolean retval = false;

		if( this.head.succ == this.tail )
			retval = true;
		return( retval );
	}

	/**
	 * Get the head of the list
	 */
	public final Node getHead()
	{
		return this.head.succ;
	}

	/**
	 * Get the tail of the list
	 */
	public final Node getTail()
	{
		return this.tail.pred;
	}
	
	/**
	 * Add a node to the head of the list
	 */
	public void addHead( Node node )
	{
		node.succ = this.head.succ;
		node.pred = this.head;
		this.head.succ.pred = node;
		this.head.succ = node;
	}

	/**
	 * Add a node to the tail of the list
	 */
	public void addTail( Node node )
	{
		node.succ = this.tail;
		node.pred = this.tail.pred;
		this.tail.pred.succ = node;
		this.tail.pred = node;
	}

	/**
	 * Remove the node at the head of the list, returns null if its empty
	 */
	public Node remHead()
	{
		Node retval = null;

		if( !this.isEmpty() )
		{
			retval = this.head.succ;
			retval.remove();
		}
		return( retval );
	}

	/**
	 * Remove the node at the tail of the list, returns null if its empty
	 */
	public Node remTail()
	{
		Node retval = null;

		if( !this.isEmpty() )
		{
			retval = this.tail.pred;
			retval.remove();
		}
		return( retval );
	}

	public String toString()
	{
		String retval = "LinkedList[list={ ";
		Node node;

		node = this.head;
		while( node != null )
		{
			retval += node.toString() + " ";
			node = node.succ;
		}
		retval += "}]";
		return retval;
	}
}
