# genmk.pl
#
# Copyright (c) 1999, 2000, 2001 University of Utah and the Flux Group.
# All rights reserved.
#
# This file is part of the Janos Java NodeOS software.
# 
# Permission to use, copy, modify, distribute, and sell this software and
# its documentation is hereby granted without fee, provided that the above
# copyright notice and this permission/disclaimer notice is retained in all
# copies or modified versions, and that both notices appear in supporting
# documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
# WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
# ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
# 
# Users are requested, but not required, to send to csl-dist@cs.utah.edu any
# improvements that they make and grant Univ. of Utah redistribution rights.
#

#
# Generate a makefile for testing an application with
# various command line arguments, and "failure modes."
#
# Expects two names on the command line, the first is the name of the
# configuration file describing the test.  The second is the name of
# the makefile to generate from that configuration.  This filename
# should end in '.mk'.  This target filename is also used to construct
# the basename of the tests.  So, a target filename of 'crusher.mk'
# creates a file called 'crusher.mk' and all of the targets in
# this makefile are prefixed with 'crusher-'.
#
# Configuration files may contain comments (just like this one).
# The following directives are required:
#  NAME=<name>  declares a new test with name <name>.
#  TIMEOUT=<val> the longest the test engine should wait for the test to
#    to finish.
#  COMMAND=<command string> the command to run
# The following directives are optional:
#  TESTNAME_MODIFIER=<string> an optional string to differenatiate tests.
#  INPUT=<file> an input file to the pipe into stdin.
#  DIFFOPTS=various arguments to be passed to the test differentiator. 
#    (see runtest.pl for the supported arguments)
#
# 
# XXX old commands no longer supported:
#  SCRIPT=<arg>
#  VMARGS=<args> the arguments to pass to the VM (before the class)
#  CLASS=<name> the name of the classfile to pass to Java.
#  ARGS=<args> the arguments to pass to the CLASS.
#
# The generated makefile contains the following targets: (basename is the name 
# of the makefile minus '.mk'.)
#  <basename>-setup: to setup all targets (compile and generate "reference" output)
#  <basename>-run: to run all targets (and compare output to the reference output)
#

$conf = shift || &usage();
$target = shift || &usage();

$basename = $target;
$basename =~ s/\.mk$//;

open(TARGET, ">$target") || die "Cannot open $target: $!\n";

print TARGET "### THIS FILE WAS AUTOMATICALLY GENERATED FROM:\n";
print TARGET "###    $conf\n";
print TARGET "### To rengerate run:\n";
print TARGET "###    'genmk $conf $target'\n";
print TARGET "\n";

$allSetupTargets = '';
$allRunTargets = '';

$testname = '';
$testname_modifier = '';
$command = '';
$diffOpts = '';
$timeout = '';  ### seconds
$inputFile = '';

open(CONF, "$conf") || die "Cannot open $conf: $!\n";

$wroteMultipleRules = 0;

while (<CONF>) {
    chop;
    s/\#.*//;			# throw away comments

    if (/NAME=(.*)/) {
      ## Write out previous test
      &writeRules() if ($testname ne '');
      
      ## Setup the new test (re-set basic variables)
      $testname = "${basename}-$1";
      $testname_modifier = '';
      $command = '';
      $diffOpts = '';
      $timeout = '';
      $inputFile = '';
      
    } elsif (/DIFFOPTS?=(.*)/) {
      &checkBlank($diffOpts, "DIFFOPTS");
      $diffOpts = $1;
      $diffOpts =~ s/\s*//;
      
    } elsif (/TIMEOUT=(.*)/) {
      &checkBlank($timeout, "TIMEOUT");
      $timeout = $1;
      
    } elsif (/INPUT=(.*)/) {
      &checkBlank($inputFile, "INPUT");
      $inputFile = $1;
      
    } elsif (/TESTNAME_MODIFIER=(.*)/) {
      &checkBlank($testname_modifier, "TESTNAME_MODIFIER");
      $testname_modifier = $1;
      
    } elsif (/COMMAND=(.*)/) {
      &checkBlank($command, "COMMAND");
      $command = $1;
      
    } elsif (/^(SCRIPT)=/
	       || /^(CLASS)=/
	       || /^(VMARGS)=/
	       || /^(ARGS)=/) {
      &fatalError("genmk: Error: Deprecated command $1.\n");

    } elsif (/^\s*$/) {
      # Skip blank lines

    } else {
      &fatalError("genmk: Error: Unknown command line: $_\n");
    }
}

## write out the last test's rules
## (Note: 'wroteMultipleRules' is set to 1 if writeRules() is called.)
if ($wroteMultipleRules == 0) {
  $testname="${basename}";
}

&writeRules();

## Write trailer
print TARGET "### Make the FORCE target do nothing\n";
print TARGET "FORCE:\n\n";

print TARGET "### Make all of the fake targets in this file phony.\n";
print TARGET ".PHONY: FORCE $allSetupTargets $allRunTargets $basename-setup $basename-run\n\n";

if ($wroteMultipleRules == 0) {
  print TARGET "### Define entrypoints that will run all the tests in this makefile\n";
  print TARGET "$basename-setup: $allSetupTargets\n\n";
  print TARGET "$basename-run: $allRunTargets\n\n";
  print TARGET "$basename-printcommand: $allPrintTargets\n";
}

close(TARGET);
close(CONF); 

# END

### ---------------------------------------------

sub writeRules {
    local($ref) = "\$(REFOUTPUT_DIR)/${testname}${testname_modifier}.\$(REFOUTPUT_SUFFIX)";
    local($diffOpt) = '';
    local($timeoutOpt) = '';
    local($inputOpt) = '';

    if ($command eq '') {
      &fatalError("genmk: ERROR: No command defined for test ${testname}.\n");
    }

    ## Cleanup command
    $command = join(' ', split(' ', $command));
    $command = "\'${command}\'";

    $diffOpt = "--diffopt=$diffOpts" if ($diffOpts ne '');
    $timeoutOpt = "--timeout=$timeout" if ($timeout ne '');
    $inputOpt = "--input=$inputFile" if ($inputFile ne '');

    $wroteMultipleRules = 1;

    $allSetupTargets .= " $testname-setup";
    $allRunTargets .= " $testname-run";
    $allPrintTargets .= " $testname-printcommand";

    print TARGET "TESTCASES+= $testname\n\n";
    print TARGET "### run and setup rules for $testname\n";
    ## Setup rule
    print TARGET "$testname-setup: FORCE\n";
    print TARGET "\trm -f $ref \n";
    print TARGET "\techo \"Generating reference output for ${testname}${testname_modifier}\";\n";
    print TARGET "\t-\$(TEST_RUNNER) \$(TEST_RUNNER_ARGS) $inputOpt $diffOpt $timeoutOpt --setup $command $ref \n";
    print TARGET "\n";

    ## Debugging rule for finding the commands
    print TARGET "$testname-printcommand: FORCE\n";
    print TARGET "\techo \"TEST: ${testname}${testname_modifier} is:\";\n";
    print TARGET "\techo \"  $command\";\n";
    print TARGET "\n";

    ## Run rule
    print TARGET "$testname-run: $ref\n";
    print TARGET "\techo \"TEST: ${testname}${testname_modifier}.\";\n";
    print TARGET "\t-\$(TEST_RUNNER) \$(TEST_RUNNER_ARGS) $inputOpt $diffOpt $timeoutOpt --logfile=\$(LOGFILE) --compare $command $ref \n";
    #print TARGET "\techo \" \";\n";
    print TARGET "\n";
}

sub checkBlank {
  local($blank, $varname) = @_;

  if ($blank ne '') {
    &fatalError("genmk: WARNING: duplicate use of $varname in conf file.  Using new value.\n");
  }
}

sub fatalError {
  local($msg) = shift;

  print STDERR "$msg";
  
  ## Remove target file
  close(TARGET);
  unlink($target);

  exit 1;
}

sub usage {
    print STDERR "Usage: genmk <input .conf file> <output .mk file>\n";
    exit 2;
}

# eof
